//
//  lcng.c
//
//  Copyright (c) 2014 Paul Nicholson
//  All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//  1. Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//  
//  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

//
//    Authors:  Pieter-Tjerk de Boer PA3FWM, Enschede, Netherlands;
//              Paul Nicholson, Todmorden, UK;
//
//  Home page:  http://abelian.org/lcnetgen/
//    Contact:  lcng01@abelian.org
//

#define VERSION "0.4e"

//
//  This code uses C99 with GNU extensions.  Compile with -std=gnu99.
//

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <complex.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/resource.h>

#define FALSE 0
#define TRUE (!FALSE)

//
//  Some physical constants.
//

#define MU (4 * M_PI * 1.0E-7)
#define EPSILON 8.854187817e-12

//
//  Definitions which set the spatial resolution of the capacitance
//  determination. The numbers specify approximately how many elementary
//  charge tiles will be used to represent each type of component.
//

#define NG_DISC 500
#define NG_RECT 500
#define NG_TOROID 500
#define NG_SPHERE 200
#define NG_CYL 200
#define NG_COIL 200
#define NG_WIRE 200
#define NG_PLATE 500
#define NG_RING 200

//
//  Flags set by command line options
//

static int VFLAG = 0;     // Verbosity level, incremented by each -v option

// Algorithms for solving for charge distribution
#define ISM_CG    1   // Conjugate gradient method
#define ISM_GE    2   // Gaussian elimination
#define ISM_GMRES 3   // Generalised minimum residuals method

static int is_method = ISM_GMRES;  // Set by -m options

// Algorithms for calculating potential distribution
#define CPM_DIR   1   // Direct calculation
#define CPM_FMM   2   // Fast multipole method

static int cp_method = CPM_DIR;

// Output options
static int OPT_spice = FALSE;    // TRUE if -o spice is given
static int OPT_list = FALSE;     // TRUE if -o list is given
static int OPT_tiles = FALSE;    // TRUE if -o tiles is given
static int OPT_povray = FALSE;   // TRUE if -o povray is given
static int OPT_gnuplot = FALSE;  // TRUE if -o gnuplot is given
static int OPT_geda = FALSE;     // TRUE if -o geda is given

// Diagnostic options
static int DIAG_bsag = FALSE;   // Diagnostic test of BSAG subroutine
static int DIAG_fmm = FALSE;    // Diagnostic test of FMM coefficients

static int FLAG_NC = FALSE;             // Skip capacitance calculation if TRUE
static int FLAG_NL = FALSE;              // Skip inductance calculation if TRUE
static int FLAG_NR = FALSE;              // Skip resistance calculation if TRUE

static int FLAG_RDC = FALSE;      // TRUE if -r dc is given

static double tile_scale = 1;     // Set by -t option.  Scales the number of
                                  // tiles of all the components

static int FLAG_PRECON = FALSE;                             // Set by -m precon

static int MXP = 6;                 // Number of degrees of spherical harmonics
                                    // Set by -m degree= option

typedef float PMTYPE;                         // Data type for potential matrix
typedef double QVTYPE;                  // Data type for charges and potentials

typedef complex float CXTYPE;  // Data type for multipole fields & coefficients
                               // Todo: switch to real coefficients

#define MAXCPU 64                          // About the most we can get on a PC
static int ncpu = 1;                 // Number of CPUs to use. Set by -c option

#define CCTOL 0.01               // Completion tolerance for the norm of charge
                                 // calculation residual

///////////////////////////////////////////////////////////////////////////////
//  Utilities                                                                //
///////////////////////////////////////////////////////////////////////////////

//
//  Output a message, if the verbosity level is high enough.
//

static pthread_mutex_t w_lock = PTHREAD_MUTEX_INITIALIZER;

static void __attribute__ ((format (printf, 2, 3)))
            logmsg( int level, char *format, ...)
{
   if( VFLAG < level) return;

   pthread_mutex_lock( &w_lock);

   va_list va;
   va_start( va, format); vfprintf( stderr, format, va); va_end( va); 

   fprintf( stderr, "\n");
   fflush( stderr);

   pthread_mutex_unlock( &w_lock);
}

//
//  Routine to be called when a fatal error occurs.
//  Outputs the message out on stderr before calling exit().
//

static void __attribute__ ((format (printf, 1, 2)))
            __attribute__ ((noreturn)) bailout( char *format, ...)
{
   if( format)
   {
      va_list va;
      char temp[ 500];

      va_start( va, format); vsprintf( temp, format, va); va_end( va);
      logmsg( -1, "bailout: %s", temp);
   }

   exit( 1);
}

//
//  Allocate a block of zeroed memory.
//

static void * __attribute__ ((malloc)) safe_malloc( size_t size)
{
   void *p = malloc( size);
 
   if( !p) bailout( "not enough memory [%zd]", size);

   memset( p, 0, size);
   return p;
}

#define allocate( A) ((A *) safe_malloc( sizeof( A)));
#define array( N, A) ((A *) safe_malloc( (N) * sizeof( A)))

static void *malloc2D( size_t N, int esize)
{
   char *p = safe_malloc( N * N * esize);
   char *r = safe_malloc( N * sizeof( void *));

   for( size_t i = 0; i < N; i++, p += N * esize) ((void **) r)[i] = p;
   return r; 
}

#define array2D( N, A) ((A **) malloc2D( (N), sizeof( A)))

static void free2D( void *p)
{
   char **q = p;
   free( q[0]);
   free( p);
}

#if 0  // Only needed for testing
static double lcng_time( void)
{
   struct rusage r;
   if( getrusage( RUSAGE_SELF, &r)) bailout( "rusage, %s", strerror( errno));

   return r.ru_utime.tv_sec + 1e-6 * r.ru_utime.tv_usec +
          r.ru_stime.tv_sec + 1e-6 * r.ru_stime.tv_usec;
}
#endif

//
//  Integer min and max.
//

static inline int imax( int a, int b)
{
   return a > b ? a : b;
}

static inline int imin( int a, int b)
{
   return a < b ? a : b;
}

///////////////////////////////////////////////////////////////////////////////
//  Components & Electrodes                                                  //
///////////////////////////////////////////////////////////////////////////////

struct AXIS
{
   double pa;       // Polar angle, angle between zenith (+z) and axis, radians
   double azim;        // Azimuth, radians between +x and projection of axis on
                       // xy plane, +y = +pi/2.
};

struct POINT        // For points and vectors
{
   double x, y, z;
};

static struct POINT origin = { 0, 0, 0};

//
// Component types
//

#define CT_TOROID 1
struct TOROID
{
   struct POINT *center;
   double outer_radius;
   double inner_radius;
   struct AXIS *axis;
   int nt;
};

#define CT_SPHERE 2
struct SPHERE
{
   struct POINT *center;
   double radius;
   int nt;                                         // Requested number of tiles
};

#define CT_DISC 3
struct DISC
{
   struct POINT *center;
   double radius1;
   double radius2;
   struct AXIS *axis;
   int nt;                                         // Requested number of tiles
   int uniform;                               // TRUE to request uniform tiling
};

#define CT_CYL 4
struct CYL
{
   struct POINT *end1;
   struct POINT *end2;
   struct AXIS *axis;
   double radius1, radius2;
   double length;                                               // Axial length
   int nt;                                         // Requested number of tiles
   int uniform;                               // TRUE to request uniform tiling
};

#define CT_WIRE 5
struct WIRE
{
   struct POINT *end1;
   struct POINT *end2;
   struct AXIS *axis;
   double length;   // Axial length
   double wirad;
   int nt;                                         // Requested number of tiles
   int uniform;                               // TRUE to request uniform tiling
   struct WIRE *next;                  // Next wire if this is a polygonal loop
};

#define CT_RECT 6
struct RECT
{
   struct POINT *center;
   struct POINT *edge1;
   struct POINT *edge2;
   int nt;                                         // Requested number of tiles
   int uniform;                               // TRUE to request uniform tiling
};

#define CT_PLATE 7
struct PLATE
{
   struct POINT **points;
   int npoints;
   int nt;
   int uniform;
};

#define CT_RING 8
struct RING
{
   struct POINT *center;
   struct AXIS *axis;
   double radius;
   double wirad;
   int nt;
};

struct COMPONENT {
   int type;                                                     // One of CT_*
   struct ELECTRODE *electrode;

   union {
      struct SPHERE sphere;
      struct TOROID toroid;
      struct CYL cyl;
      struct DISC disc;
      struct WIRE wire;
      struct RECT rect;
      struct PLATE plate;
      struct RING ring;
  } u;

  struct COMPONENT *next;
};

struct COIL
{
   char *name;
   int ident;

   struct POINT *end1;                        // Center of coil at starting end
   struct POINT *end2;
   struct AXIS *axis;
   double radius1, radius2;                             // Radius at end1, end2

   struct POINT **end1_poly;   // List of points defining cross-section at end1
   struct POINT **end2_poly;   // List of points defining cross-section at end2
   int npoints;                               // Number of cross-section points

   double wirad;                                // Radius of conductor - metres
   double turns;

   double length;                                               // Axial length
   int loop;                                     // TRUE if this is a loop coil

   double *taps;                              // Array of taps, turns from end1
   int ntaps;                                                 // Number of taps
   int *tsect;                               // Section where each tap connects

   double dphi;                                   // BSAG integration step size

   struct COIL *image;                             // Mirror image of this coil

   struct CSECT {
      int unit;                                           // 0 ... coil->Ns - 1
      struct POINT end1, end2;
      struct POINT *end1_poly; // List of points defining cross-section at end1
      struct POINT *end2_poly; // List of points defining cross-section at end2
      struct AXIS axis;
      double length;
      double turns;                          // Number of turns in this section
      double start;                         // Turns at start (end1) of section
      double radius1, radius2;         // If this is has circular cross-section

      struct ELECTRODE *e;               // Electrode representing this section
      struct CSECT *image;             // Image of this section in ground plane
   } *sections;                          // Information about each coil section

   int Ns;                                   // Number of sections for modeling
   int nt;                                       // Number of tiles per section

   struct COIL *next;                           // Next in linked list of coils
};

//
//  A structure for each independent electrode.
//

struct ELECTRODE 
{
   char *name;
   int ident;

   QVTYPE volts;                             // The potential on this electrode
   QVTYPE charge;                         // The total charge on this electrode

   struct COMPONENT *components;           // Linked list of all the components

   struct COIL *coil;                    // If this electrode is a coil section
   int unit;                           // Unit number if this is a coil section

   struct ELECTRODE *next;            // To next in a linked list of electrodes
   struct ELECTRODE *image;
};

//
//  The electrodes is decomposed into a large number of tiles.  Each tile has
//  a back pointer to the electrode that owns it.
//

struct TILE
{
   struct POINT c;                                            // Center of tile
   struct POINT ci;                                   // Center of ground image 

   PMTYPE spc;                                    // Self-potential coefficient

   struct POINT ux, uy;               // Unit vectors spanning rectangular tile
   double sx, sy;                 // Extent of the tile along both unit vectors
                           // sx==0 is flag for tile with unknown size or shape
   double area;                           // Area of the TILE, equal to 4*sx*sy

   struct ELECTRODE *electrode;           // The electrode which owns this tile
   int idx;                  // Index of this tile in tile_list[], V[], and Q[]
};

//
//  A ground plane, specified in Hessian normal form.
//

struct GROUND {
   struct POINT *center;                         // Arbitray point on the plane
   struct AXIS  *axis;                                   // Normal to the plane

   struct POINT norm;                 // Normal vector in cartesian coordinates
   double offset;                // x*norm.x + y*norm.y + z*norm.z + offset = 0
};

struct WORLD 
{
   char sysname[100];                    // System name from the input filename

   struct ELECTRODE *electrodes;               // Linked list of all electrodes
   struct COIL *coils;                              // Linked list of all coils

   struct AXIS *Efield;                           // Incident E-field direction
   struct AXIS *Bfield;                           // Incident B-field direction

   struct GROUND *g;                          // If a ground plane is specified
   struct POINT rc;                                   // Centroid of the system
   double msize;                             // Maximum dimension of the system
} WORLD;    

//
//  Create an electrode description structure. Initialise it and add it to
//  the head of the list of electrodes W->electrodes
//
//  If the electrode does not belong to a coil then give it a default name.
//  This might be replaced by the name parsed from the input file.
//

struct ELECTRODE *create_electrode( struct WORLD *W, struct COIL *c)
{
   static int counter = 0;
   struct ELECTRODE *e = allocate( struct ELECTRODE);

   e->next = W->electrodes;  W->electrodes = e;

   e->coil = c;
   if( !c)
   {
      e->ident = ++counter;
      if( asprintf( &e->name, "E%d", e->ident) < 0) bailout( "out of memory");
   }
      
   return e;
}

//
//  Create a component and link it to its parent electrode.
//

struct COMPONENT *create_component( struct ELECTRODE *e, int type)
{
   struct COMPONENT *c = allocate( struct COMPONENT);

   c->next = e->components; e->components = c;
   c->type = type;

   return c;
}

//
//  Create a coil and add it to the list.
//

struct COIL *create_coil( struct WORLD *W)
{
   static int counter = 0;

   struct COIL *c = allocate( struct COIL);
   c->next = W->coils; W->coils = c;

   c->ident = ++counter;
   if( asprintf( &c->name, "L%d", c->ident) < 0) bailout( "out of memory");

   return c;
}
 
//
//  Create and initialise a tile and add it to the tile_list[] array.
//

static int ntiles = 0;                         // Total number of tiles created
static struct TILE **tile_list = NULL;                    // Array of all tiles
static int tiles_alloc = 0;                        // Allocated size of tiles[]

static struct TILE *create_tile( struct ELECTRODE *e,
                                 struct POINT center, QVTYPE spc)
{
   if( isnan( spc)) bailout( "self potential is nan in %s", e->name);
   if( spc <= 0) bailout( "bad self potential %.3e in %s", spc, e->name);

   struct TILE *t = allocate( struct TILE);

   t->electrode = e;
   t->c = center;
   t->idx = ntiles;
   t->spc = spc;

   // Extend the allocation of tile_list[] if necessary, in blocks of 1000

   if( ntiles + 1 > tiles_alloc)
   {
      tiles_alloc += 1000;
      tile_list = realloc( tile_list, sizeof( struct TILE *) * tiles_alloc);
      if( !tile_list) bailout( "out of memory for tiles");
   }

   tile_list[ntiles++] = t;

   return t;
}

///////////////////////////////////////////////////////////////////////////////
//  Expression Evaluation                                                    //
///////////////////////////////////////////////////////////////////////////////

//
//  Input file parser produces postfix expressions represented as a linked
//  list of nodes.
//

// Node types
#define NT_CONST 0                                        // A numeric constant
#define NT_OPER1 1                                          // A unary operator
#define NT_OPER2 2                                         // A binary operator
#define NT_OPER3 3                                       // A tertiary operator
#define NT_DEFN  4                                          // A defined symbol
#define NT_AFCN1 5             // Call an arithmetic function with one argument
#define NT_UNIT  6                                       // A units description

struct NODE
{
   int type;                                // One of the above type indicators
   union
   {
      double val;
      int op;
      struct DEFN *defn;
      double (*func)();
      struct UNIT *unit;
   } u;

   struct NODE *next;                                    // Linked list pointer
};

static struct NODE *alloc_node( int type)
{
   struct NODE *n = allocate( struct NODE);

   n->type = type;
   n->next = NULL;
   return n;
}

//
//  Return a pointer to the last node of an expression.
//

static struct NODE *end_of( struct NODE *n)
{
   while( n->next) n = n->next;
   return n;
}

//
//  A variable definition binds a symbol name to its defining expression.
//

static struct DEFN
{
   char name[100];                            // Symbol name of this definition
   struct NODE *expr;                                // Its defining expression
   struct DEFN *next;                                    // Linked list pointer
}
 *defns, **defn_ep = &defns;                    // A linked list of definitions

static struct DEFN *find_defn( char *name)
{
   struct DEFN *d = defns;
   for( ; d && strcmp( d->name, name); d = d->next);
   return d;
}

static struct DEFN *alloc_defn( char *name)
{
   struct DEFN *d = *defn_ep = allocate( struct DEFN);

   *(defn_ep = &d->next) = 0;
   strcpy( d->name, name);

   return d;
}

//
//  To handle units conversions, we jam a special node into the expression,
//  in order to hold conversion data.
//

struct UNIT
{
   double factor;                         // Factor for conversion to metric
   double (*cvt)(double);                     // Conversion function, if any
   char *symbol;                                            // Symbol string
};

//
//  Take an unscaled postfix expression and prefix it with a units conversion.
//

static struct NODE *scale_expression( struct NODE *n, double factor,
                                      double (*cvt)(), char *symbol)
{
   struct NODE *t = alloc_node( NT_UNIT);
   t->u.unit = allocate( struct UNIT);
   t->u.unit->factor = factor;
   t->u.unit->cvt = cvt;
   t->u.unit->symbol = symbol;
   t->next = n;
   return t;
}

//
//  Extract the units conversion for the given expression.
//

static struct UNIT *extract_units( struct NODE *n)
{
   for( ; n; n = n->next) if( n->type == NT_UNIT) return n->u.unit;
   return 0;
}

//
//  A stack of doubles for expression evaluation, along with a stack pointer.
//

static double estack[1000], *estack_p = estack;

//
//  A recursive function to evaluate an expression.
//

static double evaluate( struct NODE *n)
{
   double d1, d2, d3;

   for( ; n; n = n->next)
   {
      switch( n->type)
      {
         case NT_CONST: // 
                        //  Just push a constant value onto the stack.
                        //

                        *estack_p++ = n->u.val; break;

         case NT_OPER1: //
                        //  Unary operator.
                        //

                        d1 = *--estack_p;

                        switch( n->u.op)
                        {
                           case '+': *estack_p++ = d1;  break;
                           case '-': *estack_p++ = -d1;  break;
                        }
                        break;

         case NT_OPER2: //
                        //  Binary operator. Pop the top two values off the
                        //  stack, apply the appropriate operator to them,
                        //  and push the result back onto the stack.
                        //
 
                        d2 = *--estack_p;
                        d1 = *--estack_p;
               
                        switch( n->u.op)
                        {
                           case '+':  *estack_p++ = d1 + d2; break;
                           case '-':  *estack_p++ = d1 - d2; break;
                           case '*':  *estack_p++ = d1 * d2; break;
                           case '/':  *estack_p++ = d1 / d2; break;
                           case '>':  *estack_p++ = d1 > d2; break;
                           case '<':  *estack_p++ = d1 < d2; break;
                           case '=':  *estack_p++ = d1 == d2; break;
                        }
                        break;  
               
         case NT_OPER3: //
                        //  Tertiary operator. Pop the three values off the
                        //  stack, and push the selected result back on. No
                        //  lazy evaluation here, I'm afraid.
                        //
 
                        d3 = *--estack_p;
                        d2 = *--estack_p;
                        d1 = *--estack_p;
                        *estack_p++ = d1 ? d2 : d3;
                        break; 
               
         case NT_DEFN:  //
                        //  Evaluate the defining expression for the symbol
                        //  and push the result onto the stack.
                        //

                        *estack_p++ = evaluate( n->u.defn->expr);  break;

         case NT_AFCN1: //
                        //  Call an arithmetic function.
                        //

                        d1 = *--estack_p; *estack_p++ = n->u.func(d1);  break;
      }
   }

   //
   //  By now the result is the top item on the stack.
   //

   return *--estack_p;
}

//
//  Evaluate an expression, as above, but then apply the units conversion, if
//  any.
//

static double evaluate_scaled( struct NODE *n)
{
   struct UNIT *unit = extract_units( n);

   if( !unit) return evaluate( n);

   if( unit->cvt) return unit->cvt( evaluate( n));
   else return unit->factor * evaluate( n);
}

//
//  Pre-define some common constants.
//

static void setup_constants( void)
{
   (alloc_defn( "PI")->expr = alloc_node( NT_CONST))->u.val = M_PI;
   (alloc_defn( "MU")->expr = alloc_node( NT_CONST))->u.val = MU;
   (alloc_defn( "EPSILON")->expr = alloc_node( NT_CONST))->u.val = EPSILON;
}

///////////////////////////////////////////////////////////////////////////////
//  R/L/C Storage                                                            //
///////////////////////////////////////////////////////////////////////////////

//
//  Data structures to hold calculated R, L, and C values until they are
//  needed for output.   The store functions need to be thread-safe.
//

//
//  R_STORE: series resistance of each coil section.
//

static struct R_STORE {
   struct COIL *c;
   int s;
   double R;
}
 *Rstore = NULL;

static int Rstore_cnt = 0;                                 // Number of entries
static int Rstore_alloc = 0;                         // Allocated size of array
static pthread_mutex_t Rstore_lock = PTHREAD_MUTEX_INITIALIZER;

static void store_R( struct COIL *c, int s, double R)
{
   pthread_mutex_lock( &Rstore_lock);

   struct R_STORE *p = Rstore;

   for( int i = 0; i < Rstore_cnt; i++, p++)
      if( p->c == c && p->s == s)
      {
         p->R = R;
         pthread_mutex_unlock( &Rstore_lock);
         return;
      }

   if( Rstore_cnt == Rstore_alloc)
   {
      Rstore_alloc += 100;
      Rstore = realloc( Rstore, Rstore_alloc * sizeof( struct R_STORE));
   }

   p = Rstore + Rstore_cnt++;
   p->c = c;
   p->s = s;
   p->R = R;

   pthread_mutex_unlock( &Rstore_lock);
}

static double find_R( struct COIL *c, int s)
{
   struct R_STORE *p = Rstore;

   for( int i = 0; i < Rstore_cnt; i++, p++)
      if( p->c == c && p->s == s) return p->R;

   bailout( "cannot find coil resistance, %s", c->name);
}

//
//  L_STORE: Mutual and self inductance of coil sections.
//
//  Self inductance if c1 == c2 and s1 == s2, otherwise mutual.
//

static struct L_STORE {
   struct COIL *c1;
   struct COIL *c2;
   int s1;
   int s2;
   double L;
}
 *Lstore = NULL;

static int Lstore_cnt = 0;                                 // Number of entries
static int Lstore_alloc = 0;                         // Allocated size of array
static pthread_mutex_t Lstore_lock = PTHREAD_MUTEX_INITIALIZER;

static void store_L( struct COIL *c1, int s1,
                     struct COIL *c2, int s2,
                     double L)
{
   pthread_mutex_lock( &Lstore_lock);

   int i;
   struct L_STORE *p = Lstore;

   for( i = 0; i < Lstore_cnt; i++, p++)
      if( p->c1 == c1 && p->s1 == s1 && p->c2 == c2 && p->s2 == s2 ||
          p->c1 == c2 && p->s1 == s2 && p->c2 == c1 && p->s2 == s1)
         bailout( "duplicate L store");

   if( Lstore_cnt == Lstore_alloc)
   {
      Lstore_alloc += 100;
      Lstore = realloc( Lstore, Lstore_alloc * sizeof( struct L_STORE));
   }

   p = Lstore + Lstore_cnt++;
   p->c1 = c1;
   p->c2 = c2;
   p->s1 = s1;
   p->s2 = s2;
   p->L = L;

   pthread_mutex_unlock( &Lstore_lock);
}

static double find_L( struct COIL *c1, int s1, struct COIL *c2, int s2)
{
   int i;
   struct L_STORE *p = Lstore;

   for( i = 0; i < Lstore_cnt; i++, p++)
      if( p->c1 == c1 && p->s1 == s1 && p->c2 == c2 && p->s2 == s2 ||
          p->c1 == c2 && p->s1 == s2 && p->c2 == c1 && p->s2 == s1) return p->L;

   if( c1 == c2 && s1 == s2)
      bailout( "cannot find self L of %s/%d", c1->name, s1);

   bailout( "cannot find mutual L %s/%d to %s/%d", c1->name, s1, c2->name, s2);
}

//
//  C_STORE: Mutual and self capacitances of electrodes (including coil
//           sections).
//
//  Self capacitance (ie to infinity) if e1 == e2, otherwise mutual.
//

static struct C_STORE {
   struct ELECTRODE *e1;
   struct ELECTRODE *e2;
   double C;
}
 *Cstore = NULL;

static int Cstore_cnt = 0;                                 // Number of entries
static int Cstore_alloc = 0;                         // Allocated size of array
static pthread_mutex_t Cstore_lock = PTHREAD_MUTEX_INITIALIZER;

static void store_C( struct ELECTRODE *e1, struct ELECTRODE *e2, double C)
{
   pthread_mutex_lock( &Cstore_lock);

   struct C_STORE *p = Cstore;

   for( int i = 0; i < Cstore_cnt; i++, p++)
      if( p->e1 == e1 && p->e2 == e2 ||
          p->e1 == e2 && p->e2 == e1)
      {
         p->C = (p->C + C)/2;
         pthread_mutex_unlock( &Cstore_lock);
         return;
      }

   if( Cstore_cnt == Cstore_alloc)
   {
      Cstore_alloc += 100;
      Cstore = realloc( Cstore, Cstore_alloc * sizeof( struct C_STORE));
   }

   p = Cstore + Cstore_cnt++;
   p->e1 = e1;
   p->e2 = e2;
   p->C = C;

   pthread_mutex_unlock( &Cstore_lock);
}

static double find_C( struct ELECTRODE *e1, struct ELECTRODE *e2)
{
   struct C_STORE *p = Cstore;

   for( int i = 0; i < Cstore_cnt; i++, p++)
      if( p->e1 == e1 && p->e2 == e2 ||
          p->e1 == e2 && p->e2 == e1) return p->C;

   if( e2) bailout( "cannot find mutual C, %s to %s", e1->name, e2->name);
   bailout( "cannot find self C, %s", e1->name);
}

//
//  Q_STORE: Charge on each electrode from incident E-field.
//

static struct Q_STORE {
   struct ELECTRODE *e;
   double Q;
}
 *Qstore = NULL;

static int Qstore_cnt = 0;                                 // Number of entries
static int Qstore_alloc = 0;                         // Allocated size of array
static pthread_mutex_t Qstore_lock = PTHREAD_MUTEX_INITIALIZER;

static void store_Q( struct ELECTRODE *e, double Q)
{
   pthread_mutex_lock( &Qstore_lock);

   struct Q_STORE *p = Qstore;

   for( int i = 0; i < Qstore_cnt; i++, p++)
      if( p->e == e) bailout( "duplicate Q store");

   if( Qstore_cnt == Qstore_alloc)
   {
      Qstore_alloc += 100;
      Qstore = realloc( Qstore, Qstore_alloc * sizeof( struct Q_STORE));
   }

   p = Qstore + Qstore_cnt++;
   p->e = e;
   p->Q = Q;
   pthread_mutex_unlock( &Qstore_lock);
}

static double find_Q( struct ELECTRODE *e)
{
   struct Q_STORE *p = Qstore;

   for( int i = 0; i < Qstore_cnt; i++, p++) if( p->e == e) return p->Q;

   bailout( "cannot find induced charge, %s", e->name);
}

//
//  F_STORE: Linked flux from incident magnetic field for each coil section.
//

static struct F_STORE {
   struct COIL *c;
   int s;
   double F;
}
 *Fstore = NULL;

static int Fstore_cnt = 0;                                 // Number of entries
static int Fstore_alloc = 0;                         // Allocated size of array
static pthread_mutex_t Fstore_lock = PTHREAD_MUTEX_INITIALIZER;

static void store_F( struct COIL *c, int s, double F)
{
   pthread_mutex_lock( &Fstore_lock);

   struct F_STORE *p = Fstore;

   for( int i = 0; i < Fstore_cnt; i++, p++)
      if( p->c == c && p->s == s)
      {
         p->F = F;
         pthread_mutex_unlock( &Fstore_lock);
         return;
      }

   if( Fstore_cnt == Fstore_alloc)
   {
      Fstore_alloc += 100;
      Fstore = realloc( Fstore, Fstore_alloc * sizeof( struct F_STORE));
   }

   p = Fstore + Fstore_cnt++;
   p->c = c;
   p->s = s;
   p->F = F;
   pthread_mutex_unlock( &Fstore_lock);
}

static double find_F( struct COIL *c, int s)
{
   int i;
   struct F_STORE *p = Fstore;

   for( i = 0; i < Fstore_cnt; i++, p++)
      if( p->c == c && p->s == s) return p->F;

   bailout( "cannot find flux for L %s/%d", c->name, s);
}

///////////////////////////////////////////////////////////////////////////////
//  Geometric Stuff                                                          //
///////////////////////////////////////////////////////////////////////////////

//
//  Calculate the point p which is at distance d from point o along
//  the axis a.  o may be NULL, indicating origin.
//

static void forepoint( struct AXIS *a, double d,
                       struct POINT *p, struct POINT *o)
{
   // Rotate unit vector (0,0,d) by the polar angle around the y axis
   double x1 = sin(a->pa) * d;
   double z1 = cos(a->pa) * d;

   // Rotate by the azimuth angle around the z axis
   p->x = cos(a->azim) * x1;
   p->y = sin(a->azim) * x1;
   p->z = z1;

   if( o)
   {
      p->x += o->x;
      p->y += o->y;
      p->z += o->z;
   }
}

//
//  Difference between two vectors.
//

static inline struct POINT sub_3D( struct POINT p1, struct POINT p2)
{
   return (struct POINT) { p1.x - p2.x, p1.y - p2.y, p1.z - p2.z };
}

//
//  Sum of two vectors.
//

static inline void sum_3D( struct POINT *sum, struct POINT p)
{
   sum->x += p.x;   sum->y += p.y;   sum->z += p.z;
}

static inline struct POINT add_3D( struct POINT p1, struct POINT p2)
{
   return (struct POINT) { p1.x + p2.x, p1.y + p2.y, p1.z + p2.z };
}

//
//  Scale a vector.
//

static inline void scale_3D( struct POINT *p, double m)
{
   p->x *= m;   p->y *= m;   p->z *= m;
}

//
//  Inner product between two vectors.
//

static inline double inner_prod_3D( struct POINT p1, struct POINT p2)
{
   return p1.x * p2.x + p1.y * p2.y + p1.z * p2.z;
}

//
//  Cross product b x c between two vectors.
//

static inline struct POINT xprod_3D( struct POINT b, struct POINT c)
{
   return (struct POINT) { b.y * c.z - b.z * c.y,
                           b.z * c.x - b.x * c.z,
                           b.x * c.y - b.y * c.x };
}

//
//  Magnitude of a vector.
//

static inline double mag_3D( struct POINT p)
{
   return sqrt( inner_prod_3D( p, p));
}

//
//  Normalise a vector
//

static inline void norm_3D( struct POINT *p)
{
   double d = mag_3D( *p);
   p->x /= d; p->y /= d; p->z /= d;
}

//
//  Distance between two points.
//

static inline double pythagoras(struct POINT p, struct POINT q)
{
   return mag_3D( sub_3D( p, q));
}

//
//  Construct vector in spherical coordinates pointing from p2 to p1.
//

static void to_polar( struct POINT *p1, struct POINT *p2,
                      double *pa, double *azim, double *mag)
{
   double dx = p1->x - p2->x;
   double dy = p1->y - p2->y;
   double dz = p1->z - p2->z;

   double r = sqrt(dx*dx + dy*dy);

   *pa = atan2( r, dz);
   *azim = atan2( dy, dx);
   if( mag) *mag = sqrt(dx*dx + dy*dy + dz*dz);
}

//
//  Rotate the world (well, a point at a time) so that vertical +z moves to
//  axis given by pa, azim.
//

struct POINT rotate( struct POINT p, double pa, double azim)
{
   // Rotate by polar angle first
   struct POINT t = {  cos(pa) * p.x + sin(pa) * p.z,
                       p.y,
                       -sin(pa) * p.x + cos(pa) * p.z
                    };

   // Then rotate in azimuth
   return (struct POINT) {  cos( azim) * t.x - sin(azim) * t.y,
                            sin( azim) * t.x + cos(azim) * t.y,
                            t.z
                         };
}

//
//  Rotate the world so that the axis pa,azim becomes vertical (+z)
//

struct POINT unrotate( struct POINT p, double pa, double azim)
{
   // Undo the azimuth rotation first
   struct POINT t = { cos( -azim) * p.x - sin(-azim) * p.y,
                      sin( -azim) * p.x + cos(-azim) * p.y,
                      p.z };

   // Then the polar angle
   return (struct POINT) { cos(-pa) * t.x + sin(-pa) * t.z,
                           t.y,
                           -sin(-pa) * t.x + cos(-pa) * t.z};
}

//
//  Rotate a point around an axis (which passes through the origin) by a
//  given angle.
//

static void rotate_point( struct POINT *p, struct AXIS axis, double angle)
{
   //  Rotate the point around the origin so that the axis is vertical
   struct POINT v  = unrotate( *p, axis.pa, axis.azim);

   //  Rotate in the xy plane around the +z axis
   struct POINT t = {   v.x * cos(angle) - v.y * sin(angle),
                        v.x * sin(angle) + v.y * cos(angle),
                        v.z };

   //  Rotate back to original
   *p = rotate( t, axis.pa, axis.azim);
}

//
//  Rotate an axis around another axis by a given angle.
//

static void rotate_axis( struct AXIS *a, struct AXIS axis, double angle)
{
   // Unit vector along axis a
   struct POINT p; forepoint( a, 1.0, &p, NULL);

   // Rotate it
   rotate_point( &p, axis, angle);

   to_polar( &p, &origin, &a->pa, &a->azim, NULL);
}

//
//  Work out the coordinates ci of the image of point c in the ground
//  plane.   Return FALSE if point c is below the ground.
//

static int image_point( struct POINT *c, struct POINT *ci)
{
   // Perpendicular distance from point c to ground plane
   double d = inner_prod_3D( WORLD.g->norm, *c) + WORLD.g->offset;
    
   // Position of mirror point
   forepoint( WORLD.g->axis, -2*d, ci, c);

   // A negative 'd' indicates that point c is on the wrong side of the ground
   return d >= 0;
}

//
//  Distance between centers of two tiles.  MINDIST prevents complete failure
//  when two tiles happen to have the same center.   This happens sometimes
//  when two components are touching.  Not a problem so long as they are part
//  of the same electrode.
//
//  Todo: maybe warn if two electrodes are touching
//

#define MINDIST 1e-4            // Hope nobody is modelling anything very small

static inline double tile_distance( struct TILE *d, struct TILE *s)
{
   double r = pythagoras( d->c, s->c);
   if( r < MINDIST) r = MINDIST;
   return r;
}

//
//  Distance to a destination tile from the ground image of a source tile.
//

static inline double tile_mirror_distance( struct TILE *d, struct TILE *s)
{
   double r = pythagoras( d->c, s->ci);
   if( r < MINDIST) r = MINDIST;
   return r;
}

//
//  Axis of a polygon
//

static struct AXIS polygon_axis( struct POINT *poly, int N)
{
   struct POINT va = origin;

   for( int i = 1; i < N - 1; i++)
   {
      struct POINT d1 = sub_3D( poly[i+0], poly[i-1]);
      struct POINT d2 = sub_3D( poly[i+1], poly[i-1]);
      struct POINT xp = xprod_3D( d1, d2);
      sum_3D( &va, xp);
   }

   norm_3D( &va);
   struct AXIS axis;
   to_polar( &va, &origin, &axis.pa, &axis.azim, NULL);
   return axis;
}

//
//  Area of a polygon
//

static double polygon_area( struct AXIS *axis, struct POINT **plist, int N)
{
   double sum = 0;

   struct POINT t = unrotate( *plist[N-1], axis->pa, axis->azim);

   for( int i = 0; i < N; i++)
   {
      struct POINT s = unrotate( *plist[i], axis->pa, axis->azim);
      sum += (t.x + s.x) * (t.y - s.y); 
      t = s;
   }

   return fabs(sum/2);
}

///////////////////////////////////////////////////////////////////////////////
//  Shifts & Rotations                                                       //
///////////////////////////////////////////////////////////////////////////////

//
//  These functions only need to modify the parsed values and values calculated
//  in the parse_ functions.  Should not be called after parsing has finished.
//

static void shift_component( struct COMPONENT *c, struct POINT offset)
{
   switch( c->type)
   {
      case CT_TOROID: sum_3D( c->u.toroid.center, offset);
                      break;
      case CT_SPHERE: sum_3D( c->u.sphere.center, offset);
                      break;
      case CT_DISC:   sum_3D( c->u.disc.center, offset);
                      break;
      case CT_WIRE:   sum_3D( c->u.wire.end1, offset);
                      sum_3D( c->u.wire.end2, offset);
                      break;
      case CT_RECT:   sum_3D( c->u.rect.center, offset);
                      sum_3D( c->u.rect.edge1, offset);
                      sum_3D( c->u.rect.edge2, offset);
                      break;
      case CT_PLATE:  for( int i = 0; i < c->u.plate.npoints; i++)
                         sum_3D( c->u.plate.points[i], offset);
                      break;
      case CT_RING:   sum_3D( c->u.ring.center, offset);
                      break;

      default: bailout( "case not handled");
   }
}

static void shift_coil( struct COIL *c, struct POINT offset)
{
   if( c->end1) sum_3D( c->end1, offset);
   if( c->end2) sum_3D( c->end2, offset);
   for( int i = 0; i < c->npoints; i++)
   {
      sum_3D( c->end1_poly[i], offset);
      sum_3D( c->end2_poly[i], offset);
   }
}

static void rotate_component( struct COMPONENT *c,
                              struct AXIS axis, double angle)
{
   switch( c->type)
   {
      case CT_TOROID: rotate_point( c->u.toroid.center, axis, angle);
                      rotate_axis( c->u.toroid.axis, axis, angle);
                      break;
      case CT_SPHERE: rotate_point( c->u.sphere.center, axis, angle);
                      break;
      case CT_DISC:   rotate_point( c->u.disc.center, axis, angle);
                      rotate_axis( c->u.disc.axis, axis, angle);
                      break;
      case CT_WIRE:   rotate_point( c->u.wire.end1, axis, angle);
                      rotate_point( c->u.wire.end2, axis, angle);
                      rotate_axis( c->u.wire.axis, axis, angle);
                      break;
      case CT_RECT:   rotate_point( c->u.rect.center, axis, angle);
                      rotate_point( c->u.rect.edge1, axis, angle);
                      rotate_point( c->u.rect.edge2, axis, angle);
                      break;
      case CT_PLATE:  for( int i = 0; i < c->u.plate.npoints; i++)
                         rotate_point( c->u.plate.points[i], axis, angle);
                      break;
      case CT_RING:   rotate_point( c->u.ring.center, axis, angle);
                      rotate_axis( c->u.ring.axis, axis, angle);
                      break;

      default: bailout( "case not handled");
   }
}

static void rotate_coil( struct COIL *c, struct AXIS axis, double angle)
{
   if( c->end1) rotate_point( c->end1, axis, angle);
   if( c->end2) rotate_point( c->end2, axis, angle);
   for( int i = 0; i < c->npoints; i++)
   {
      rotate_point( c->end1_poly[i], axis, angle);
      rotate_point( c->end2_poly[i], axis, angle);
   }
   rotate_axis( c->axis, axis, angle);
}

//
//  Translate everything in the world to a new origin given by center.
//

static void shift_world( struct WORLD *W, struct POINT offset)
{
   for( struct ELECTRODE *e = W->electrodes; e; e = e->next)
      for( struct COMPONENT *c = e->components; c; c = c->next)
         shift_component( c, offset);

   for( struct COIL *c = W->coils; c; c = c->next)
      shift_coil( c, offset);

   if( W->g) sum_3D( W->g->center, offset);
}

static void rotate_world( struct WORLD *W, struct AXIS axis, double angle)
{
   for( struct ELECTRODE *e = W->electrodes; e; e = e->next)
      for( struct COMPONENT *c = e->components; c; c = c->next)
         rotate_component( c, axis, angle);

   for( struct COIL *c = W->coils; c; c = c->next)
      rotate_coil( c, axis, angle);

   if( W->g) rotate_axis( W->g->axis, axis, angle);

   if( W->Efield) rotate_axis( W->Efield, axis, angle);
   if( W->Bfield) rotate_axis( W->Bfield, axis, angle);
}

//
//  Move everything in the temporary world WT into the real world W.
//

static void sum_world( struct WORLD *W, struct WORLD *WT)
{
   struct ELECTRODE **ep = &W->electrodes;
   while( *ep) ep = &(*ep)->next;
   *ep = WT->electrodes;

   struct COIL **cp = &W->coils;
   while( *cp) cp = &(*cp)->next;
   *cp = WT->coils;

   if( W->Efield && WT->Efield) bailout( "more than one Efield specified");
   if( WT->Efield) W->Efield = WT->Efield;

   if( W->Bfield && WT->Bfield) bailout( "more than one Bfield specified");
   if( WT->Bfield) W->Bfield = WT->Bfield;

   if( W->g && WT->g) bailout( "more than one ground specified");
   if( WT->g) W->g = WT->g;
}

///////////////////////////////////////////////////////////////////////////////
//  Output File Utilities                                                    //
///////////////////////////////////////////////////////////////////////////////

//
//  Just wrappers around the system calls to trap all the errors.
//  Only one file opened at a time so the name and handle can be global.
//

static char *outfilename;
static FILE *fdo;   

static void output_open( char *ext)
{
   if( asprintf( &outfilename, "%s.%s", WORLD.sysname, ext) < 0)
      bailout( "out of memory");

   fdo = fopen( outfilename, "w");
   if( !fdo) bailout( "cannot open %s, %s", outfilename, strerror( errno));
}

static void __attribute__ ((format (printf, 1, 2))) outputs( char *format, ...)
{
   va_list va;

   va_start( va, format);
   if( vfprintf( fdo, format, va) < 0)
      bailout( "cannot write %s, %s", outfilename, strerror( errno));

   va_end( va);
}

static void output_close( void)
{
   if( fclose( fdo))
      bailout( "error closing %s, %s", outfilename, strerror( errno));
   free( outfilename);
}

///////////////////////////////////////////////////////////////////////////////
//  Output Listing                                                           //
///////////////////////////////////////////////////////////////////////////////

//
//  Called when -o list is given.  Produce a data file listing all the
//  reactances and resistances.
//

static void output_listing( void)
{
   int i;
   char id1[50], id2[50];

   output_open( "list");

   //
   // Output capacitances.
   //

   struct C_STORE *cp = Cstore;

   for( i = 0; i < Cstore_cnt; i++, cp++)
   {
      if( cp->e1->coil)
         sprintf( id1, "%s_%d", cp->e1->coil->name, cp->e1->unit);
      else
         strcpy( id1, cp->e1->name);

      if( cp->e2->coil)
         sprintf( id2, "%s_%d", cp->e2->coil->name, cp->e2->unit);
      else
         strcpy( id2, cp->e2->name);

      if( cp->e1 == cp->e2)
         outputs( "C %s inf %.5e\n", id1, cp->C);
      else
         outputs( "C %s %s %.5e\n", id1, id2, cp->C);
   }

   //
   //  Output inductances.
   //

   // Inductance, self and mutual, of each coil section
   struct L_STORE *lp;

   for( i = 0, lp = Lstore; i < Lstore_cnt; i++, lp++)
   {
      sprintf( id1, "%s_%d", lp->c1->name, lp->s1);

      if( lp->c1 == lp->c2 && lp->s1 == lp->s2) strcpy( id2, "self");
      else
         sprintf( id2, "%s_%d", lp->c2->name, lp->s2);

      outputs( "L %s %s %.5e\n", id1, id2, lp->L);
   }

   // Self inductance of each coil
   struct COIL *c;
   for( c = WORLD.coils; c; c = c->next)
   {
      double L = 0;

      for( i = 0, lp = Lstore; i < Lstore_cnt; i++, lp++)
      {
         if( lp->c1 != c || lp->c2 != c) continue;

         if( lp->s1 == lp->s2) L += lp->L;
         else L += 2 * lp->L;
      }
      outputs( "L %s self %.5e\n", c->name, L);
   }
   
   // Mutual inductance of each pair of coils
   struct COIL *c1, *c2;
   for( c1 = WORLD.coils; c1; c1 = c1->next)
      for( c2 = c1->next; c2; c2 = c2->next)
      {
         double M = 0;

         for( i = 0, lp = Lstore; i < Lstore_cnt; i++, lp++)
         {
            if( lp->c1 == c1 && lp->c2 == c2 ||
                lp->c2 == c1 && lp->c1 == c2) M += lp->L;
         }

         outputs( "L %s %s %.5e\n", c1->name, c2->name, M);
      }

   //
   //  Output DC resistances.
   //

   struct R_STORE *rp = Rstore;

   for( i = 0; i < Rstore_cnt; i++, rp++)
      outputs( "R %s_%d %.5e\n", rp->c->name, rp->s+1, rp->R);

   //
   //  Output induced charge.
   //

   if( WORLD.Efield)
   {
      struct Q_STORE *qp = Qstore;

      for( i = 0; i < Qstore_cnt; i++, qp++)
      {
         if( qp->e->coil)
            sprintf( id1, "%s_%d", qp->e->coil->name, qp->e->unit);
         else
            strcpy( id1, qp->e->name);
   
         outputs( "Q %s %.5e\n", id1, qp->Q);
      }
   }

   //
   //  Output linked flux.
   //

   if( WORLD.Bfield)
   {
      struct F_STORE *fp = Fstore;

      for( i = 0; i < Fstore_cnt; i++, fp++)
         outputs( "F %s_%d %.5e\n", fp->c->name, fp->s, fp->F);
   }

   output_close();
}

///////////////////////////////////////////////////////////////////////////////
//  Output Spice Sub-Circuit                                                 //
///////////////////////////////////////////////////////////////////////////////

//  Skin depth s = sqrt(1.75e-8/(pi * hertz * mu))

//  effective_area ~= s * pi * 2 * wirad
//  total_area = pi * wirad * wirad

//  ratio = total_area/effective_area
//        = pi * wirad * wirad / (s * 2 * pi * wirad)
//        = wirad / (s * 2)
//        = wirad/2 * sqrt(pi * hertz * mu/1.75e-8)
//        = wirad/2 * sqrt(pi * mu/1.75e-8) * sqrt(hertz)

//  If ratio <= 1 then use DC resistance:
//  ratio = 1 = wirad/2 * sqrt(pi * mu/1.75e-8) * sqrt(hertz_min)
//  ratio^2 = 1 = wirad^2/4 * pi * mu/1.75e-8 * hertz_min
//  hertz_min = 4/(wirad^2)/pi/mu * 1.75e-8
//  ac_res = Rdc * ratio
//         = Rdc * wirad/2 * sqrt(pi * mu/1.75e-8) * sqrt(hertz) 
//  R = hertz > hertz_min ? ac_res : dc_res

static char *ac_res_expr( struct COIL *c, int section)
{
   static char expr[100];

   double Rdc = find_R( c, section);

   double hertz_min = 4.0/(c->wirad * c->wirad)/M_PI/MU * 1.75e-8;
   logmsg( 2, "hertz min %.4e", hertz_min);

   if( FLAG_RDC)
      sprintf( expr, "%.4e", Rdc);
   else
      sprintf( expr, "R = '(HERTZ > %.4e) ? (%.4e * sqrt(HERTZ)) : %.4e'",
                   hertz_min, Rdc * c->wirad/2 * sqrt(M_PI * MU/1.75e-8), Rdc);

   return expr;
}

//
//  Called when -o spice is given.
//

static void output_spice( void)
{
   // Counters to generate unique component IDs within the sub-circuit
   static int Ccount = 0, Kcount = 0, Rcount = 0,
              Fcount = 0, Ecount = 0;

   struct ELECTRODE *e, *et;
   struct COIL *c;
   int i, j;

   output_open( "spice");

   //
   //  Generate header with pins to each electrode and coil.  If this is
   //  altered, then alter output_geda() to match the pin sequence.
   //

   outputs( ".SUBCKT %s INF", WORLD.sysname);

   for( e = WORLD.electrodes; e; e = e->next)
      if( !e->coil) outputs( " E%d", e->ident);

   for( c = WORLD.coils; c; c = c->next)
   {
      for( i = 0; i<c->ntaps; i++)
         outputs( " L%d_%d", c->ident, c->tsect[i]);
   }

   if( WORLD.Efield) outputs( " EF");
   if( WORLD.Bfield) outputs( " BF");

   outputs( "\n*\n");

   int pin = 0;

   outputs( "* pin %d: zero of potential at infinity\n", ++pin);
   for( e = WORLD.electrodes; e; e = e->next)
      if( !e->coil) outputs( "* pin %d: %s\n", ++pin, e->name);

   for( c = WORLD.coils; c; c = c->next)
   {
      outputs( "* pin %d: %s end1, 0.00 turns\n", ++pin, c->name);
      for( i = 1; i<c->ntaps - 1; i++)
         outputs( "* pin %d: %s tap, %.2f turns\n", ++pin, c->name, c->taps[i]);
      outputs( "* pin %d: %s end2, %.2f turns\n", ++pin, c->name, c->turns);
   }

   if( WORLD.Efield) outputs( "* pin %d: E-field, 1 volt = 1 volt/metre\n", ++pin);
   if( WORLD.Bfield) outputs( "* pin %d: B-field, 1 amp = 1 Tesla\n", ++pin);

   outputs( "*\n* Generated by lcng version %s\n*\n\n", VERSION);

   if( WORLD.Efield)
   {
      outputs( "* E-field generator\n");
      outputs( "CE EF EF1 1.0\n");
      outputs( "VE EF1 0 AC 0 DC 0\n\n");
   }

   if( WORLD.Bfield)
   {
      outputs( "* B-field generator\n");
      outputs( "LB BF 0 1.0\n\n");
   }

   //
   //  Output electrode capacitances.  If an E-field is present, include a
   //  current source to inject charge from infinity into each electrode.
   //
   //  Zero or negative capacitances (which can arise from limited convergence
   //  of the solution) are discarded.
   //

   for( e = WORLD.electrodes; e; e = e->next)
   {
      if( e->coil) continue;   // Do electrodes which are coil sections later

      // Self capacitances

      double C = find_C( e, e);
      if( C > 0)
      {
         outputs( "* Self capacitance of %s\n", e->name);
         outputs( "C%d E%d INF %.5e\n", ++Ccount, e->ident, C);

         if( WORLD.Efield)
         {
            outputs( "* E-field coupling to %s\n", e->name);
            outputs( "F%d E%d INF VE %.5e\n", ++Fcount, e->ident, find_Q( e));
         }

         outputs( "\n");
      }

      // Mutual capacitances

      for( et = e->next; et; et = et->next)
      {
         if( et->coil) continue;

         double C = find_C( e, et);
         if( C > 0)
         {
            outputs( "* Capacitance between %s and %s\n",
                           e->name, et->name);
            outputs( "C%d E%d E%d %.5e\n",
                          ++Ccount, e->ident, et->ident, C);
         }
      }
    
      outputs( "\n");
   }
   
   //
   //  Output coils.  If a B-field is present, include a voltage source in
   //  series with each coil section to represent the induced EMF.
   //

   for( c = WORLD.coils; c; c = c->next)
   {
      outputs( "* Coil %s in %d sections\n", c->name, c->Ns);

      for( i = 0; i < c->Ns; i++)
      {
         double L = find_L( c, i, c, i);
         outputs( "L%d_%d L%d_%d L%d_%dR %.5e\n",
                        c->ident, i, c->ident, i, c->ident, i, L);

         if( !WORLD.Bfield)
            outputs( "R%d L%d_%dR L%d_%d %s\n",
                     ++Rcount, c->ident, i, c->ident, i+1, ac_res_expr( c, i));
         else
         {
            outputs( "R%d L%d_%dR L%d_%dB %s\n",
                       ++Rcount, c->ident, i, c->ident, i, ac_res_expr( c, i));
            outputs( "E%d L%d_%d L%d_%dB BF 0 %.5e\n",
                          ++Ecount, c->ident, i+1, c->ident, i, find_F( c, i));
         }
      }
    
      outputs( "\n");

      if( c->Ns > 1)
      {
         outputs( "* Coil %s, internal L\n", c->name);
         for( i = 0; i < c->Ns; i++)
         {
            double L1 = find_L( c, i, c, i);
   
            for( j = i+1; j < c->Ns; j++)
            {
               double L2 = find_L( c, j, c, j);
               double M = find_L( c, i, c, j);

               outputs( "K%d L%d_%d L%d_%d %.5e\n",
                           ++Kcount, c->ident, i, c->ident, j, M/sqrt(L1 * L2));
            }
         }
         outputs( "\n");
      }

      outputs( "* Coil %s, external C\n", c->name);

      struct C_STORE *cp;
      for( cp = Cstore, i = 0; i < Cstore_cnt; i++, cp++)
      {
         if( cp->C <= 0) continue;

         if( cp->e1->coil == c && cp->e2 == cp->e1)
         {
            if( !WORLD.Efield)
               outputs( "C%d L%d_%d INF %.5e\n",
                              ++Ccount, c->ident, cp->e1->unit+1, cp->C); 
            else
            {
               outputs( "C%d L%d_%d LI%d_%d %.5e\n", ++Ccount,
                         c->ident, cp->e1->unit+1,
                         c->ident, cp->e1->unit+1, cp->C);
               double Q = find_Q( cp->e1);
               outputs( "F%d LI%d_%d INF VE %.5e\n", ++Fcount,
                         c->ident, cp->e1->unit+1, Q/sqrt(2));

               // Extra R so that spice can find DC point
               outputs( "R%d LI%d_%d INF 100e9\n", ++Rcount, 
                        c->ident, cp->e1->unit+1);
            }
         }

         if( cp->e1->coil == c && cp->e2->coil == NULL)
            outputs( "C%d L%d_%d E%d %.5e\n",
                           ++Ccount, c->ident, cp->e1->unit+1, 
                           cp->e2->ident, cp->C); 

         if( cp->e2->coil == c && cp->e1->coil == NULL)
            outputs( "C%d L%d_%d E%d %.5e\n",
                           ++Ccount, c->ident, cp->e2->unit+1,
                           cp->e1->ident, cp->C); 
      }

      outputs( "\n");

      if( c->Ns > 1)
      {
         outputs( "* Coil %s, internal C\n", c->name);
         for( cp = Cstore, i = 0; i < Cstore_cnt; i++, cp++)
            if( cp->e1->coil == c && cp->e2->coil == c &&
                cp->e1 != cp->e2 && cp->C > 0)
               outputs( "C%d L%d_%d L%d_%d %.5e\n",
                              ++Ccount,
                              c->ident, cp->e1->unit+1,
                              c->ident, cp->e2->unit+1,
                              cp->C); 

         outputs( "\n");
      }
   }

   //
   //  Inductive coupling between coils.
   //

   for( c = WORLD.coils; c; c = c->next)
      for( struct COIL *ct = c->next; ct; ct = ct->next)
      {
         outputs( "* Mutual inductance, coil %s to %s\n", c->name, ct->name);

         for( i = 0; i < c->Ns; i++)
         {
            double L1 = find_L( c, i, c, i);

            for( j = 0; j < ct->Ns; j++)
            {
               double L2 = find_L( ct, j, ct, j);
               double M = find_L( c, i, ct, j);

               outputs( "K%d L%d_%d L%d_%d %.5e\n",
                         ++Kcount, c->ident, i, ct->ident, j, M/sqrt(L1 * L2));
            }
         }
         outputs( "\n");
      }

   //
   //  Capacitive coupling between coils.
   //

   for( c = WORLD.coils; c; c = c->next)
      for( struct COIL *ct = c->next; ct; ct = ct->next)
      {
         outputs( "* Mutual capacitance, coil %s to %s\n", c->name, ct->name);

         for( int i1 = 0; i1 < c->Ns; i1++)
            for( int i2 = 0; i2 < ct->Ns; i2++)
            {
               struct CSECT *s1 = c->sections + i1;
               struct CSECT *s2 = ct->sections + i2;
               outputs( "C%d L%d_%d L%d_%d %.5e\n",
                             ++Ccount,
                             c->ident, s1->unit+1,
                             ct->ident, s2->unit+1, find_C( s1->e, s2->e));
            }
         outputs( "\n");
      }

   outputs( ".ENDS %s\n", WORLD.sysname);
   output_close();
}

///////////////////////////////////////////////////////////////////////////////
//  Output Tile Data                                                         //
///////////////////////////////////////////////////////////////////////////////

//
//  Called when -o tiles is given.  Produce a data file listing the x,y,z
//  coordinates of every tile.
//

static void output_tiles( void)
{
   output_open( "tiles");

   for( int i = 0; i < ntiles; i++)
   {
      struct TILE *t = tile_list[i];
      outputs( "%.6e %.6e %.6e\n", t->c.x, t->c.y, t->c.z);
#if 0
      outputs( "%.6e %.6e %.6e   %.6e %.6e %.6e   %.6e %.6e %.6e\n",
        t->c.x, t->c.y, t->c.z,
        t->ux.x*t->sx, t->ux.y*t->sx, t->ux.z*t->sx,
        t->uy.x*t->sy, t->uy.y*t->sy, t->uy.z*t->sy
        );
#endif
   }

   output_close();
}

///////////////////////////////////////////////////////////////////////////////
//  Output to gnuplot                                                        //
///////////////////////////////////////////////////////////////////////////////

//
//  Called when -o gnuplot is given.  Launches gnuplot and sends it the tile
//  points to plot in 3D mode.
//
//  gnuplot stays interactive 
static void output_gnuplot( void)
{
   FILE *p = popen( "gnuplot", "w");
   if( !p) bailout( "unable to run gnuplot: %s", strerror( errno));

   fprintf( p, "set terminal x11 size 640,640\n"
               "set style data points\n"
               "set view equal xyz\n"
               "set xyplane 0\n"
               "set xlabel 'X'\n"
               "set ylabel 'Y'\n"
               "set zlabel 'Z'\n"
               "set grid\n"
               "unset key\n");

   fprintf( p, "set title '%s'\n", WORLD.sysname);

   double range = WORLD.msize * 1.05;
   fprintf( p, "set xrange [%.6e:%.6e]\n", WORLD.rc.x - range,
                                           WORLD.rc.x + range);
   fprintf( p, "set yrange [%.6e:%.6e]\n", WORLD.rc.y - range,
                                           WORLD.rc.y + range);
   fprintf( p, "set zrange [%.6e:%.6e]\n", WORLD.rc.z - range,
                                           WORLD.rc.z + range);

   fprintf( p, "splot '-' using 1:2:3\n");
 
   for( int i = 0; i < ntiles; i++)
   {
      struct TILE *t = tile_list[i];
      fprintf( p, "%.6e %.6e %.6e\n", t->c.x, t->c.y, t->c.z);
   }

   fprintf( p, "e\n");
   fflush( p);

   fgetc( stdin); // Wait here for a key press

   pclose( p);
}


///////////////////////////////////////////////////////////////////////////////
//  Output Povray File                                                       //
///////////////////////////////////////////////////////////////////////////////

//
//  Product an input file for povray which describes the system.
//
//  Takes a guess at a suitable camera position and 'look at' point.
//

//
//  Convert from our right handed coordinate system to the left handed system
//  used by povray.
//

static char *pov_point( struct POINT *p)
{
   static char temp[50];

   sprintf( temp, "<%.6e, %.6e, %.6e>", p->x, p->z, p->y);
   return temp;
}

static void output_povray( void)
{
   output_open( "pov");

   outputs( "global_settings { ambient_light color rgb <1,1,1>\n"
            "}\n\n");

   double fov = 30;   // Field width, degrees

   double r = 4 * WORLD.msize/tan(fov * M_PI/180);
   struct POINT cam = WORLD.rc;   cam.z += r/3;   cam.y -= r;

   outputs( "background { rgb <0.3,0.3,0.3> }\n"
            "\n"
            "camera {\n"
            "   location %s\n", 
            pov_point( &cam));

   // Aim at the centroid
   outputs( "   up <0,1,0>\n   right <1.78,0,0>\n"
            "   angle %.1f\n"
            "   look_at %s\n"
            "}\n",
            fov, pov_point( &WORLD.rc));

   outputs( "\n");

   outputs( "#declare f_metal = finish {\n"
            "   ambient 0.2 diffuse 0.2 specular 1.5\n"
            "   roughness 0.02 brilliance 2.0\n"
            "   reflection { 0.6 metallic 1 } metallic\n"
            "}\n\n");

   outputs( "#declare t_wire = texture {\n"
            "  pigment { rgb <0.82, 0.25, 0.10> }\n"
            "   normal { waves 0.10 frequency 3 scale 0.005 }\n"
            "   finish { f_metal }\n"
            "}\n\n");

   outputs( "#declare t_metal = texture {\n"
            "   pigment { rgb <0.8,0.8,0.85> }\n"
            "   finish { f_metal }\n"
            "}\n\n");

   outputs( "light_source { <3, 4, -3> color rgb <1,1,1>}\n"
            "light_source { <-4.5, 2, -1.5> color rgb <1,1,1>}\n"
            "light_source { <0, 8, 0> color rgb <1,1,1>}\n"
            "\n");

   if( WORLD.g)
   {
      outputs( "// Ground plane\n");
      outputs( "plane { %s, %.6e\n", pov_point( &WORLD.g->norm), -WORLD.g->offset);
      outputs( "   pigment { checker color rgb <0.55,0.55,0.55>,\n");
      outputs( "                     color rgb <0.50,0.50,0.50>\n");
      outputs( "   }\n"
               "   finish { ambient 0.3 specular 0.3 roughness 0.01 }\n"
               "}\n");
   }
#if 0
   else
      outputs( "// Transparent grid in xy plane\n"
               "plane { <0,1,0>,0\n"
               "   pigment {\n"
               "      boxed scale 0.5 translate <0.5,0.5,0>\n"
               "      warp {repeat x} warp {repeat y} warp {planar z, 0}\n"
               "      color_map{\n"
               "         [0.01 color rgb <1,1,1>]    // Grid color \n"
               "         [0.06 color rgbt <0,0,0,1>] // Transparent\n" 
               "      }\n"
               "      rotate x*90\n"
               "   }\n"
               "}\n");
#endif

   for( struct ELECTRODE *e = WORLD.electrodes; e; e = e->next)
   {
      outputs( "// Electrode %s\n", e->name);

      char *texture = e->coil ? "t_wire" : "t_metal";

      for( struct COMPONENT *comp = e->components; comp; comp = comp->next)
      {
         switch( comp->type)
         {
            case CT_SPHERE:
               outputs( "sphere {\n");
               outputs( "   %s %.6e\n", pov_point( comp->u.sphere.center),
                                        comp->u.sphere.radius);
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");
               break;

            case CT_CYL:
               outputs( "cone {\n");
               outputs( "   %s,\n", pov_point( comp->u.cyl.end1));
               outputs( "   %.6e,\n", comp->u.cyl.radius1);
               outputs( "   %s,\n", pov_point( comp->u.cyl.end2));
               outputs( "   %.6e\n", comp->u.cyl.radius2);
               outputs( "   open\n");
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");
               break;

            case CT_TOROID:
               outputs( "torus {\n");
               double tube_diam = comp->u.toroid.outer_radius - 
                                  comp->u.toroid.inner_radius;
               outputs( "   %.6e, %.6e\n",
                        comp->u.toroid.inner_radius + tube_diam/2,
                        tube_diam/2);
               outputs( "   rotate <0,0,%.6e>\n",
                        comp->u.toroid.axis->pa * 180/M_PI);
               outputs( "   rotate <0,%.6e,0>\n",
                        comp->u.toroid.axis->azim * 180/M_PI);
               outputs( "   translate %s\n", pov_point( comp->u.toroid.center));
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");
               break;

            case CT_DISC:
               outputs( "disc {\n");
               struct POINT norm;
               forepoint( comp->u.disc.axis, 1.0, &norm, NULL);
               outputs( "   %s,\n", pov_point( comp->u.disc.center));
               outputs( "   %s,\n", pov_point( &norm));
               outputs( "   %.6e, %.6e\n", comp->u.disc.radius2,
                                           comp->u.disc.radius1);
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");
               break;

            case CT_RECT:
               outputs( "polygon {\n");
               struct POINT d1 = sub_3D( *comp->u.rect.edge1,
                                         *comp->u.rect.center);
               struct POINT d2 = sub_3D( *comp->u.rect.edge2,
                                         *comp->u.rect.center);
               struct POINT p1 = sub_3D( *comp->u.rect.center, d1);
                            p1 = sub_3D( p1, d2);
               
               struct POINT p2 = add_3D( *comp->u.rect.center, d1);
                            p2 = sub_3D( p2, d2);
                            
               struct POINT p3 = add_3D( *comp->u.rect.center, d1);
                            p3 = add_3D( p3, d2);

               struct POINT p4 = sub_3D( *comp->u.rect.center, d1);
                            p4 = add_3D( p4, d2);

               outputs( "   4,\n");
               outputs( "   %s\n", pov_point( &p1));
               outputs( "   %s\n", pov_point( &p2));
               outputs( "   %s\n", pov_point( &p3));
               outputs( "   %s\n", pov_point( &p4));
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");
               break;

            case CT_PLATE:
               outputs( "polygon {\n");

               // Povray insists points are *exactly* coplanar, so have to
               // enforce this
               int N = comp->u.plate.npoints;
               struct POINT *list = array( N, struct POINT);
               for( int i = 0; i < N; i++)
                  list[i] = *comp->u.plate.points[i];

               struct POINT center = origin;
               for( int i = 0; i < N; i++)
                  sum_3D( &center, list[i]);
               scale_3D( &center, 1.0/N);
               for( int i = 0; i < N; i++) list[i] = sub_3D( list[i], center);

               struct AXIS axis = polygon_axis( list, comp->u.plate.npoints);
               for( int i = 0; i < N; i++)
               {
                  list[i] = unrotate( list[i], axis.pa, axis.azim);
                  list[i].z = 0;   // Make exactly in xy plane
               }

               outputs( "   %d,\n", N+1);
               for( int i = 0; i < N; i++)
                  outputs( "   %s\n", pov_point( &list[i]));
               outputs( "   %s\n", pov_point( &list[0]));
               outputs( "   texture { %s }\n", texture);

               // Rotate and translate to its proper position
               outputs( "   rotate <0,0,%.6e>\n", -axis.pa * 180/M_PI);
               outputs( "   rotate <0,%.6e,0>\n", -axis.azim * 180/M_PI);
               outputs( "   translate %s\n", pov_point( &center));
               outputs( "}\n");
               break;

            case CT_WIRE:
               outputs( "cylinder {\n");
               outputs( "%s,\n", pov_point( comp->u.wire.end1));
               outputs( "%s,\n", pov_point( comp->u.wire.end2));
               outputs( "%.6e\n", comp->u.wire.wirad);
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");

               if( comp->u.wire.next)  // Part of a polygonal loop coil?
               {
                  outputs( "sphere {\n");
                  outputs( "   %s %.6e\n", pov_point( comp->u.wire.end2),
                                           comp->u.wire.wirad);
                  outputs( "   texture { %s }\n", texture);
                  outputs( "}\n");
               }
               break;

            case CT_RING:
               outputs( "torus {\n");
               outputs( "   %.6e, %.6e\n", comp->u.ring.radius,
                                           comp->u.ring.wirad);
               outputs( "   rotate <0,0,%.6e>\n",
                        comp->u.ring.axis->pa * 180/M_PI);
               outputs( "   rotate <0,%.6e,0>\n",
                        comp->u.ring.axis->azim * 180/M_PI);
               outputs( "   translate %s\n", pov_point( comp->u.ring.center));
               outputs( "   texture { %s }\n", texture);
               outputs( "}\n");
               break;
         }
      }

      outputs( "\n");     
   }

   output_close();
}

///////////////////////////////////////////////////////////////////////////////
//  Output gEDA Symbol                                                       //
///////////////////////////////////////////////////////////////////////////////

#define SIDE_TOP    1
#define SIDE_LEFT   2
#define SIDE_RIGHT  3
#define SIDE_BOTTOM 4

#define PIN_GAP 800
#define PPC  130    // Points per character.  Font is variable width so this is
                    // a guess at the max.

static struct PIN {
   int pinseq;
   int side;
   char *label;
} *pins = NULL;

static int npins = 0;

static void __attribute__ ((format (printf, 2, 3)))
    add_pin( int side, char *format, ...)
{
   pins = realloc( pins, (npins + 1) * sizeof( struct PIN));
   if( !pins) bailout( "out of memory");
   memset( pins + npins, 0, sizeof( struct PIN));

   char temp[100];
   va_list va; va_start( va, format); vsprintf( temp, format, va); va_end( va);


   pins[npins].label = strdup( temp);
   pins[npins].side = side;
   pins[npins].pinseq = npins + 1;
   npins++;
}

static void output_geda( void)
{
   output_open( "sym");

   // Assign pins.  Must be in the same order as in output_spice()

   int side = SIDE_LEFT;
   void toggle_side( void)
       { side = side == SIDE_LEFT ? SIDE_RIGHT : SIDE_LEFT; }

   add_pin( SIDE_BOTTOM, "INF");

   for( struct ELECTRODE *e = WORLD.electrodes; e; e = e->next)
      if( !e->coil)
      {
         add_pin( side, "%s", e->name);
         toggle_side();
      }

   for( struct COIL *c = WORLD.coils; c; c = c->next)
   {
      for( int i = 0; i<c->ntaps; i++)
         add_pin( side, "%s_%d", c->name, c->tsect[i]);
      toggle_side();   // All pins for a coil together on same side
   }

   if( WORLD.Efield) add_pin( side, "EF");
   if( WORLD.Bfield) add_pin( side, "BF");

   int size_left = PIN_GAP, size_right = PIN_GAP,
       size_top = PIN_GAP, size_bottom = PIN_GAP;

   int wl_left = 0, wl_right = 0, wl_bottom = 0;
   for( int i = 0; i < npins; i++)
   {
      int w = strlen( pins[i].label);

      switch ( pins[i].side)
      {
         case SIDE_LEFT:   size_left += PIN_GAP;
                           if( w > wl_left) wl_left = w;
                           break;
         case SIDE_RIGHT:  size_right += PIN_GAP;
                           if( w > wl_right) wl_right = w;
                           break;
         case SIDE_TOP:    size_top += PIN_GAP;
                           break;
         case SIDE_BOTTOM: size_bottom += PIN_GAP;
                           if( w > wl_bottom) wl_bottom = w;
                           break;
      }
   }

   int box_width = imax( size_top, size_bottom);
   box_width = imax( box_width, (wl_left + wl_right + 3) * PPC);
   box_width = imax( box_width,
                     2 * PIN_GAP + PPC*strlen( WORLD.sysname));
   
   int box_height = imax( size_left, size_right);

   // gschem likes to have pins, etc at multiples of 100
   box_width = box_width / 100 * 100;
   box_height = box_height / 100 * 100;

   int xmargin = size_left > PIN_GAP ? 300 : 0;
   int ymargin = size_bottom > PIN_GAP ? 300 : 0;

   outputs( "v 20110115 2\n");
   outputs( "B %d %d %d %d 3 0 0 0 -1 -1 0 -1 -1 -1 -1 -1\n",
            xmargin, ymargin, box_width, box_height);

   int xp = PIN_GAP, yp_left = PIN_GAP, yp_right = PIN_GAP;

   for( int i = 0; i < npins; i++)
   {
      if( pins[i].side == SIDE_LEFT)
      {
         int x = 0, y = yp_left + ymargin;
         outputs( "P %d %d %d %d 1 0 0\n", x, y, x + 300, y);
         outputs( "{\n");
         outputs( "T %d %d 5 10 0 0 0 0 1\n", x, y);
         outputs( "pintype=passive\n");
         outputs( "T %d %d 5 10 1 1 0 0 1\n", x + 355, y - 5);
         outputs( "pinlabel=%s\n", pins[i].label);
         outputs( "T %d %d 5 10 0 1 0 6 1\n", x + 205, y + 45);
         outputs( "pinnumber=%d\n", pins[i].pinseq);
         outputs( "T %d %d 5 10 0 0 0 0 1\n", x, y);
         outputs( "pinseq=%d\n", pins[i].pinseq);
         outputs( "}\n");
         yp_left += PIN_GAP; 
      }
      else
      if( pins[i].side == SIDE_RIGHT)
      {
         int x = xmargin + box_width + 300, y = yp_right + ymargin;
         outputs( "P %d %d %d %d 1 0 0\n", x, y, x - 300, y);
         outputs( "{\n");
         outputs( "T %d %d 5 10 0 0 0 0 1\n", x, y);
         outputs( "pintype=passive\n");
         outputs( "T %d %d 5 10 1 1 0 0 1\n",
                   x - 355 - PPC * (int)strlen( pins[i].label) , y - 5);
         outputs( "pinlabel=%s\n", pins[i].label);
         outputs( "T %d %d 5 10 0 1 0 6 1\n", x - 205, y + 45);
         outputs( "pinnumber=%d\n", pins[i].pinseq);
         outputs( "T %d %d 5 10 0 0 0 0 1\n", x, y);
         outputs( "pinseq=%d\n", pins[i].pinseq);
         outputs( "}\n");
         yp_right += PIN_GAP; 
      }
      else
      if( pins[i].side == SIDE_BOTTOM)
      {
         int x = xmargin + xp, y = 0;
         outputs( "P %d %d %d %d 1 0 0\n", x, y, x, y + 300);
         outputs( "{\n");
         outputs( "T %d %d 5 10 0 0 0 0 1\n", x, y);
         outputs( "pintype=passive\n");
         outputs( "T %d %d 5 10 1 1 0 0 1\n",
                   x - PPC * (int)strlen( pins[i].label)/2 , y + 345 );
         outputs( "pinlabel=%s\n", pins[i].label);
         outputs( "T %d %d 5 10 0 1 0 6 1\n", x, y);
         outputs( "pinnumber=%d\n", pins[i].pinseq);
         outputs( "T %d %d 5 10 0 0 0 0 1\n", x, y + 45);
         outputs( "pinseq=%d\n", pins[i].pinseq);
         outputs( "}\n");
         xp += PIN_GAP; 
      }

   }

   outputs( "T %d %d 8 10 1 1 0 0 1\n",
             xmargin + box_width/2 - 100 * (int)strlen( WORLD.sysname)/2,
             ymargin + box_height - 200);
   outputs( "device=%s\n", WORLD.sysname);

   outputs( "T %d %d 15 10 1 1 0 0 1\n",
             xmargin,
             ymargin + box_height + 300);
   outputs( "refdes=X?\n");

   output_close();
}

///////////////////////////////////////////////////////////////////////////////
//  Matrix and Vector Functions                                              //
///////////////////////////////////////////////////////////////////////////////

//
//  Duplicate a matrix.
//

static PMTYPE **dup_2D( int N, PMTYPE **A)
{
   int i, j;
   PMTYPE **p = array2D( N, PMTYPE);

   for( i = 0; i < N; i++) for( j = 0; j < N; j++) p[i][j] = A[i][j];
   return p;
}

//
//  Inner product between two arrays.
//

static double inner_prod( int N, QVTYPE *v1, QVTYPE *v2)
{
   int i;
   double t = 0;

   for( i = 0; i < N; i++) t += v1[i] * v2[i];
   return t;
}

//
//  Solve [a][f] + [c] = [0] by Gaussian elimination to obtain [f].
//

static void gaussian_elim( int Neq, PMTYPE **A, QVTYPE *c, QVTYPE *f)
{
   PMTYPE **AX = dup_2D( Neq, A);    // Working copy because it gets destroyed

   int *col_used;               // Flags to remember which rows have been used
   int *row_use;                               // The row used for each column

   col_used = array( Neq, int);
   row_use = array( Neq, int);

   int row, col;
   for( row = 0; row < Neq; row++) col_used[ row] = -1;
 
   for( col = 0; col < Neq-1; col++)      // For each column except the last...
   {
      double t, g;   int pivot = 0;

      //
      //  Select a pivot row. Look for the unused row with the highest
      //  magnitude in this column.
      //
 
      for( t = 0.0, row = 0; row < Neq; row++) 
         if( col_used[ row] < 0 && (g = fabs( AX[ row][col])) > t) 
         { 
            t = g;
            pivot = row;
         }
 
      if( t == 0.0) bailout( "error: empty column %d in elimination", col);
 
      //
      //  Mark the row used, so we dont use it again, and remember the
      //  column it was used to eliminate, for use when back substituting.
      //

      col_used[ pivot] = col; 
      row_use[ col] = pivot;

      //
      //  Subtract multiples of the pivot row from all the other unused rows.
      //
 
      for( row = 0; row < Neq; row++)
         if( col_used[ row] < 0)
         {
            int ct;
            double f = AX[ row][col]/AX[ pivot][col];

            for( ct = col; ct < Neq; ct++)
               AX[ row][ct] = AX[ row][ct] - f * AX[ pivot][ct];
 
            c[ row] = c[ row] - f * c[ pivot];
         }
   }
 
   //
   //  One un-used row remains, which is the starting point for back
   //  substitution. Find it and add to the used list.
   //
 
   for( row = 0; row < Neq && col_used[ row] >= 0; row++);

   col_used[ row] = Neq - 1; 
   row_use[ Neq-1] = row; 
  
   //
   //  Back-substitute to obtain the [f] values.
   //
 
   for( col = Neq-1; col >= 0; col--)
   {
      row = row_use[ col];
      col_used[ row] = -1;

      if( AX[ row][col] == 0.0)
         bailout( "empty coefficient, col=%d\n", col); 

      f[ col] = - c[row]/ AX[ row][col];

      //
      //  Substitute for f[col] in the remaining rows.
      //

      for( row = 0; row < Neq; row++)
         if( col_used[ row] >= 0)
            c[ row] = c[ row]+ AX[ row][col]* f[ col];
   }

   free2D( AX); free( col_used); free( row_use);
}

static void gmres( int size, void eval(QVTYPE *, QVTYPE *),
                   QVTYPE *r, QVTYPE *q)
{
   int maxiter = 150;
   int iter, i, j;
   double rnorm, norm;
   double hi, hip1, length;

   QVTYPE *ap = array( size, QVTYPE);
   QVTYPE *p = array( size, QVTYPE);

   static double *c = NULL, *s = NULL, *g = NULL, *y = NULL;
   static QVTYPE **bv = NULL;
   static double **bh = NULL;
  
   if( !c)  // Allocation first time through
   {
      c = array( size, double);
      s = array( size, double);
      g = array( size, double);
      y = array( size, double);
      bv = array( maxiter, QVTYPE *);
      bh = array( maxiter, double *);
   }
 
   rnorm = sqrt( inner_prod( size, r, r));

   for( i = 0; i < size; i++) 
   {
      p[i] = r[i] / rnorm;
      g[i] = 0.0;
   }
   g[0] = rnorm;

   for( iter = 0; iter < maxiter && rnorm > CCTOL; iter++)
   {
      // allocate the back vectors if they haven't been already
      if( !bv[iter]) 
      {
         bv[iter] = array( size, QVTYPE);
         bh[iter] = array( iter+2, double);
      }
    
      for( i = 0; i < size; i++) bv[iter][i] = p[i];
    
      // Compute potential ap from charge vector p
      eval( ap, p); 

      // Initialize v^{iter+1} to Av^{iter}
      for( i = 0; i < size; i++) p[i] = ap[i];
    
      // Make v^{iter+1} orthogonal to v^{i}, i <= iter
      for( j = 0; j <= iter; j++) 
      {
         hi = inner_prod( size, ap, bv[j]);
         for( i = 0; i < size; i++) p[i] -= hi * bv[j][i];
         bh[iter][j] = hi;
      }
    
      // Normalize v^{iter+1}
      norm = sqrt( inner_prod( size, p, p));

      for( i = 0; i < size; i++) p[i] /= norm;
      bh[iter][iter+1] = norm;
    
      // Apply rotations to new h column
      for( i = 0; i < iter; i++) 
      {
         hi = bh[iter][i];
         hip1 = bh[iter][i+1];
         bh[iter][i] = c[i] * hi - s[i] * hip1;
         bh[iter][i+1] = c[i] * hip1 + s[i] * hi;
      }
    
      // Compute new rotations
      hi = bh[iter][iter];
      hip1 = bh[iter][iter+1];
      length = sqrt(hi * hi + hip1 * hip1);
      c[iter] = hi/length;
      s[iter] = -hip1/length;
    
      // Apply new rotations
      bh[iter][iter] = c[iter] * hi - s[iter] * hip1;
      bh[iter][iter+1] = c[iter] * hip1 + s[iter] * hi;
      hi = g[iter];
      g[iter] = c[iter] * hi;
      g[iter+1] = s[iter] * hi;    
    
      rnorm = fabs(g[iter+1]);
      logmsg( 3, "gmres iter %d %.4f", iter, rnorm);
   }

   iter--;

   for( i = 0; i <= iter; i++) y[i] = g[i];
   for( i = iter; i >= 0; i--) 
   {
      y[i] /= bh[i][i];
      for( j = i-1; j >= 0; j--) y[j] -= bh[i][j]*y[i];
   }

   for( i = 0; i < size; i++) 
   {
      q[i] = 0.0;
      for( j = 0; j <= iter; j++) q[i] += y[j] * bv[j][i];
   }

   free( ap); free( p);

   logmsg( 2, "gmres used %d iterations", iter);
   if( rnorm > CCTOL)
      logmsg( 0, "gmres failed to converge: rnorm=%.3f", rnorm);
}

//
//  Conjugate gradient method.
//

static void cgm( int size, void eval(QVTYPE *, QVTYPE *), QVTYPE *b, QVTYPE *x)
{
   QVTYPE *r = array( size, QVTYPE);
   QVTYPE *rn = array( size, QVTYPE);
   double *xn = array( size, double);
   QVTYPE *p = array( size, QVTYPE);
   double *pn = array( size, double);
   QVTYPE *t = array( size, QVTYPE);

   int i, k = 0;
   int maxiter = 4000;

   // r0 = b - Ax
   eval( r, x);
   for( i = 0; i<size; i++) r[i]  = b[i] - r[i];

   // p0 = r0
   for( i = 0; i<size; i++) p[i] = r[i];

   k = 0;
   while( 1)
   {
      // alpha = rk.rk / (pk. Apk)
      eval( t, p);   // t = A*pk
      double alpha = inner_prod( size, r, r) / inner_prod( size, p, t);

      // x(k+1) = x(k) + alpha * pk
      for( i = 0; i<size; i++) xn[i] = x[i] + alpha * p[i];

      // r(k+1) = r(k) - alpha * Apk;
      for( i = 0; i<size; i++) rn[i] = r[i] - alpha * t[i];

      double rnorm = sqrt( inner_prod( size, rn, rn));
      if( isnan( rnorm)) bailout( "failed to solve in cgm");

      if( rnorm < CCTOL) break;
      if( k == maxiter)
      {
         logmsg( 0, "cgm failed to converge: rnorm=%.3f", rnorm);
         break;
      }

      // beta = r(k+1).r(k+1)/(rk.rk)
      double beta = inner_prod( size, rn,rn) / inner_prod( size, r, r);

      // p(k+1) = r(k+1) + beta * p(k)
      for( i = 0; i < size; i++) pn[i] = rn[i] + beta * p[i];
      
      k++;
      for( i = 0; i < size; i++)
      {
         x[i] = xn[i];
         r[i] = rn[i];
         p[i] = pn[i];
      }

      logmsg( 4, "cgm %d %.3f", k, rnorm);
   }

   // Result is x(k+1)
   for( i = 0; i < size; i++) x[i] = xn[i];

   logmsg( 2, "cgm used %d iterations", k);

   free( r); free( rn); free( p);
   free( pn); free( xn); free( t);
}

//
//  LU decomposition and matrix inverse. From Numerical Recipies.
//

#define   LUTINY 1.0e-20

static void lubksb(PMTYPE **a, int n, int *indx, PMTYPE *b)
{
   for( int i = 0, ii = -1; i < n; i++)
   {
      int ip = indx[i];
      double sum = b[ip];
      b[ip] = b[i];
      if( ii >= 0) for( int j = ii; j < i; j++) sum -= a[i][j] * b[j];
      else if( sum) ii = i;
      b[i] = sum;
   }

   for( int i = n-1; i >= 0; i--)
   {
      double sum = b[i];
      for( int j = i+1; j < n; j++) sum -= a[i][j] * b[j];
      b[i] = sum / a[i][i];
   }
}

static void ludcmp( PMTYPE **a, int n, int *indx, PMTYPE *d)
{
   double *vv = array( n, double);

   *d = 1.0;
   for( int i = 0; i < n; i++)
   {
      double big = 0.0, temp;
      for( int j = 0; j < n; j++) if((temp=fabs(a[i][j])) > big) big = temp;

      if( big == 0.0) bailout( "singular matrix in ludcmp");
      vv[i] = 1.0/big;
   }

   int imax = 0;
   for( int j = 0; j < n; j++)
   {
      for( int i = 0; i < j; i++)
      {
         double sum = a[i][j];
         for( int k = 0; k < i; k++) sum -= a[i][k] * a[k][j];
         a[i][j] = sum;
      }

      double big = 0.0;
      for( int i = j; i < n; i++)
      {
         double d, sum = a[i][j];
         for( int k = 0; k < j; k++) sum -= a[i][k] * a[k][j];
         a[i][j] = sum;
         if( (d = vv[i] * fabs(sum)) >= big)
         {
            big = d;
            imax = i;
         }
      }

      if (j != imax)
      {
         for( int k = 0; k < n; k++)
         {
            double d = a[imax][k];
            a[imax][k] = a[j][k];
            a[j][k] = d;
         }

         *d = -(*d);
         vv[imax] = vv[j];
      }

      indx[j] = imax;
      if ( a[j][j] == 0.0) a[j][j] = LUTINY;
      if ( j != n-1)
      {
        double d = 1.0 / a[j][j];
        for( int i = j+1; i < n; i++) a[i][j] *= d;
      }
   }

   free(vv);
}

static void invert_matrix( int N, PMTYPE **mat)
{
   PMTYPE **y = array2D( N, PMTYPE);
   PMTYPE *col = array( N, PMTYPE);
   int *indx = array( N, int);

   PMTYPE d;
   ludcmp( mat, N, indx, &d);

   for( int j = 0; j < N; j++)
   {
      for( int i = 0; i < N; i++) col[i] = 0.0;

      col[j] = 1.0;
      lubksb(mat,N,indx,col);
      for( int i = 0; i < N; i++) y[i][j] = col[i];
   }

   for( int i = 0; i < N; i++) for( int j = 0; j < N; j++) mat[i][j] = y[i][j];

   free2D( y); free( col); free( indx);
}

///////////////////////////////////////////////////////////////////////////////
//  Input File Parsing                                                       //
///////////////////////////////////////////////////////////////////////////////

//
//  Stuff for reading and parsing the input file.  This is a very simple top
//  down recursive descent parser, so it's easy to add new bits of syntax as
//  and when required.
//

static char pbuff[ 500], *pb;          // Parsing buffer, and a pointer into it
static int lino;                       // Current line number in the input file
static int eof_flag = FALSE;          // TRUE when end-of-file has been reached
static FILE *fdi;

//
//  Throw out a parser error and exit.  Keep things simple by giving up at the
//  first error.
//

static void __attribute__ ((format (printf, 1, 2)))
            __attribute__ ((noreturn)) parse_error( char *msg, ...)
{
   va_list va;
   char temp[100];

   va_start( va, msg); vsprintf( temp, msg, va); va_end( va);

   fprintf( stderr, "error in input file: line %d: %s\n", lino, temp);
   exit( 1);
}

static void __attribute__ ((format (printf, 1, 2)))
            __attribute__ ((noreturn)) parse_error_context( char *msg, ...)
{
   va_list va;
   char temp[100];

   va_start( va, msg); vsprintf( temp, msg, va); va_end( va);

   fprintf( stderr, "error in input file: line %d: %s", lino, temp);
   if( !eof_flag)
      fprintf( stderr, " [%.8s...]\n", pb);
   else
      fprintf( stderr, "\n");
   exit( 1);
}

//
//  Read the next line of the input file into the parsing buffer pbuff.
//

static int parse_reload( void)
{
   *(pb = pbuff) = 0;

   if( fgets( pbuff, sizeof(pbuff), fdi)) { lino++; return TRUE; }

   eof_flag = TRUE;
   return FALSE;
}

//
//  Skip over arbitrary white space and blank lines.  Comments are discarded
//  by truncating the parse buffer at the ';'.
//

static void skip_whitespace( void)
{
   while( 1)
      if( !*pb)                                             // Skip blank lines
      {
         if( !parse_reload()) return;
      }
      else
      if( isspace( *pb)) pb++;                               // Skip whitespace
      else
      if( *pb == ';') *pb = 0;                                 // Skip comments
      else
         return;
}

static int is_separator( char c)
{
   return !c || isspace(c) ||
           c == '{' || c == '}' || c == '[' || c == ']' ||
           c == '(' || c == ')' ||
           c == ',';
}

//
//  Parse a variable name: A letter followed by a string of letters, digits,
//  and underscores.  If no_reserved is TRUE then don't accept any reserved
//  words.
//

char *reserved_words[] = { "FIELD", "COIL", "ELECTRODE", "GROUND",
                           "SHIFT", "ROTATE", "OFFSET", "AXIS",
                           "TRANSFORM", NULL};
                           
static int parse_variable( char *s, int no_reserved)
{
   char *t = s, *spb = pb;
   skip_whitespace();

   if( isalpha(*pb))
      while( *pb &&
             (isalnum( *pb) || *pb == '_')) *t++ = toupper(*pb++);

   *t = 0;

   if( no_reserved)
      for( char **r = reserved_words; *r; r++)
         if( !strcmp( s, *r)) { *s = 0; pb = spb; return 0; }
   return t - s;
}

static int parse_name( char **p)
{
   char name[100];

   *p = NULL;
   if( parse_variable( name, FALSE))
   {
      *p = strdup( name);
      return TRUE;
   }

   return FALSE;
}

//
//  Test for a character symbol.
//

static int parse_char( char c)
{
   skip_whitespace();
   if( *pb == c) { pb++; return c; }
   return 0;
}

//
//  Return TRUE if the given target string matches (case ignored) the next
//  token in the parse buffer.
//

static int parse_word( char *target)
{
   int n = strlen( target);

   skip_whitespace();
   if( strncasecmp( pb, target, n) || !is_separator( *(pb+n))) return FALSE;

   pb += n;
   return TRUE;
}

//
//  Read a float and integer number from the parse buffer.  We skip our own
//  whitespace, rather than leaving it to strtod, so we can jump blank lines
//  too.
//

static double parse_float( void)
{
   skip_whitespace();
   return strtod( pb, &pb);
}

//
//  Parse arithmetic functions.
//

static struct NODE * parse_tertiary_expression( void);

static struct NODE * parse_function( char *name)
{
   struct NODE *n = 0;

   if( !strcmp( name, "SQRT") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = sqrt;
   else
   if( !strcmp( name, "SIN") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = sin;
   else
   if( !strcmp( name, "ASIN") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = asin;
   else
   if( !strcmp( name, "COS") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = cos;
   else
   if( !strcmp( name, "ACOS") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = acos;
   else
   if( !strcmp( name, "EXP") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = exp;
   else
   if( !strcmp( name, "ABS") && parse_char( '('))
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = fabs;
   else
      return 0;

   if( !parse_char( ')')) parse_error( "missing closing ')'");
   return n;
}

//
//  Parse a value: this is either a bracketed sub-expression, a defined
//  variable, a function call, or a numeric constant.
//

static struct NODE * parse_value( void)
{
   char *s, name[100];
   struct NODE *n;

   if( parse_char( '('))
   {
      n = parse_tertiary_expression();
      if( !parse_char( ')')) parse_error( "missing )");
   }
   else
   if( parse_variable( name, FALSE))
   {
      if( (n = parse_function( name)) == 0 &&
          ((n = alloc_node( NT_DEFN))->u.defn = find_defn(name)) == 0)
             parse_error( "undefined symbol [%s]", name);
   }
   else
   {
      skip_whitespace(); s = pb;
      (n = alloc_node( NT_CONST))->u.val = parse_float();
      if( s == pb) parse_error_context( "not a valid expression");
   }

   return n;
}

static struct NODE * parse_unary_expression( void)
{
   struct NODE *n;
   int c;

   if( (c = parse_char( '-')) != 0 ||
       (c = parse_char( '+')) != 0)
      (end_of(n = parse_value())->next = alloc_node( NT_OPER1))->u.op = c;
   else
      n = parse_value();

   return n;
}

//
//  Parse high-precedence operators.
//

static struct NODE * parse_mul_expression( void)
{
   int c;
   struct NODE *n = parse_unary_expression();

   while( (c = parse_char( '*')) != 0 ||
          (c = parse_char( '/')) != 0)
   {
      struct NODE *e = end_of(n)->next = parse_unary_expression();
      (end_of(e)->next = alloc_node( NT_OPER2))->u.op = c;
   }

   return n;
}

//
//  Parse low-precedence operators.
//

static struct NODE * parse_sum_expression( void)
{
   int c;
   struct NODE *n = parse_mul_expression();

   while( (c = parse_char( '+')) != 0 ||
          (c = parse_char( '-')) != 0)
   {
      struct NODE *e = end_of(n)->next = parse_mul_expression();
      (end_of(e)->next = alloc_node( NT_OPER2))->u.op = c;
   }

   return n;
}

//
//  Parse very low precedence operators.
//

static struct NODE * parse_test_expression( void)
{
   int c;
   struct NODE *n = parse_sum_expression();

   while( (c = parse_char( '>')) != 0 ||
          (c = parse_char( '<')) != 0 ||
          (c = parse_char( '=')) != 0)
   {
      struct NODE *e = end_of(n)->next = parse_test_expression();
      (end_of(e)->next = alloc_node( NT_OPER2))->u.op = c;
   }

   return n;
}

//
//  Tertiary expressions,  A ? B : C
//

static struct NODE * parse_tertiary_expression( void)
{
   struct NODE *n = parse_test_expression();

   if( !parse_char( '?')) return n;

   end_of( n)->next = parse_tertiary_expression();

   if( !parse_char( ':')) parse_error( "expecting ':' in tertiary operator");
   end_of( n)->next = parse_tertiary_expression();

   end_of( n)->next = alloc_node( NT_OPER3);
   return n;
}

//
//  Parse a variable definition:  = expression
//

static void parse_defn( char *name)
{
   if( !parse_char( '=')) parse_error_context( "syntax error in definition");
   if( find_defn( name)) parse_error( "duplicate definition of %s", name);
   alloc_defn( name)->expr = parse_tertiary_expression();
}

//
//  Parse an expression followed by an optional units symbol.
//

static struct NODE *parse_scaled_expression( void)
{
   static char *unit_str = "m";      // Units indicator for output measurements
   static double unit_scale = 1.0; // Multiplier applied to input measurements,
                                   // and divider for output measurements 

   struct NODE *n = parse_tertiary_expression();

   //  Convert AWG to metres.
   double awg_to_m( double a) { return 7.348e-3/pow(1.122932, a-1)/2; }

   if( parse_word( "in") ||
       parse_word( "inch") ||
       parse_word( "inches")) return scale_expression( n, 1/39.37, 0, "in");

   if( parse_word( "ft") ||
       parse_word( "feet") ||
       parse_word( "foot")) return scale_expression( n, 12/39.37, 0, "ft");

   if( parse_word( "mm")) return scale_expression( n, 1e-3, 0, "mm");
   if( parse_word( "cm")) return scale_expression( n, 1e-2, 0, "cm");

   if( parse_word( "awg")) return scale_expression( n, 0, awg_to_m, "awg");

   if( parse_word( "kv")) return scale_expression( n, 1e3, 0, "kV");
   if( parse_word( "v")) return scale_expression( n, 1, 0, "V");
   if( parse_word( "pf")) return scale_expression( n, 1e-12, 0, "V");
   if( parse_word( "nf")) return scale_expression( n, 1e-9, 0, "V");
   if( parse_word( "uf")) return scale_expression( n, 1e-6, 0, "V");

   return scale_expression( n, unit_scale, 0, unit_str);
}

static double parse_scaled_value( void)
{
   struct NODE *n = parse_scaled_expression();
   return evaluate_scaled( n);
}

static double parse_tertiary_value( void)
{
   struct NODE *n = parse_scaled_expression();
   return evaluate( n);
}

static struct POINT * parse_point( void)
{
   struct POINT *p = allocate( struct POINT);

   p->x = parse_scaled_value();
   if( !parse_char( ',')) parse_error( "expecting comma after x coordinate");
   p->y = parse_scaled_value();
   if( !parse_char( ',')) parse_error( "expecting comma after y coordinate");
   p->z = parse_scaled_value();

   return p;
}

static struct AXIS *parse_axis( void)
{
   struct AXIS *a = allocate( struct AXIS);

   a->pa = M_PI/180 * parse_tertiary_value();
   if( !parse_char( ',')) parse_error( "expecting comma after polar angle");
   a->azim = M_PI/180 * parse_tertiary_value();

   return a;
}

//
//  Parse list of points defining the vertices of a coil or plate.
//

static int parse_pointlist( struct POINT ***points, int *npoints)
{
   while( parse_char( '{'))
   {
      if( eof_flag) parse_error( "unexpected end of file");

      *points = realloc( *points, (*npoints+1) * sizeof( struct POINT *));
      (*points)[*npoints] = parse_point();
      (*npoints)++;

      if( !parse_char( '}')) parse_error( "missing closing }");
   }

   return *npoints != 0;
}

//
//  Parse a TOROID description clause:  TOROID { ... }
//

static void parse_toroid( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_TOROID);
   struct TOROID *t = &comp->u.toroid;

   t->nt = NG_TOROID;

   if( !parse_char( '{')) parse_error( "expecting { after TOROID");

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "outer_radius"))
      {
         t->outer_radius = parse_scaled_value();
         if( t->outer_radius <= 0) parse_error( "invalid radius for toroid");
      }
      else
      if( parse_word( "inner_radius"))
         t->inner_radius = parse_scaled_value();
      else
      if( parse_word( "center")) t->center = parse_point();
      else 
      if( parse_word( "axis")) t->axis = parse_axis();
      else
      if( parse_word( "tiles")) t->nt = (int) parse_tertiary_value();
      else
         parse_error_context( "unrecognised TOROID parameter");
   }

   if( !t->outer_radius) parse_error( "missing outer radius of toroid");
   if( !t->center) parse_error( "toroid needs center point");
   if( !t->axis) parse_error( "toroid needs axis");

   if( t->nt <= 0) parse_error( "invalid number of tiles");
   t->nt *= tile_scale;
}

//
//  Parse a SPHERE description clause:  SPHERE { ... }
//

static void parse_sphere( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_SPHERE);
   struct SPHERE *s = &comp->u.sphere;

   s->nt = NG_SPHERE;

   if( !parse_char( '{')) parse_error( "expecting { after SPHERE");

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "radius"))
      {
         s->radius = parse_scaled_value();
         if( s->radius <= 0) parse_error( "invalid radius for sphere");
      }
      else
      if( parse_word( "center")) s->center = parse_point();
      else
      if( parse_word( "tiles")) s->nt = (int) parse_tertiary_value();
      else
         parse_error_context( "unrecognised SPHERE parameter");
   }

   if( !s->center) parse_error( "sphere needs center point");
   if( !s->radius) parse_error( "sphere needs radius");

   if( s->nt <= 0) parse_error( "invalid number of tiles");
   s->nt *= tile_scale;
}

//
//  Parse a DISC description clause:   DISC { ... }
//

static void parse_disc( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_DISC);
   struct DISC *d = &comp->u.disc;
  
   d->nt = NG_DISC;

   if( !parse_char( '{')) parse_error( "expecting { after DISC");

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "radius1")) d->radius1 = parse_scaled_value();
      else
      if( parse_word( "radius2")) d->radius2 = parse_scaled_value();
      else
      if( parse_word( "radius"))
      {
         d->radius1 = 0;
         d->radius2 = parse_scaled_value();
      }
      else
      if( parse_word( "center")) d->center = parse_point();
      else 
      if( parse_word( "axis")) d->axis = parse_axis();
      else
      if( parse_word( "uniform")) d->uniform = TRUE;
      else
      if( parse_word( "tiles")) d->nt = (int) parse_tertiary_value();
      else
         parse_error_context( "unrecognised DISC parameter");
   }

   if( !d->center) parse_error( "disc needs center point");
   if( !d->axis) parse_error( "disc needs axis");

   if( d->nt <= 0) parse_error( "invalid number of tiles");
   d->nt *= tile_scale;
}

//
//  Parse a CYLINDER description clause:  CYL[INDER] { ... }
//

static void parse_cyl( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_CYL);
   struct CYL *c = &comp->u.cyl;

   c->nt = NG_CYL;

   if( !parse_char( '{')) parse_error( "expecting { after CYL");

   int has_length = FALSE;

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "radius1")) c->radius1 = parse_scaled_value();
      else
      if( parse_word( "radius2")) c->radius2 = parse_scaled_value();
      else
      if( parse_word( "radius")) 
         c->radius1 = c->radius2 = parse_scaled_value();
      else
      if( parse_word( "end1")) c->end1 = parse_point();
      else
      if( parse_word( "end2")) c->end2 = parse_point();
      else 
      if( parse_word( "axis")) c->axis = parse_axis();
      else
      if( parse_word( "length"))
      {
         if( has_length) parse_error( "length already specified");
         c->length = parse_scaled_value();
         has_length = TRUE;
      }
      else
      if( parse_word( "tiles")) c->nt = (int) parse_tertiary_value();
      else
      if( parse_word( "uniform")) c->uniform = TRUE;
      else
         parse_error_context( "unrecognised CYL parameter");
   }

   if( !c->end1) parse_error( "must specify end1");
   if( c->axis && c->end2) parse_error( "both axis and end2 given");
   if( !c->axis && !c->end2) parse_error( "neither axis nor end2 given");

   if( c->axis && !has_length) parse_error( "needs length with axis");
   if( !c->axis && has_length) parse_error( "length not required");

   if( c->nt <= 0) parse_error( "invalid number of tiles");
   c->nt *= tile_scale;

   if( !c->axis)
   {
      c->axis = allocate( struct AXIS);
      to_polar( c->end2, c->end1, &c->axis->pa, &c->axis->azim, &c->length);
   }
   else
   {
      c->end2 = allocate( struct POINT);
      forepoint( c->axis, c->length, c->end2, c->end1);
   }
}

//
//  Parse a RECTANGLE description clause:   RECT[ANGLE] { ... }
//

static void parse_rectangle( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_RECT);
   struct RECT *d = &comp->u.rect;
   d->nt = NG_RECT;

   if( !parse_char( '{')) parse_error( "expecting { after RECTANGLE");

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "center")) d->center = parse_point();
      else 
      if( parse_word( "edge1")) d->edge1 = parse_point();
      else 
      if( parse_word( "edge2")) d->edge2 = parse_point();
      else
      if( parse_word( "tiles")) d->nt = (int) parse_tertiary_value();
      else
      if( parse_word( "uniform")) d->uniform = TRUE;
      else
         parse_error_context( "unrecognised RECTANGLE parameter");
   }

   if( !d->center) parse_error( "rectangle needs center point");

   if( !d->edge1 && !d->edge2)
      parse_error( "rectangle needs edge center points");

   if( !d->edge1) parse_error( "rectangle needs edge1");
   if( !d->edge2) parse_error( "rectangle needs edge2");

   struct POINT u = sub_3D( *d->edge1, *d->center);
   struct POINT v = sub_3D( *d->edge2, *d->center);

   if( fabs(inner_prod_3D(u,v)) > 1e-3 * mag_3D(u) * mag_3D(v))
      parse_error( "rectangle edges are not orthogonal");

   if( d->nt <= 0) parse_error( "invalid number of tiles");
   d->nt *= tile_scale;
}

//
//  Parse a PLATE description clause: PLATE { ... }
//

static void parse_plate( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_PLATE);
   struct PLATE *p = &comp->u.plate;
   p->nt = NG_PLATE;

   if( !parse_char( '{')) parse_error( "expecting { after PLATE");

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "points"))
          parse_pointlist( &p->points, &p->npoints);
      else
      if( parse_word( "tiles")) p->nt = (int) parse_tertiary_value();
      else
      if( parse_word( "uniform")) p->uniform = TRUE;
      else
         parse_error_context( "unrecognised PLATE parameter");
   }

   if( p->npoints < 3) parse_error( "at least three points required");
   if( p->nt <= 0) parse_error( "invalid number of tiles");
   p->nt *= tile_scale;
}

static void parse_wire( struct ELECTRODE *e)
{
   struct COMPONENT *comp = create_component( e, CT_WIRE);
   struct WIRE *w = &comp->u.wire;

   w->nt = NG_WIRE;

   if( !parse_char( '{')) parse_error( "expecting { after WIRE");

   int has_length = FALSE;

   while( !parse_char( '}'))
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "wirad")) w->wirad = parse_scaled_value();
      else
      if( parse_word( "end1")) w->end1 = parse_point();
      else
      if( parse_word( "end2")) w->end2 = parse_point();
      else 
      if( parse_word( "axis")) w->axis = parse_axis();
      else
      if( parse_word( "length"))
      {
         if( has_length) parse_error( "length already specified");
         w->length = parse_scaled_value();
         has_length = TRUE;
      }
      else
      if( parse_word( "tiles")) w->nt = (int) parse_tertiary_value();
      else
      if( parse_word( "uniform")) w->uniform = TRUE;
      else
         parse_error_context( "unrecognised WIRE parameter");
   }

   if( !w->end1) parse_error( "must specify end1");
   if( w->axis && w->end2) parse_error( "both axis and end2 given");
   if( !w->axis && !w->end2) parse_error( "neither axis nor end2 given");

   if( w->axis && !has_length) parse_error( "needs length with axis");
   if( !w->axis && has_length) parse_error( "length not required");

   if( w->wirad <= 0) parse_error( "missing or invalid wire radius");

   if( w->nt <= 0) parse_error( "invalid number of tiles");
   w->nt *= tile_scale;

   if( !w->axis)
   {
      w->axis = allocate( struct AXIS);
      to_polar( w->end2, w->end1, &w->axis->pa, &w->axis->azim, &w->length);
   }
   else
   {
      w->end2 = allocate( struct POINT);
      forepoint( w->axis, w->length, w->end2, w->end1);
   }
}

//
//  Parse a COIL description clause:  COIL { ... }
//

static void parse_coil( struct WORLD *W)
{
   if( !parse_char( '{')) parse_error( "expecting { after COIL");

   struct COIL *c = create_coil( W);
   c->nt = NG_COIL;

   int has_length = FALSE;

   //
   //  c->taps[] lists all the connections to the coil, including the two ends
   //

   c->taps = allocate( double);                  // Allocate one entry for end1
   c->ntaps = 1;        // Leave c->taps[0] == 0 to account for end1 at 0 turns

   while( !parse_char( '}')) 
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "radius1"))
         c->radius1 = parse_scaled_value();
      else
      if( parse_word( "radius2"))
         c->radius2 = parse_scaled_value();
      else
      if( parse_word( "radius")) c->radius1 = 
                                 c->radius2 = parse_scaled_value();
      else
      if( parse_word( "end1"))
      {
         if( c->end1 || c->end1_poly) parse_error( "end1 given twice");
         int n = 0;
         if( parse_pointlist( &c->end1_poly, &n))
         {
            if( c->npoints && n != c->npoints)
               parse_error( "unequal number of polygon points");
            else c->npoints = n;
         }
         else
            c->end1 = parse_point();
      }
      else
      if( parse_word( "end2"))
      {
         if( c->end2 || c->end2_poly) parse_error( "end2 given twice");
         int n = 0;
         if( parse_pointlist( &c->end2_poly, &n))
         {
            if( c->npoints && n != c->npoints)
               parse_error( "unequal number of polygon points");
            else c->npoints = n;
         }
         else
            c->end2 = parse_point();
      }
      else 
      if( parse_word( "axis")) c->axis = parse_axis();
      else
      if( parse_word( "length"))
      {
         if( has_length) parse_error( "length already specified");
         c->length = parse_scaled_value();
         has_length = TRUE;
      }
      else
      if( parse_word( "name")) parse_name( &c->name);
      else
      if( parse_word( "wirad")) c->wirad = parse_scaled_value();
      else
      if( parse_word( "turns")) c->turns = parse_tertiary_value();
      else
      if( parse_word( "tap"))
      {
         c->taps = realloc( c->taps, (c->ntaps+1) * sizeof( double));
         c->taps[c->ntaps] = parse_tertiary_value();
         if( c->taps[c->ntaps] <= 0) bailout( "invalid tap position");
         c->ntaps++;
      }
      else
      if( parse_word( "sections")) c->Ns = parse_tertiary_value();
      else
      if( parse_word( "tiles")) c->nt = (int) parse_tertiary_value();
      else
      if( parse_word( "dphi")) c->dphi = parse_tertiary_value();
      else
         parse_error_context( "unrecognised COIL parameter");
   }

   //
   //  Some sanity checks to the coil description.
   //

   if( c->turns == 0) parse_error( "missing or invalid number of turns");
   if( c->wirad <= 0) parse_error( "missing or invalid wire radius");

   if( !c->npoints)   // Circular cross-section?
   {
      if( c->radius1 <= 0||
          c->radius2 <= 0) parse_error( "missing or invalid coil radius");

      if( !c->end1) parse_error( "must specify end1");
      if( c->axis && c->end2) parse_error( "both axis and end2 given");
      if( !c->axis && !c->end2) parse_error( "neither axis nor end2 given");

      if( c->axis && !has_length) parse_error( "needs length with axis");
      if( !c->axis && has_length) parse_error( "length not required");
   }
   else   // Polygonal cross-section
   {
      if( c->npoints < 3) parse_error( "too few points for polygonal winding");
      if( c->axis) parse_error( "axis not required with polygonal winding");
      if( !has_length &&
          (!c->end1_poly || !c->end2_poly))
          parse_error( "needs length with polygonal winding");
      if( has_length && c->end1_poly && c->end2_poly)
          parse_error( "length not required");
   }

   //
   //  Fill in any missing dependent values.
   //

   if( c->npoints)  // Polygonal winding?
   {
      // Set c->end1 to the approximate center of the cross section
      if( !c->end1 && c->end1_poly)
      {
         c->end1 = allocate( struct POINT);
         for( int i = 0; i < c->npoints; i++)
            sum_3D( c->end1, *c->end1_poly[i]);
         scale_3D( c->end1, 1.0/c->npoints);
      }

      // Set c->end2 to the approximate center of the cross section
      if( !c->end2 && c->end2_poly)
      {
         c->end2 = allocate( struct POINT);
         for( int i = 0; i < c->npoints; i++)
            sum_3D( c->end2, *c->end2_poly[i]);
         scale_3D( c->end2, 1.0/c->npoints);
      }

      // Calculate axis from the plane and right-handed ordering of the points
      // Use end1 if given, otherwise end2
      struct POINT v = origin;
      for( int i = 0; i < c->npoints - 1; i++)
      {
         struct POINT d1, d2;
         if( c->end1_poly)
         {
            d1 = sub_3D( *c->end1_poly[i+0], *c->end1);
            d2 = sub_3D( *c->end1_poly[i+1], *c->end1);
         }
         else
         {
            d1 = sub_3D( *c->end2_poly[i+0], *c->end2);
            d2 = sub_3D( *c->end2_poly[i+1], *c->end2);
         }
         struct POINT xp = xprod_3D( d1, d2);
         sum_3D( &v, xp);
      }
      norm_3D( &v);

      c->axis = allocate( struct AXIS);
      to_polar( &v, &origin, &c->axis->pa, &c->axis->azim, NULL);
      logmsg( 2, "%s axis %.3f,%.3f", c->name,
                         c->axis->pa * 180/M_PI, c->axis->azim * 180/M_PI);

      if( has_length && !c->end2_poly)
      {
         c->end2_poly = array( c->npoints, struct POINT *);
         for( int i = 0; i < c->npoints; i++)
         {
            c->end2_poly[i] = allocate( struct POINT);
            forepoint( c->axis, c->length, c->end2_poly[i], c->end1_poly[i]);
         }
      }

      if( !has_length)
      {
         c->length = mag_3D( sub_3D( *c->end2, *c->end1));
         has_length = TRUE;
      }
   }

   if( !c->axis)
   {
      //
      //  No axis given, so calculate axis and length from the given ends
      //

      c->axis = allocate( struct AXIS);
      to_polar( c->end2, c->end1, &c->axis->pa, &c->axis->azim, &c->length);
   }
   else
   {
      //
      //  Axis given, so calculate end2 from the given axis and length.
      //

      c->end2 = allocate( struct POINT);
      forepoint( c->axis, c->length, c->end2, c->end1);
   }

   if( c->nt <= 0) parse_error( "invalid number of tiles");
   c->nt *= tile_scale;

   if( !c->Ns) c->Ns = 10;                   // Default number of coil sections
   if( fabs( c->turns) < c->Ns) c->Ns = fabs( c->turns);

   //
   //  Determine if we have a loop coil:  length is zero and the two ends are
   //  the same.
   //

   if( !c->length)
   {
      if( !c->npoints && c->radius1 == c->radius2) c->loop = TRUE;

      if( c->npoints)
      {
         int i;
         for( i = 0; i < c->npoints; i++)
            if( c->end1_poly[i]->x != c->end2_poly[i]->x ||
                c->end1_poly[i]->y != c->end2_poly[i]->y ||
                c->end1_poly[i]->z != c->end2_poly[i]->z) break;
         if( i == c->npoints) c->loop = TRUE;
      }

      if( c->loop)
      {
         logmsg( 1, "loop coil %s", c->name);
         c->Ns = 1;
      }
   }

   //
   // Append end2 to the c->taps[] array
   //

   c->taps = realloc( c->taps, (c->ntaps+1) * sizeof( double));
   c->taps[c->ntaps] = c->turns;
   c->ntaps++;

   // Todo: sort taps[] into increasing order and check all same polarity.

   // Check pitch and radius
   // Todo: check pitch on polygon coils
   if( !c->loop && !c->npoints)
   {
      double dr = c->radius2 - c->radius1;
      double slant_length = sqrt( dr*dr + c->length*c->length);
      if( slant_length/c->turns < 2 * c->wirad)
         logmsg( 0, "coil %s pitch less than wire diameter", c->name);
   }
}

static void parse_field( struct WORLD *W)
{
   if( !parse_char( '{')) parse_error( "expecting { after FIELD");

   while( !parse_char( '}')) 
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "electric"))
      {
         if( WORLD.Efield) parse_error( "cannot have more than one E-field");
         W->Efield = parse_axis();
      }
      else
      if( parse_word( "magnetic"))
      {
         if( WORLD.Bfield) parse_error( "cannot have more than one B-field");
         W->Bfield = parse_axis();
      }
      else
         parse_error_context( "unrecognised statement");
   }
}

static void parse_ground( struct WORLD *W)
{
   if( !parse_char( '{')) parse_error( "expecting { after GROUND");

   if( W->g) parse_error( "cannot have more than one ground");
   W->g = allocate( struct GROUND);

   while( !parse_char( '}')) 
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "center")) W->g->center = parse_point();
      else
      if( parse_word( "axis")) W->g->axis = parse_axis();
      else
         parse_error_context( "unrecognised statement");
   }

   if( !W->g->axis) bailout( "ground needs axis");
   if( !W->g->center) bailout( "ground needs center");

   forepoint( W->g->axis, 1.0, &W->g->norm, NULL); 
   W->g->offset = -inner_prod_3D( *W->g->center, W->g->norm);
}

//
//  Parse an electrode description:  ELECTRODE { ... }
//

static void parse_electrode( struct WORLD *W)
{
   if( !parse_char( '{')) parse_error( "expecting { after ELECTRODE");

   struct ELECTRODE *e = create_electrode( W, NULL);

   while( !parse_char( '}')) 
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "name")) parse_name( &e->name);
      else
      if( parse_word( "sphere")) parse_sphere( e);
      else
      if( parse_word( "toroid")) parse_toroid( e);
      else
      if( parse_word( "disc")) parse_disc( e);
      else
      if( parse_word( "cyl") ||
          parse_word( "cylinder")) parse_cyl( e);
      else
      if( parse_word( "rect") ||
          parse_word( "rectangle")) parse_rectangle( e);
      else
      if( parse_word( "plate")) parse_plate( e);
      else
      if( parse_word( "wire")) parse_wire( e);
      else
         parse_error_context( "unrecognised electrode component");
   }

   if( !e->components) parse_error( "electrode has no components");
}

static void parse_world( struct WORLD *W);

static void parse_transform_world( struct WORLD *W)
{
   if( !parse_char( '{')) parse_error( "expecting { after shift");

   struct WORLD *WT = allocate( struct WORLD);

   while( !parse_char( '}')) 
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_word( "shift"))
      {
         struct POINT *offset = parse_point();
         shift_world( WT, *offset);
      }
      else
      if( parse_word( "rotate"))
      {
         struct AXIS *axis = parse_axis();
         double angle = M_PI/180 * parse_tertiary_value();
         rotate_world( WT, *axis, angle); 
      }
      else
         parse_world( WT);
   }

   sum_world( W, WT);
}

//
//  Parse world
//

static void parse_world( struct WORLD *W)
{
   char name[ 100];

   while( skip_whitespace(), !eof_flag)
   {
      if( parse_word( "field")) parse_field( W);
      else
      if( parse_word( "coil")) parse_coil( W);
      else
      if( parse_word( "electrode")) parse_electrode( W);
      else
      if( parse_word( "ground")) parse_ground( W);
      else
      if( parse_word( "transform")) parse_transform_world( W);
      else
      if( parse_variable( name, TRUE)) parse_defn( name);
      else
         break;
   }
}

//
//  Parse the input file.
//

static void parse_input_file( void)
{
   char temp[ 100];
   pbuff[ 0] = 0; pb = pbuff;
   eof_flag = lino = 0;

   sprintf( temp, "%s.in", WORLD.sysname);
   if( (fdi = fopen( temp, "r")) == NULL)
      bailout( "cannot open input file %s, %s", temp, strerror( errno));

   parse_world( &WORLD);

   if( !eof_flag) parse_error_context( "unrecognised statement");

   fclose( fdi);
}

//
//  Set up coils.
//  
//  Each coil is broken into segments by its taps.  If no taps then the entire
//  coil is one segment, otherwise for N taps there are N+1 segments.
//
//  Each coil is constructed from c->Ns sections.  Ns is shared amongst the
//  segments and section sizes adjusted so that the taps coincide with section
//  boundaries.  c->Ns will be adjusted slightly if necessary.
//
//  Note: A coil with no taps has c->ntaps == 2 because ntaps counts the ends.
//

static void decompose_coils( void)
{
   struct SEG {
      int Ns;
      
      double radius1, radius2;
      struct POINT end1;
      struct POINT end2;
      double length;
      double turns;
      double start;
   };

   for( struct COIL *c = WORLD.coils; c; c = c->next)
   {
      int nsegs = c->ntaps - 1;  // Number of segments between ends and taps
      struct SEG *segs = array( nsegs, struct SEG);
      int seg;

      c->tsect = array( c->ntaps, int);

      if( nsegs == 1) // No taps, so entire coil in one segment
      {
         c->tsect[0] = 0;
         c->tsect[1] = c->Ns;

         segs[0].Ns = c->Ns;
         segs[0].radius1 = c->radius1;
         segs[0].radius2 = c->radius2;

         segs[0].end1.x = c->end1->x;
         segs[0].end1.y = c->end1->y;
         segs[0].end1.z = c->end1->z;

         segs[0].end2.x = c->end2->x;
         segs[0].end2.y = c->end2->y;
         segs[0].end2.z = c->end2->z;

         segs[0].length = c->length; 

         segs[0].turns = c->turns;
         segs[0].start = 0;
      }
      else // Coil has more than one segment
      {
         // Decide how many sections for each segment
         int new_Ns = 0;

         for( seg = 0; seg < nsegs; seg++)
         {
            c->tsect[seg] = new_Ns;

            struct SEG *sp = segs + seg;
            sp->start = c->taps[seg];
            sp->turns = c->taps[seg+1] - c->taps[seg];

            sp->Ns = (int) ceil( fabs( sp->turns) * c->Ns/(double) c->turns);
            if( !sp->Ns) sp->Ns = 1;
            new_Ns += sp->Ns;
            logmsg( 1, "coil %s seg %d: turns %.1f from %.1f Ns %d",
                c->name, seg, sp->turns, sp->start, sp->Ns);
         }

         if( new_Ns != c->Ns)
         {
            logmsg( 1, "revised %s Ns %d to %d", c->name, c->Ns, new_Ns);
            c->Ns = new_Ns;
         }

         c->tsect[nsegs] = c->Ns;

         // Increments per turn
         double dr = (c->radius2 - c->radius1)/c->turns;
         double dx = (c->end2->x - c->end1->x)/c->turns;
         double dy = (c->end2->y - c->end1->y)/c->turns;
         double dz = (c->end2->z - c->end1->z)/c->turns;

         for( seg = 0; seg < nsegs; seg++)
         {
            struct SEG *sp = segs + seg;

            sp->radius1 = c->radius1 + dr * sp->start;
            sp->radius2 = c->radius1 + dr * (sp->start + sp->turns);
   
            sp->end1.x = c->end1->x + dx * sp->start;
            sp->end1.y = c->end1->y + dy * sp->start;
            sp->end1.z = c->end1->z + dz * sp->start;
   
            sp->end2.x = sp->end1.x + dx * sp->turns;
            sp->end2.y = sp->end1.y + dy * sp->turns;
            sp->end2.z = sp->end1.z + dz * sp->turns;
   
            sp->length = c->length * sp->turns/c->turns; 
         }
      }

      //
      //  Now break up each coil segment into sections.
      //

      int unit = 0;                    // Numbering of sections within the coil
      c->sections = array( c->Ns, struct CSECT);

      for( seg = 0; seg < nsegs; seg++)
      {
         struct SEG *sp = segs + seg;
         int n;
  
         // Increments per section 
         double dr = (sp->radius2 - sp->radius1)/sp->Ns;
         double dx = (sp->end2.x - sp->end1.x)/sp->Ns;
         double dy = (sp->end2.y - sp->end1.y)/sp->Ns;
         double dz = (sp->end2.z - sp->end1.z)/sp->Ns;

         double start = sp->start; 
         for( n = 0; n < sp->Ns; n++)
         {
            struct CSECT *s = c->sections + unit;
            s->unit = unit++;

            s->end1 = (struct POINT) { sp->end1.x + n * dx,
                                       sp->end1.y + n * dy,
                                       sp->end1.z + n * dz };
            s->end2 = (struct POINT) { s->end1.x + dx,
                                       s->end1.y + dy,
                                       s->end1.z + dz };
            s->axis = *c->axis;
            s->length = sp->length / sp->Ns;

            s->turns = sp->turns / sp->Ns;
            s->start = start;   start += s->turns;

            s->radius1 = sp->radius1 + n * dr;
            s->radius2 = sp->radius1 + (n+1) * dr;

            if( c->npoints) // Polygonal coil?
            {
               s->end1_poly = array( c->npoints, struct POINT);
               s->end2_poly = array( c->npoints, struct POINT);
               for( int i = 0; i < c->npoints; i++)
               {
                  struct POINT tp, dp = sub_3D( *c->end2_poly[i],
                                                *c->end1_poly[i]);
                  tp = dp;
                  scale_3D( &tp, s->start/c->turns);
                  s->end1_poly[i] = add_3D( *c->end1_poly[i], tp);

                  tp = dp;
                  scale_3D( &tp, (s->start + s->turns)/c->turns);
                  s->end2_poly[i] = add_3D( *c->end1_poly[i], tp);
               }
            }
         }
      }

      free( segs);
   }
}

//
//  Create electrodes to represent the capacitive effects of each coil.
//
//  One electrode for each coil section.
//
//  Coils with circular cross-section have one cylinder per coil section.
//  Polygonal coils have each section built from rectangles.
//

static void build_coil_electrodes( void)
{
   for( struct COIL *c = WORLD.coils; c; c = c->next)
      for( int n = 0; n < c->Ns; n++)
      {
         struct CSECT *s = c->sections + n;

         s->e = create_electrode( &WORLD, c);
         s->e->unit = s->unit;

         if( asprintf( &s->e->name, "L%d-%d", c->ident, s->unit) < 0)
                bailout( "out of memory");

         if( c->loop && c->npoints)   // Polygonal loop
         {
            double wirad = c->wirad * sqrt( s->turns * 2);
            struct WIRE *list[c->npoints];

            for( int i1 = 0; i1 < c->npoints; i1++)
            {
               int i2 = (i1 + 1) % c->npoints;

               struct COMPONENT *comp = create_component( s->e, CT_WIRE);
               struct WIRE *w = &comp->u.wire;
               list[i1] = w;
               w->uniform = TRUE;
               w->wirad = wirad;
               w->nt = c->nt / c->npoints;
               w->end1 = allocate( struct POINT);
               *w->end1 = s->end1_poly[i1];
               w->end2 = allocate( struct POINT);
               *w->end2 = s->end1_poly[i2];

               w->axis = allocate( struct AXIS);
               to_polar( w->end2, w->end1,
                         &w->axis->pa, &w->axis->azim, &w->length);
            }

            // Link each wire to next in the loop
            for( int i1 = 0; i1 < c->npoints; i1++)
            {
               int i2 = (i1 + 1) % c->npoints;
               list[i1]->next = list[i2];
            }
         }
         else
         if( c->loop && !c->npoints)  // Circular loop
         {
            struct COMPONENT *comp = create_component( s->e, CT_RING);
            struct RING *r = &comp->u.ring;
            r->center = &s->end1;
            r->axis = c->axis;
            r->radius = s->radius1;
            r->wirad = c->wirad * sqrt( s->turns * 2);
            r->nt = c->nt;
         }
         else
         if( !c->npoints && s->length != 0) // Circular cross-section solenoid?
         {
            struct COMPONENT *comp = create_component( s->e, CT_CYL);
            struct CYL *cyl = &comp->u.cyl;
   
            cyl->nt = c->nt;                     // Number of tiles per section
            cyl->uniform = TRUE;

            cyl->radius1 = s->radius1;
            cyl->radius2 = s->radius2;

            cyl->axis = c->axis; 
            cyl->length = s->length;
            cyl->end1 = &s->end1;
            cyl->end2 = &s->end2;
         }
         else
         if( !c->npoints)  // Circular cross-section planar coil?
         {
            struct COMPONENT *comp = create_component( s->e, CT_DISC);
            struct DISC *disc = &comp->u.disc;
   
            disc->nt = c->nt;                    // Number of tiles per section
            disc->uniform = TRUE;
            disc->radius1 = s->radius1;
            disc->radius2 = s->radius2;

            disc->axis = c->axis; 
            disc->center = &s->end1;
         }
         else  // Polygonal cross-section
            for( int i1 = 0; i1 < c->npoints; i1++)
            {
               int i2 = (i1 + 1) % c->npoints;

               struct COMPONENT *comp = create_component( s->e, CT_PLATE);
               struct PLATE *p = &comp->u.plate;
               p->uniform = TRUE;
               p->npoints = 4;
               p->nt = c->nt / c->npoints;
               p->points = array( 4, struct POINT *);
               p->points[0] = allocate( struct POINT);
               *p->points[0] = s->end1_poly[i1];
               p->points[1] = allocate( struct POINT);
               *p->points[1] = s->end2_poly[i1];
               p->points[2] = allocate( struct POINT);
               *p->points[2] = s->end2_poly[i2];
               p->points[3] = allocate( struct POINT);
               *p->points[3] = s->end1_poly[i2];
            }
   }
}

//
//  Create a mirror image of a coil for magnetic
//  calculations in the presence of a ground plane.  Just do the minimum
//  setup required to be able to call Lmutual().
//

static void create_image_coils( void)
{
   for( struct COIL *c = WORLD.coils; c; c = c->next)
   {
      struct COIL *ci = allocate( struct COIL);

      ci->end1 = allocate( struct POINT);
      ci->end2 = allocate( struct POINT);
      image_point( c->end1, ci->end1);
      image_point( c->end2, ci->end2);

      ci->axis = allocate( struct AXIS);
      struct POINT p; forepoint( c->axis, 1.0, &p, NULL);
      struct POINT pi;  image_point( &p, &pi);
      to_polar( &pi, &origin,  &ci->axis->pa, &ci->axis->azim, &ci->length);
      ci->length = c->length;

      ci->radius1 = c->radius1;
      ci->radius2 = c->radius2;
      ci->turns = c->turns;
      ci->length = c->length;
      ci->Ns = c->Ns;
      ci->nt = c->nt;

      ci->taps = c->taps;
      ci->ntaps = c->ntaps;
      ci->tsect = c->tsect;

      ci->dphi = c->dphi;

      ci->npoints = c->npoints;
      ci->end1_poly = array( ci->npoints, struct POINT *);
      ci->end2_poly = array( ci->npoints, struct POINT *);
      for( int i = 0; i < ci->npoints; i++)
      {
         ci->end1_poly[i] = allocate( struct POINT);
         image_point( c->end1_poly[c->npoints - i - 1], ci->end1_poly[i]);

         ci->end2_poly[i] = allocate( struct POINT);
         image_point( c->end2_poly[c->npoints - i - 1], ci->end2_poly[i]);
      }

      ci->sections = array( ci->Ns, struct CSECT);
      for( int i = 0; i < ci->Ns; i++)
      {
         struct CSECT *s = c->sections + i;
         struct CSECT *si = ci->sections + i;

         image_point( &s->end1, &si->end1);
         image_point( &s->end2, &si->end2);
         si->axis = *ci->axis;
         si->length = s->length;
         si->turns = s->turns;
         si->start = s->start;
         si->radius1 = s->radius1;
         si->radius2 = s->radius2;

         si->end1_poly = array( ci->npoints, struct POINT);
         si->end2_poly = array( ci->npoints, struct POINT);
         for( int j = 0; j < ci->npoints; j++)
         {
            int ji = c->npoints - j - 1;
            image_point( &s->end1_poly[ji], &si->end1_poly[j]);
            image_point( &s->end2_poly[ji], &si->end2_poly[j]);
         }
         s->image = si;
      }

      c->image = ci;
   }
}

///////////////////////////////////////////////////////////////////////////////
//  Parallelisation Functions                                                //
///////////////////////////////////////////////////////////////////////////////

//
//  An array of worker threads for parallel calculations.
//

static sem_t w_cnt;   // Number of workers waiting for jobs

static struct WORK {
   sem_t claim;                        // Unlocked until this worker is claimed
   sem_t wait;    // Worker waits on this locked semaphore until work available

   void (*fcn)();
   void *arg1, *arg2, *arg3, *arg4, *arg5;

} work[MAXCPU];

static void *worker( void *arg)
{
   struct WORK *w = (struct WORK *) arg;

   while( 1)
   {
      sem_post( &w->claim);        // Indicate this worker is available
      sem_post( &w_cnt);           // Increment count of idle workers

      sem_wait( &w->wait);         // Wait for a job

      w->fcn( w->arg1, w->arg2, w->arg3, w->arg4, w->arg5);
   }

   return 0;
}

static struct WORK *worker_claim( void)
{
   sem_wait( &w_cnt);            // Block here until a worker becomes available

   // Claim the first available worker 
   for( int i = 0; i < ncpu; i++)
      if( sem_trywait( &work[i].claim) == 0) return work + i;

   bailout( "error in worker_claim"); // Never reached, unless w_cnt mishandled
}

//
//  Hand a function to a worker thread to get on with.
//

static void worker_run( void (*fcn)(),
                        void *arg1, void *arg2, void *arg3,
                        void *arg4, void *arg5)
{
   struct WORK *w = worker_claim();

   w->fcn = fcn;
   w->arg1 = arg1;
   w->arg2 = arg2;
   w->arg3 = arg3;
   w->arg4 = arg4;
   w->arg5 = arg5;

   sem_post( &w->wait);    // Set the worker going
}

//
//  Wait for all workers to finish by claiming them all, then releasing them.
//

static void worker_rendezvous( void)
{
   for( int i = 0; i < ncpu; i++) worker_claim();

   // Release them all.  In a separate loop in case workers themselves are
   // claiming workers.

   for( int i = 0; i < ncpu; i++)
   {
      sem_post( &work[i].claim);
      sem_post( &w_cnt);
   }
}

//
//  Create some worker threads.
//

static void worker_init( void)
{
   sem_init( &w_cnt, 0, 0);

   for( int i = 0; i < ncpu; i++)
   {
      memset( &work[i], 0, sizeof( struct WORK));
      sem_init( &work[i].claim, 0, 0);
      sem_init( &work[i].wait, 0, 0);

      pthread_t pid;
      if( pthread_create( &pid, NULL, worker, (void *) &work[i]) != 0)
         bailout( "cannot start worker thread, %s", strerror( errno));
   }
}

///////////////////////////////////////////////////////////////////////////////
//  Resistance Calculation                                                   //
///////////////////////////////////////////////////////////////////////////////

//
//  Compute the DC resistance of all coil sections.   We leave it to spice to
//  calculate the AC resistance.
//

static void do_resistance( void)
{
   for( struct COIL *c = WORLD.coils; c; c = c->next)
      if( !c->npoints)    // Circular cross-section?
         for( int s = 0; s < c->Ns; s++)
         {
            double dr = (c->radius2 - c->radius1)/c->Ns;
            double mr = c->radius1 + dr * (s + 0.5);
            double nt = fabs(c->turns)/c->Ns;
         
            double a = M_PI * c->wirad * c->wirad;   // Cross section area
         
            // 1.75e-8 is the resistivity of copper at 25 deg C
            double R = 2 * M_PI * mr * nt/a * 1.75e-8;
            store_R( c, s, R);
         }
      else   // Polygonal cross-section
         for( int s = 0; s < c->Ns; s++)
         {
            double nt = fabs(c->turns)/c->Ns;
            double a = M_PI * c->wirad * c->wirad;   // Cross section area
            double len = 0;  // Length per turn

            for( int i = 0; i < c->npoints; i++)
            {
               int j = (i + 1) % c->npoints;
               len += mag_3D( sub_3D( *c->end1_poly[j], *c->end1_poly[i]));
            }
            
            // 1.75e-8 is the resistivity of copper at 25 deg C
            double R = len * nt/a * 1.75e-8;
            store_R( c, s, R);
         }
}

///////////////////////////////////////////////////////////////////////////////
//  Capacitance Calculation                                                  //
///////////////////////////////////////////////////////////////////////////////

//
//  Self potential of rectangular tile. Exact expression based on Bancroft,
//    "A Note on the Moment Method Solution for the Capacitance
//     of a Conducting Flat Plate",
//  IEEE Transactions on Antennas and Propagation, 1997
//

static PMTYPE self_potential_rectangle( double h1, double h2)
{
   double s = sqrt( h1*h1 + h2*h2);
   PMTYPE p = h2*log((s+h1)/(s-h1)) + h1*log((s+h2)/(s-h2));
   return p / (4 * M_PI * EPSILON * h1*h2);
}

//
//  Self potential coefficient at center of a wire tile. 
//

static PMTYPE self_potential_wire( double h, double r)
{
   double hstep = h / 200;

   double z, sum = 0;
   for( z = -h/2; z < h/2; z += hstep) sum += 1/sqrt(z*z + r*r);

   return sum/(4 * M_PI * EPSILON)/200;
}

//
//  Self potential coefficient at center of a sphere tile.
//  Integrate over a spherical cap.
//

static PMTYPE self_potential_sphere( double r, double a)
{
   double h = a/(2 * M_PI * r);       // Cap height
   double b = sqrt( h * (2*r - h));   // Base radius

   double sum = 0, w, wstep = b/50;
   for( w = wstep; w < b; w += wstep)
   {
      double ring_area = 2 * M_PI * (w + wstep/2) * wstep;
      double charge_fraction = ring_area/a;

      double z = r - sqrt(r*r - w*w);
      sum += charge_fraction / sqrt(z*z + w*w);
   }

   return sum/(4 * M_PI * EPSILON);
}

//
//  Sum the charge on the electrodes and coil sections.
//
//  esrc indicates which electrode is the source (set to 1 volt), all the
//  other electrodes are destinations (set to 0 volts).
//
//  Iteration often terminates with some electrodes carrying small charges of
//  the wrong sign.  These are discarded.  If they are more than a small
//  amount they are counted and reported as 'sign errors'.  This tends to
//  happen when some component is substantially shielded from the rest of the
//  system.
//  

static void sum_charges( struct ELECTRODE *esrc)
{
   QVTYPE total_dst_charge = 0.0;       // Total charge on zero volt electrodes
   QVTYPE total_src_charge = 0.0;          // Total charge on 1 volt electrodes

   struct ELECTRODE *e;

   int sign_errs = 0;
   double sign_tol = 0.2e-12;

   for( e = WORLD.electrodes; e; e = e->next) 
   {
      if( e == esrc) 
      {
         if( e->charge < 0)
         {
            if( e->charge < -sign_tol) sign_errs++;
            e->charge = 0;
         }
         total_src_charge += e->charge;
      }
      else
      {
         if( e->charge > 0)
         {
            if( e->charge > sign_tol) sign_errs++;
            e->charge = 0;
         }
         total_dst_charge += e->charge;
         store_C( esrc, e, -e->charge);
      }

      logmsg( 2, "Q %s %.3f", e->name, e->charge * 1e12);
   }

   QVTYPE Qinf = total_src_charge + total_dst_charge;
   store_C( esrc, esrc, Qinf);

   if( sign_errs) logmsg( 0, "sign errors %d", sign_errs);
   logmsg( 2, "solve %s Qsrc %.6f Qdst %.6f", esrc->name,
           total_src_charge * 1e12, -total_dst_charge * 1e12);
}

//
//  Distribute tiles fairly evenly over a sphere using the spiral algorithm.
//

static void setup_tiles_sphere( struct ELECTRODE *e, struct SPHERE *s)
{
   int k;
   double phi_k = 0;
   double a = 4 * M_PI * s->radius * s->radius / s->nt;    // Average tile area

   PMTYPE spc = self_potential_sphere( s->radius, a);

   for( k = 1; k <= s->nt; k++)
   {
      double hk = -1 + 2.0 * (k-1)/(double)(s->nt - 1);
      double theta = acos( hk);
      double phi;

      if( k == 1 || k == s->nt) phi = 0;
      else
         phi = phi_k + 3.6/sqrt(s->nt)/sqrt(1.0 - hk*hk);

      double r = s->radius * sin(theta);
      struct POINT p = { s->center->x + r * sin(phi),
                         s->center->y + r * cos(phi),
                         s->center->z + s->radius * sin(theta - M_PI/2) };
  
      create_tile( e, p, spc);
      phi_k = phi;
   }

   logmsg( 1, "tiles for %s (sphere): %d", e->name, s->nt);
}

static struct TILE *create_oriented_tile( struct ELECTRODE *e,
      struct POINT *o,
      struct POINT *c,
      struct AXIS *a, PMTYPE spc)
{
   // Rotate by the polar angle around the y axis
   double x1 = cos(a->pa) * o->x + sin(a->pa) * o->z;
   double y1 = o->y;
   double z1 = -sin(a->pa) * o->x + cos(a->pa) * o->z;

   // Rotate by the azimuth angle around the z axis
   double x2 = cos(a->azim) * x1 - sin(a->azim)*y1;
   double y2 = sin(a->azim) * x1 + cos(a->azim)*y1;
   double z2 = z1;

   struct POINT p = { c->x + x2, c->y + y2, c->z + z2 };
   return create_tile( e, p, spc);
}

static struct TILE *create_oriented_tile_rect( struct ELECTRODE *e,
      struct POINT *o,
      struct POINT *c,
      struct AXIS *a, PMTYPE spc,
      struct POINT *ux,
      struct POINT *uy
      )
{
   struct TILE *t = create_oriented_tile( e, o, c, a, spc);

   // same rotations for unit x vector
   double x1,y1,z1;
   o=ux;
   x1 = cos(a->pa) * o->x + sin(a->pa) * o->z;
   y1 = o->y;
   z1 = -sin(a->pa) * o->x + cos(a->pa) * o->z;
   t->ux.x = cos(a->azim) * x1 - sin(a->azim)*y1;
   t->ux.y = sin(a->azim) * x1 + cos(a->azim)*y1;
   t->ux.z = z1;

   // and for unit y vector
   o=uy;
   x1 = cos(a->pa) * o->x + sin(a->pa) * o->z;
   y1 = o->y;
   z1 = -sin(a->pa) * o->x + cos(a->pa) * o->z;
   t->uy.x = cos(a->azim) * x1 - sin(a->azim)*y1;
   t->uy.y = sin(a->azim) * x1 + cos(a->azim)*y1;
   t->uy.z = z1;
   return t;
}

static void setup_tiles_frustum( struct ELECTRODE *e, int NT,
     double r1, double r2,
     struct POINT *c,
     double h, struct AXIS *axis, int uniform)
{
   double sh = sqrt((r2 - r1) * (r2 - r1) + h * h);  // Slant height
   double slant = atan2( h, r2 - r1);  // 0 for disc, pi/2 for cylinder

   // Special case to handle circular loops as a single circular filament
   int ring = FALSE;
   if( e->coil && e->coil->loop)
   {
      ring = TRUE;
      NT = NG_RING * tile_scale;
      sh = r1;
      logmsg( 3, "using ring for %s", e->name);
      logmsg( 3, "r2 %.3f r1 %.3f h %.3f", r2, r1, h);
   }

   // correction of the number of tiles in case only one end of the frustum
   // has zero radius (and therefore no need for extra narrow tiles)
   if (!uniform && (r1==0 || r2==0)) NT/=1.5;

   double rmid = (r1 + r2)/2;
   double total_area = 2 * M_PI * rmid * sh;
   double tile_area = total_area / NT;
   double nominal_step = sqrt( tile_area);

   int nt1 = sh / nominal_step;    // Number of rings

   // nt1 can end up zero when length/diameter is very small.  Set a minimum
   // of 3 rings of tiles and rework the nominal_step.
   if( nt1 < 3)
   {
      nt1 = 3;
      nominal_step = sh/nt1;
   }

   logmsg( 2, "frustum nt1 %d sh %.3e nom step %.3e", nt1, sh, nominal_step);

   // prepare the q1[] array with values from 0 to 1 describing the
   // coordinates of the borders of the tiles along the cylinders axis

   double q1[nt1+1];
   if (r1==0) {
      for( int i = 0; i <= nt1 ; i++)
      {
         double f = (double)i / nt1;
         if (!uniform) f *= f * f;
         q1[i] = 1 - f;
      }
   } else if (r2==0) {
      for( int i = 0; i <= nt1 ; i++)
      {
         double f = (double)(nt1-i) / nt1;
         if (!uniform) f *= f * f;
         q1[i] = f;
      }
   } else {
      for( int i = 0; i < nt1/2 + 1; i++)
      {
         double f = i / (0.5*nt1);
         if (!uniform) f *= f * f;
         q1[i] = 1 - 0.5*f;
         q1[nt1-i] = 0.5*f;
      }
   }

   // Estimate number of tiles
   int nte = 0;
   for( int i = 0; i < nt1; i++)
   {
      double t = (q1[i]+q1[i+1])/2;
      double r = r1 + (r2 - r1) * t;
      int na = 2*M_PI*r/nominal_step;
      if( na < 2) na = 2;
      nte += na;
   }

   // Factor to adjust number of tiles closer to requested number
   double na_factor = NT/(double) nte;
   
   logmsg( 2, "frustum nte %d na_factor %.4f", nte, na_factor);

   int nt = 0;   // To count actual number of tiles

   for( int i = 0; i < nt1; i++)
   {
      double t = (q1[i]+q1[i+1])/2;
      double s = sh * t;

      double sh_step = (q1[i]-q1[i+1])*sh;   // step size along the frustum's slant height

      double r = r1 + (r2 - r1) * t;
      double z = s * sin(slant);

      int na = 2*M_PI*r/nominal_step * na_factor;
      if( na < 2) na = 2;
      double ca = 2*M_PI/na;
      double ring_step = ca*r;     // step size along this ring's circumference
 
      // Self potentials of all the tiles in this ring are the same. If h is
      // zero then we have a flat disc and rectangular tiles, otherwise
      // calculate for a curved tile.

      PMTYPE spc = self_potential_rectangle( sh_step, ring_step);

      double p = 0;
      for( int i = 0; i < na; i++, p += ca)
      {
         struct POINT o = { r * cos(p), r * sin(p), z };
         struct POINT ux = { sin(p), -cos(p), 0 };
         struct POINT uy = { cos(p)*cos(slant), sin(p)*cos(slant), sin(slant) };
         struct TILE *t = create_oriented_tile_rect( e, &o, c, axis, spc,
                                                     &ux, &uy);
         t->sx = ring_step/2;
         t->sy = sh_step/2;
         t->area = t->sx*t->sy*4;

         nt++;
      }

      if( ring) break;
   }

   logmsg( 1, "tiles for %s (frustum): %d", e->name, nt);
   if( !nt) bailout( "zero tiles for %s", e->name);
}

//
//  Cylinders and discs are just special cases of a right circular frustum.
//

static void setup_tiles_disc( struct ELECTRODE *e, struct DISC *d)
{
   setup_tiles_frustum( e, d->nt, d->radius1, d->radius2,
                           d->center, 0, d->axis, d->uniform);
}

static void setup_tiles_cyl( struct ELECTRODE *e, struct CYL *c)
{
   setup_tiles_frustum( e, c->nt,
          c->radius1, c->radius2,
          c->end1, c->length, c->axis, c->uniform);
}

static void setup_tiles_rect( struct ELECTRODE *e, struct RECT *d)
{
   // Perpendicular distance from center to each edge
   double r1 = pythagoras( *d->center, *d->edge1);
   double r2 = pythagoras( *d->center, *d->edge2);

   // Compute unit vectors spanning the rectangle
   struct POINT ux = sub_3D( *d->center, *d->edge1);
   struct POINT uy = sub_3D( *d->center, *d->edge2);

   norm_3D( &ux);
   norm_3D( &uy);

   //
   //  Try to cover the rectangle with approximately square tiles.  Total
   //  number of tiles may be slightly less than specified by d->nt.
   //

   double total_area = r1 * r2 *4;
   double nta = sqrt(d->nt/total_area);
   int nt1 = nta*r1*2;
   int nt2 = nta*r2*2;

   //
   //  Compute the actual corners of the tiles in a way which makes them
   //  narrower proportional to the square of the distance to the edge.
   //  (i.e., the corner points are given by a third-power function)
   //

   double q1[nt1+1];
   double q2[nt2+1];

   for( int i = 0; i < nt1/2 + 1; i++)
   {
      double f = i / (0.5*nt1);
      if (!d->uniform) f *= f * f;
      q1[i] = 1 - f;
      q1[nt1-i] = -1 + f;
   }

   for( int i = 0; i < nt2/2 + 1; i++)
   {
      double f = i / (0.5*nt2);
      if (!d->uniform) f *= f * f;
      q2[i] = 1 - f;
      q2[nt2-i] = -1 + f;
   }

   double xd1 = d->edge1->x - d->center->x;
   double yd1 = d->edge1->y - d->center->y;
   double zd1 = d->edge1->z - d->center->z;

   double xd2 = d->edge2->x - d->center->x;
   double yd2 = d->edge2->y - d->center->y;
   double zd2 = d->edge2->z - d->center->z;

   for( int i = 0; i < nt1; i++)
      for( int j = 0; j < nt2; j++)
      {
         double a = (q1[i] + q1[i+1])/2;
         double b = (q2[j] + q2[j+1])/2;

         struct POINT p = { d->center->x + a*xd1 + b*xd2,
                            d->center->y + a*yd1 + b*yd2,
                            d->center->z + a*zd1 + b*zd2 };

         PMTYPE spc =
            self_potential_rectangle( r1*(q1[i]-q1[i+1]), r2*(q2[j]-q2[j+1]));

         struct TILE *t = create_tile( e, p, spc);
         t->area= total_area * (q1[i]-q1[i+1]) * (q2[j]-q2[j+1]) / 4;

         t->ux = ux;
         t->uy = uy;
         t->sx = r1 * (q1[i+1]-q1[i])/2;
         t->sy = r2 * (q2[j+1]-q2[j])/2;
      }
}

static void setup_tiles_toroid( struct ELECTRODE *e, struct TOROID *t)
{
   double cr = (t->outer_radius + t->inner_radius)/2;  // Centerline radius
   double tr = (t->outer_radius - t->inner_radius)/2;  // Tube radius

   double total_area = 2 * M_PI * tr * 2 * M_PI * cr;
   double da = sqrt(total_area/t->nt);    // Tile size

   double hp = da/tr;                 // Angular step around tube circumference
   int nrings = 2 * M_PI / hp;

   logmsg( 1, "toroid nrings %d", nrings);

   int i, nt = 0;
   for( i = 0; i < nrings; i++)   // nrings around the tube circumference
   {
      double theta = i * hp;
      double r = cr + tr * cos(theta);
      double z = tr * sin(theta);

      int ntr = 2 * M_PI * r/da;
      if( ntr == 0) ntr = 1;

      // Todo: proper self potential of toroid tile.
      PMTYPE spc = self_potential_rectangle( 2 * M_PI * r/ntr, da);

      int j;
      double xj = i % 2 /2.0;
      for( j = 0; j < ntr; j++)
      {
         double phi = (j+xj) * 2 * M_PI/ntr;
         struct POINT o = { r * cos(phi), r * sin(phi), z };

         create_oriented_tile( e, &o, t->center, t->axis, spc);
         nt++;
      }
   }

   logmsg( 1, "tiles for %s (toroid): %d", e->name, nt);
}

static void setup_tiles_wire( struct ELECTRODE *e, struct WIRE *w)
{
   // Todo: increase tile density near ends
   PMTYPE spc = self_potential_wire( w->length/w->nt, w->wirad);

   double dx = (w->end2->x - w->end1->x)/w->nt;
   double dy = (w->end2->y - w->end1->y)/w->nt;
   double dz = (w->end2->z - w->end1->z)/w->nt;

   for( int i = 0; i < w->nt; i++)
   {
      struct POINT p = { w->end1->x + i * dx,
                         w->end1->y + i * dy,
                         w->end1->z + i * dz };

      create_tile( e, p, spc);
   }

   logmsg( 1, "tiles for %s (wire): %d", e->name, w->nt);
}

static void setup_tiles_ring( struct ELECTRODE *e, struct RING *r)
{
   double arclen = 2 * M_PI * r->radius / r->nt;

   // Each tile is approximately a straight wire
   PMTYPE spc = self_potential_wire( arclen, r->wirad);
  
   for( int i = 0; i < r->nt; i++)
   {
      double p = i * 2 * M_PI/r->nt;
      struct POINT o = { r->radius * cos(p), r->radius * sin(p), 0 };

      o = rotate( o, r->axis->pa, r->axis->azim);
      sum_3D( &o, *r->center);
      create_tile( e, o, spc);
   } 
}

static void setup_tiles_plate( struct ELECTRODE *e, struct PLATE *p)
{
   struct POINT *list = array( p->npoints, struct POINT);
   for( int i = 0; i < p->npoints; i++) list[i] = *p->points[i];

   struct AXIS axis = polygon_axis( list, p->npoints);

   double area = polygon_area( &axis, p->points, p->npoints);
   double step = sqrt( area/p->nt);

   // Todo: implement non-uniform step near edges

   //
   //  Rotate polygon to become parallel to xy plane - this make it simple to
   //  calculate new intersections of edges as the polygon shrinks.
   //

   for( int i = 0; i < p->npoints; i++)
      list[i] = unrotate( list[i], axis.pa, axis.azim);

   //
   //  Array of unit vectors v[] indicating which direction to move each edge
   //  as the polygon shrinks.
   //
   
   struct POINT *v = array( p->npoints, struct POINT);
   for( int i1 = 0; i1 < p->npoints; i1++)
   {
      int i2 = (i1 + 1) % p->npoints;

      struct POINT d1 = sub_3D( list[i2], list[i1]);

      v[i1] = xprod_3D( (struct POINT){0,0,1}, d1);
      norm_3D( v + i1);
   }

   double spc = self_potential_rectangle( step, step);

   int tnp = 0;    // Total number of tiles allocated

   int NP = p->npoints;
   struct POINT *tlist = array( NP, struct POINT);
   struct POINT *diffs = array( NP, struct POINT);

   for( int i = 0; i < NP; i++)
   {
      int j = (i + 1) % NP;
      diffs[i] = sub_3D( list[j], list[i]);
   }

   // Function to shrink the polygon boundary
   void shrink( double step)
   {
      for( int i1 = 0; i1 < NP; i1++)
      {
         int i2 = (i1 + 1) % NP;
         int i3 = (i1 - 1 + NP) % NP;
         int i4 = i1;

         struct POINT p1 = list[i1];   // The point about to be moved
         struct POINT p2 = list[i2];   // Next point along
         struct POINT p3 = list[i3];   // Previous point
         struct POINT p4 = list[i4];   // A copy of p1

         struct POINT vt;

         // Move line p1,p2 inwards by v[i1] * step
         vt = v[i1];
         scale_3D( &vt, step);
         sum_3D( &p1, vt);
         sum_3D( &p2, vt);

         // Move line p3,p4 inwards by v[i3] * step
         vt = v[i3];
         scale_3D( &vt, step);
         sum_3D( &p3, vt);
         sum_3D( &p4, vt);

         double d1 = (p1.x - p2.x)*(p3.y - p4.y) - (p1.y - p2.y)*(p3.x - p4.x);
         double d2 = (p1.x - p2.x)*(p3.y - p4.y) - (p1.y - p2.y)*(p3.x - p4.x);

         if( fabs(d1) < 1e-6 || fabs(d2) < 1e-6) { NP = 0; break; }

         // Calculate new point of intersection between lines p1,p2 and p3,p4
         tlist[i1] = (struct POINT) {
               ((p1.x*p2.y - p1.y*p2.x)*(p3.x - p4.x) -
                (p1.x - p2.x)*(p3.x*p4.y - p3.y*p4.x)) / d1,
               ((p1.x*p2.y - p1.y*p2.x) * (p3.y - p4.y) -
                (p1.y - p2.y) * (p3.x*p4.y - p3.y*p4.x)) / d2,
               (p1.z + p2.z + p3.z + p4.z)/4};
      }

      for( int i = 0; i < NP; i++) list[i] = tlist[i];
   }

   // Initial shrink before generating any tiles.  This to prevent adjacent
   // polygons from having very short or zero tile distances.

   shrink( step/4);

   while( NP >= 2)   // While at least one edge remaining
   {
      //  Create a polygonal ring of tiles along the perimeter

      int np = 0;           // Number of tiles in this ring
      for( int i = 0; i < NP; i++)
      {
         int j = (i + 1) % NP;
         struct POINT d1 = sub_3D( list[j], list[i]);
         double m = mag_3D(d1);
         norm_3D( &d1);
         int n = m / step;
         scale_3D( &d1, m / n);
  
         struct POINT t = list[i];
         for( int k = 0; k < n; k++)
         {
            create_oriented_tile( e, &t, &origin, &axis, spc);
            sum_3D( &t, d1);
         }
         np += n;
      }

      if( np == 0) break;   // Nothing left to do

      tnp += np;  // Accumulate total number of tiles

      shrink( step);

      // Remove any edges that have shrunk to nothing or reversed direction

      for( int i = 0; i < NP; i++)
      {
         int j = (i + 1) % NP;
         struct POINT d1 = sub_3D( list[j], list[i]);

         if( inner_prod_3D( d1, diffs[i]) <= 0)
         {
            for( int k = i; k < NP-1; k++)
            {
               list[k] = list[k+1];
               diffs[k] = diffs[k+1];
               v[k] = v[k+1];
            }
            i--;
            NP--;
         }
      }
   }

   logmsg( 1, "tiles for %s (plate): %d", e->name, tnp);
}


//
//  Set up TILE structures to represent the world of electrodes and coils
//

static void setup_tiles( void)
{
   for( struct ELECTRODE *e = WORLD.electrodes; e; e = e->next)
      for( struct COMPONENT *c = e->components; c; c = c->next)
         switch( c->type)
         {
            case CT_SPHERE: setup_tiles_sphere( e, &c->u.sphere);
                            break;
            case CT_DISC:   setup_tiles_disc( e, &c->u.disc);
                            break;
            case CT_CYL:    setup_tiles_cyl( e, &c->u.cyl);
                            break;
            case CT_TOROID: setup_tiles_toroid( e, &c->u.toroid);
                            break;
            case CT_RECT:   setup_tiles_rect( e, &c->u.rect);
                            break;
            case CT_WIRE:   setup_tiles_wire( e, &c->u.wire);
                            break;
            case CT_PLATE:  setup_tiles_plate( e, &c->u.plate);
                            break;
            case CT_RING:   setup_tiles_ring( e, &c->u.ring);
                            break;
         }

   logmsg( 1, "total number of tiles: %d", ntiles);
}

//
//  Called if a ground plane has been specified.  Calculates the mirror image
//  point tile->ci of the center tile->c of each tile.
//

static void calculate_image_points( void)
{
   for( int i = 0; i < ntiles; i++)
   {
      struct TILE *t = tile_list[i];

      if( !image_point( &t->c, &t->ci))
      {
         // The tile is on the wrong side of the ground plane. All components
         // should be on the side specified by the axis vector of the ground
         // plane.

         bailout( "component beneath ground, check %s", t->electrode->name);
      }

      // Adjust the self potential to allow for the tile's own reflection
      t->spc -= 1/( 4 * M_PI * EPSILON * tile_mirror_distance( t, t));
   }
}

//
//  Reset all the electrodes to zero potential.
//

static void reset_all_electrodes( void)
{
   for( struct ELECTRODE *e = WORLD.electrodes; e; e = e->next)
      e->volts = e->charge = 0;
}

static PMTYPE **PP = 0;    // Potential matrix, PP[ntiles][ntiles]
static QVTYPE *V = 0;      // Voltage vector, V[ntiles]
static QVTYPE *Q = 0;      // Charge vector, Q[ntiles]

//
//  Calculate coefficients of the potential matrix.  In all these functions,
//  TILE *d is the destination tile, *s is the source tile.
//

static PMTYPE potential_coef_twoway( struct TILE *d, struct TILE *s)
{
   double p = 1/tile_distance( d, s);

   if( WORLD.g) p -= 1/tile_mirror_distance( d, s);

   p /= 4 * M_PI * EPSILON;

   if( isnan( p) || p <= 0)
   {
      logmsg( 0, "bad potential coefficient %.3e, %s to %s",
                        p, s->electrode->name, d->electrode->name);
      logmsg( 0, "source %.4f %.4f %.4f mirror %.4f %.4f %.4f",
                        s->c.x, s->c.y, s->c.z, s->ci.x, s->ci.y, s->ci.z);
      logmsg( 0, "dest %.4f %.4f %.4f", d->c.x, d->c.y, d->c.z);
      bailout( NULL);
   }
   return p;
}

static PMTYPE potential_coef_oneway( struct TILE *d, struct TILE *s)
{
   // from Mathematica:
   //  
   //  In[2]:= Integrate[(x*x+y*y+z*z)^-(1/2), x, y]                                   
   //  
   //                        x                     x y                              2    2    2                     2    2    2
   //  Out[2]= -x + z ArcTan[-] - z ArcTan[--------------------] +  y Log[x + Sqrt[x  + y  + z ]] + x Log[y + Sqrt[x  + y  + z ]]
   //                        z                     2    2    2
   //                                      z Sqrt[x  + y  + z ]

   //
   // Compute vector from source to destination, and project it onto the tile's
   // plane. This gives the source tile's center point in the coordinates of
   // the above formula.
   //

   struct POINT v = sub_3D( d->c, s->c);
   double l = inner_prod_3D( v, v);

   // If the tiles are far apart compared to the source tile's size, then
   // don't bother with the somewhat time-consuming precise calculation

   if(l > 2 * (fabs(s->sx) + fabs(s->sy))) return potential_coef_twoway( d, s);

   double xc = inner_prod_3D( v, s->ux);
   double yc = inner_prod_3D( v, s->uy);
   double zc = l - xc*xc -yc*yc;

   if( zc < 0) zc = 0; // zc<0 may happen due to numerical inaccuracies
   else zc = sqrt(zc); 

   if (xc<0) xc=-xc;   // doesn't matter for the outcome, but avoids numerical problems in log(x+l) and log(y+l)
   if (yc<0) yc=-yc;

   double p = 0;

   for( int i = -1; i <= 1; i += 2)
      for( int j = -1 ; j <= 1;j += 2)
      {
         // For each of the source tile's 4 corner points, evaluate the above
         // indefinite integral, and add and subtract as needed to compute the
         // definite integral

         double x = xc + i*s->sx;
         double y = yc + j*s->sy;
         double z = zc;
         double l = sqrt(x*x + y*y + z*z);
         double p1 = -x + y * log(x+l) + x * log(y+l);
         if (z>0) p1+= z * atan(x/z) - z * atan(x*y/(z*l));

         p += p1 * i * j;
      } 

   p /= 4 * M_PI * EPSILON * s->area;

   if( WORLD.g) p -= 1/( 4 * M_PI * EPSILON * tile_mirror_distance( d, s));

   if( isnan( p)) bailout( "mutual potential coefficient s nan %s %s",
                            d->electrode->name, s->electrode->name);
   if( p <= 0) bailout( "bad potential coefficient %.3e %s %s",
                        p, d->electrode->name, s->electrode->name);

   return p;
}

static PMTYPE potential_coef( struct TILE *d, struct TILE *s)
{
   return s->sx ? potential_coef_oneway( d, s)
                : potential_coef_twoway( d, s);
}

static void prepare_potential_matrix_slave( int *i1, int *i2)
{
   for( int i = *i1; i < *i2; i++)     // For each collocation...
   {
      PP[i][i] = tile_list[i]->spc;

      for( int j = 0; j < ntiles; j++)  // For each source tile...
         if( i != j) PP[i][j] = potential_coef( tile_list[i], tile_list[j]);
   }
}

static void prepare_potential_matrix( void)
{
   int r = ntiles / ncpu;
   int n1[MAXCPU], n2[MAXCPU];

   for( int n = 0; n < ncpu; n++)
   {
      n1[n] = n * r;
      n2[n] = n == ncpu - 1 ? ntiles : n * r + r;
     
      worker_run( prepare_potential_matrix_slave,
                  n1 + n, n2 + n, NULL, NULL, NULL);
   }

   worker_rendezvous();
}

///////////////////////////////////////////////////////////////////////////////
//  Numeric Utilities for FMM                                                //
///////////////////////////////////////////////////////////////////////////////

//
//  Compute associated Legendre polynomial.   From Numerical Recipies.
//

static double Plm( int l, int m, double x)
{
   if( m < 0) bailout( "-ve m in Plm");
   if( m > l) bailout( "m > l in Plm");
   if( fabs(x) > 1.0) bailout( "fabs(x) in Plm");

   double pll = 0, pmm = 1.0;

   if(m > 0)
   {
      double somx2 = sqrt((1.0-x) * (1.0+x));
      double fact = 1.0;
      for( int i = 1; i <= m; i++)
      {
         pmm *= -fact * somx2;
         fact += 2.0;
      }
   }

   if( l == m) return pmm;

   double pmmp1 = x * (2*m + 1) * pmm;
   if( l == m+1) return pmmp1;

   for( int ll = m+2; ll <= l; ll++)
   {
      pll = (x * (2*ll-1) * pmmp1 - (ll +m - 1) * pmm)/(ll - m);
      pmm = pmmp1;
      pmmp1 = pll;
   }

   return pll;
}

//
//  Factorial of an integer.
//

static double fact( int n)
{
   double f = 1;
   for( int i = 1; i <= n; i++) f *= i;
   return f;
}

//
//  'i' raised to power of modulus of an integer.
//

static CXTYPE ipow( int n)
{
   if( n < 0) n = -n;

   switch( n % 4)
   {
      case 0: return 1;
      case 1: return I;
      case 2: return -1;
      default: return -I;
   }
}

//
//  -1 raised to integer power
//

static inline int npow( int n)
{
   return n % 2 ? -1 : 1;
}

//
//  Spherical harmonics.  Schmidt semi-normalized (Racah's normalization).
//  But return zero if |m| > n to eliminate some 'if' statements later.
//

static CXTYPE Ynm( int n, int m, double theta, double phi)
{
   int modm = m < 0 ? -m : m;
   if( n < modm) return 0;

   return sqrt( fact(n - modm)/(fact(n + modm))) * 
          Plm( n, modm, cos(theta)) * cexp( I * m * phi);
}

//
//  Constant Anm.
//  Ihler section 7.1.1 Eq 33.
//
 
static double **Anm_cache;

static void Anm_setup( void)
{
   Anm_cache = array( 2 * MXP + 1, double *);

   for( int n = 0; n <= 2 * MXP; n++)
   {
      Anm_cache[n] = array( 4 * MXP + 1, double);

      for( int m = -2 * MXP; m <= 2 * MXP; m++)
         Anm_cache[n][m + 2 * MXP] = npow(n)/sqrt(fact(n-m) * fact(n+m));
   }   
}

static inline double Anm( int n, int m)
{
   return Anm_cache[n][m + 2 * MXP];
}

///////////////////////////////////////////////////////////////////////////////
//  FMM Coefficient Cache                                                    //
///////////////////////////////////////////////////////////////////////////////

//
//  Thanks to the discrete translation symmetry of the octree at each level,
//  many of the coefficients are reusable.  This bit of code maintains a cache
//  of coefficients in a binary tree indexed by the coefficient type and the
//  relative polar coordinates of the transformation vector.
//

static int cachecnt[4];
static int cachehit[4];

static struct CNODE {
   int class;
   double key1, key2, key3;
   void *val;

   pthread_mutex_t lock;
   struct CNODE *left, *right;
}
 *ccache = NULL;

static int nccache = 0;

static pthread_mutex_t cache_lock = PTHREAD_MUTEX_INITIALIZER;

//
//  Returns with the cache entry locked.  Caller must unlock when done.
//

static struct CNODE *ccache_find( int class,
                                  double key1, double key2, double key3)
{
   pthread_mutex_lock( &cache_lock);

   struct CNODE **pp = &ccache;

   while( *pp)
   {
      int q = 0;
      if( class > (*pp)->class) q = 1;
      else
      if( class < (*pp)->class) q = -1;
      else
      if( key1 > (*pp)->key1) q = 1;
      else
      if( key1 < (*pp)->key1) q = -1;
      else
      if( key2 > (*pp)->key2) q = 1;
      else
      if( key2 < (*pp)->key2) q = -1;
      else
      if( key3 > (*pp)->key3) q = 1;
      else
      if( key3 < (*pp)->key3) q = -1;
      else
      {
         cachehit[class]++;
         pthread_mutex_unlock( &cache_lock);
         pthread_mutex_lock( &(*pp)->lock);
       
         return *pp;
      }

      pp = q > 0 ? &(*pp)->right : &(*pp)->left;
   }

   nccache++;
   cachecnt[class]++;
   *pp = allocate( struct CNODE);

   (*pp)->left = (*pp)->right = NULL;
   (*pp)->class = class;
   (*pp)->key1 = key1;
   (*pp)->key2 = key2;
   (*pp)->key3 = key3;

   pthread_mutex_init( &(*pp)->lock, NULL);
   pthread_mutex_lock( &(*pp)->lock);

   pthread_mutex_unlock( &cache_lock);
   return *pp;
}

static inline void ccache_release( struct CNODE *p)
{
   pthread_mutex_unlock( &p->lock);
}

///////////////////////////////////////////////////////////////////////////////
//  FMM Operators                                                            //
///////////////////////////////////////////////////////////////////////////////

//
//  References to Ihler refer to
//     Alexander T. Ihler,
//     An Overview of Fast Multipole Methods, 2004
//     http://www.ics.uci.edu/~ihler/papers/ihler_area.pdf
//
//  References to Greengard refer to
//     Greengard and Beatson
//     A short course on fast multipole methods
//     http://math.nyu.edu/faculty/greengar/shortcourse_fmm.pdf
//

// Todo: repack the fields into linear arrays and transforms into square arrays
// Todo: change to real coefficients

static CXTYPE **alloc_fmm_field( void)
{
   CXTYPE **p = array( MXP, CXTYPE *);
   for( int n = 0; n < MXP; n++) p[n] = n + array( 1 + 2*n, CXTYPE);

   return p;
}

static void zero_fmm_field( CXTYPE **p)
{
   for( int n = 0; n < MXP; n++)
      for( int m = -n; m <= n; m++) p[n][m] = 0;
}

//
//  Multipole to local expansion conversion coefficients.
//  Ihler section 7.1.2 Eq 36.
//  Greengard theorem 5.4 Eq 5.26
//

#define CNODE_M2L 1

static CXTYPE ****coef_m2lc( struct POINT *ps, struct POINT *pt)
{
   // Polar coordinates of ps with respect to pt
   double alpha, beta, r;
   to_polar( ps, pt, &alpha, &beta, &r);

   struct CNODE *cn = ccache_find( CNODE_M2L, alpha, beta, r);
   if( cn->val)
   {
      ccache_release( cn);
      return cn->val;
   }

   CXTYPE ****p = array( MXP, CXTYPE ***);

   for( int j = 0; j < MXP; j++)
   {
      p[j] = j + (CXTYPE ***) array( 1 + 2 * j, CXTYPE **);

      for( int k = -j; k <= j; k++)
      {
         p[j][k] = array( MXP, CXTYPE *);

         for( int n = 0; n < MXP; n++)
         {
            p[j][k][n] = n + array( 1 + 2 * n, CXTYPE);

            for( int m = -n; m <= n; m++)
               p[j][k][n][m] =
                     ipow(k-m)/ipow(k)/ipow(m) *
                     Anm(n,m) * Anm(j,k)/Anm(j+n,m-k) *
                     1/pow(r,j+n+1) *
                     npow(n) *
                     Ynm( j+n, m-k, alpha, beta);
         }
      }
   }

   cn->val = p;
   ccache_release( cn);
   return p;
}

//
//  Local field translation coefficients.
//  Ihler section 7.1.3 Eq 41.
//  Greengard theorem 5.5 Eq 5.30.
//

#define CNODE_LTC 2

static CXTYPE ****coef_ltc( struct POINT *ps, struct POINT *pt)
{
   double alpha, beta, r;
   to_polar( pt, ps, &alpha, &beta, &r);  // Note: need -z so use pt - ps

   struct CNODE *cn = ccache_find( CNODE_LTC, alpha, beta, r);
   if( cn->val)
   {
      ccache_release( cn);
      return cn->val;
   }

   CXTYPE ****p = array( MXP, CXTYPE ***);

   for( int j = 0; j < MXP; j++)
   {
      p[j] = j + array( 1 + 2*j, CXTYPE **);

      for( int k = -j; k <= j; k++)
      {
         p[j][k] = array( MXP, CXTYPE *);

         for( int n = 0; n < MXP; n++)
         {
            p[j][k][n] = n + array( 1 + 2*n, CXTYPE);
            if( n >= j)
               for( int m = -n; m <= n; m++)
                  p[j][k][n][m] = 
                     ipow(m)/ipow(m-k)/ipow(k) * 
                     Anm(n-j,m-k) * Anm(j,k)/Anm(n,m) *
                     1/npow(n+j) *
                     pow( r, n-j) *
                     Ynm(n-j, m-k, alpha, beta);
         }
      }
   }

   cn->val = p;
   ccache_release( cn);
   return p;
}
  
//
//  Local expansion coefficients.  No point in caching these.
//  Ihler section 7.1.3 Eq 39.
//

static CXTYPE **coef_lec( struct POINT *ps, struct POINT *pt)
{
   double alpha, beta, r;
   to_polar( ps, pt, &alpha, &beta, &r);

   CXTYPE **p = array( MXP, CXTYPE *);

   for( int j = 0; j < MXP; j++) p[j] = j + array( 1 + 2*j, CXTYPE);

   for( int j = 0; j < MXP; j++)
      for( int k = -j; k <= j; k++) p[j][k] = pow(r, j) *
                                                Ynm( j, k, alpha, beta) * 
                                                1/(EPSILON * 4 * M_PI);

   return p;
}

     
//
//  Multipole expansion coefficients.  No point in caching these.
//  Ihler section 4.3 Eq 20.
//

static CXTYPE **coef_mec( struct POINT *ps, struct POINT *pt)
{
   double alpha, beta, r;
   to_polar( ps, pt, &alpha, &beta, &r);
 
   CXTYPE **p = array( MXP, CXTYPE *);

   for( int l = 0; l < MXP; l++)
   {
      p[l] = l + array( 1 + 2*l, CXTYPE);

      for( int m = -l; m <= l; m++)
         p[l][m] = pow( r, l) * Ynm( l, -m, alpha, beta);
   }

   return p;
}

//
//  Multipole translation coefficients
//  Ihler section 7.1.1 Eq 32.
//  Greengard theorem 5.3 Eq 5.22.
//
//  But with matrix components rearranged so that a generic matrix multiply
//  can be used.
//

#define CNODE_MTC 3

static CXTYPE ****coef_mtc( struct POINT *ps, struct POINT *pt)
{
   double alpha, beta, r;
   to_polar( ps, pt, &alpha, &beta, &r);

   struct CNODE *cn = ccache_find( CNODE_MTC, alpha, beta, r);
   if( cn->val)
   {
      ccache_release( cn);
      return cn->val;
   }

   CXTYPE ****p = array( MXP, CXTYPE ***);

   for( int j = 0; j < MXP; j++)
   {
      p[j] = j + array( 1 + 2*j, CXTYPE **);

      for( int k = -j; k <= j; k++)
      {
         p[j][k] = array( MXP, CXTYPE *);

         for( int s = 0; s < MXP; s++)
         {
            p[j][k][s] = s + array( 1 + 2*s, CXTYPE);
            int n = j - s;

            if( n >= 0 && n < MXP)
               for( int t = -s; t <= s; t++)
               {
                  int m = k - t;
                  if( m >= -n && m <= n)
                     p[j][k][s][t] = 
                           ipow(k)/ipow(m)/ipow(k-m) *
                           Anm(n,m) * Anm(j-n,k-m)/Anm(j,k) *
                           pow(r,n) *
                           Ynm(n, -m, alpha, beta);
               }
         }
      }
   }

   cn->val = p;
   ccache_release( cn);
   return p;
}

//
//  Transform the spherical harmonic coefficients of a field using one of the
//  matrices calculated above.
//

static void apply_fmm_matrix( CXTYPE **field2, CXTYPE ****mat, CXTYPE **field1)
{
   for( int j = 0; j < MXP; j++)
      for( int k = -j; k <= j; k++)
         for( int s = 0; s < MXP; s++)
            for( int t = -s; t <= s; t++)
               field2[j][k] += field1[s][t] * mat[j][k][s][t];
}

///////////////////////////////////////////////////////////////////////////////
//  Adaptive Octree                                                          //
///////////////////////////////////////////////////////////////////////////////

//
//  A conventional octree structure.  If using adaptive mode, leaf boxes can
//  be placed at any depth necessary to honour the MAXLEAF value.
//
//  Note: Adaptive is disabled until I figure out how to make it work with
//  the preconditioning algorithm.
//

static int FIXDEPTH = 5;  // 0 for adaptive, non-zero for fixed depth and value
                          // sets the depth of leaf boxes.

#define MAXDEPTH 50

#define MAXLEAF  10   // Max number of tiles per leaf box. If depth is fixed
                      // then MAXLEAF is ignored.

static int autodepth = FALSE;  // If TRUE then choose a fixed depth to aim for
                               // around MAXLEAF tiles per leaf box.
                               // Set by -m depth=auto option.

static int *xlt;  // For debug only

// Todo: translate image field from mbox->mfd with suitable transform, instead
// of building the image field independently with mbox.meci and mbox.mtci

static struct MBOX {

   int id;           // Only used for debug

   int leaf;         // TRUE if this is a leaf box
   int depth;        // Zero for root

   struct POINT c;   // Coordinates of center
   struct POINT ci;  // Ground image center

   double size;      // Half width of box: side length is 2 * size

   struct MBOX *kids[8];
   struct MBOX *parent;

   // List of tiles if this is a leaf box
   struct TILE **tiles;
   int ntiles;

   // Neighbour boxes
   struct MBOX **neibs;
   int nneibs;

   // Interaction list
   struct MBOX **ilist;
   int nilist;

   struct TILE **ptiles;  // List of tiles in this box and its neighbours
   int np;                // Number of tiles in ptiles[]

   PMTYPE **PP;  // PP[ntiles][np].  Potential matrix
   PMTYPE **C;   // C[ntiles][np].  Preconditioning matrix

   // Multipole expansion coefficients for leaf boxes
   CXTYPE ***mecd; // mecd[tile][n][m] direct sources
   CXTYPE ***meci; // meci[tile][n][m] image sources

   // Multipole translation coefficients for branch boxes
   CXTYPE *****mtcd; // mtcd[kid 0..7][n][m][j][k]a direct sources
   CXTYPE *****mtci; // mtci[kid 0..7][n][m][j][k] image sources
  
   // Outgoing field components referenced to center of this box
   CXTYPE **mfd;   // mfd[l][m] field direct from source tiles
   CXTYPE **mfi;   // mfi[l][m] field from image tiles

   // Multipole to local conversion coefficients, one for each entry in the
   // interaction list
   CXTYPE *****m2lcd;  // m2lcd[nilist][n][m][j][k]
   CXTYPE *****m2lci;  // m2lci[nilist][n][m][j][k]

   // Local expansion coefficients for leaf boxes
   CXTYPE ***lec; // lec[tile][n][m]

   // Local translation coefficients for branch boxes
   CXTYPE ****ltc; // ltc[n][m][j][k]

   // Incoming field components referenced to center of this box
   CXTYPE **lf;   // lf[n][m]

   // Variables for sanity checks
   int nmf;   // Number of tiles contributing to this box's multipole field
   int nlf;   // Number of tiles contributing to this box's local field
}
 *root_mbox = NULL;

static int nmbox = 0;                                 // Total number of mboxes
static int nleafs = 0;                           // Total number of leaf mboxes
static int maxdepth = 0;                                  // Deepest level used

static struct MBOX **depthlist[MAXDEPTH];         // List of mbox at each depth
static int depthcnt[MAXDEPTH];                  // Number of mbox at each depth

struct MBOX *alloc_mbox( struct POINT *center, struct MBOX *parent,
                         double size)
{
   struct MBOX *p = allocate( struct MBOX);

   p->c = *center;
   if( WORLD.g) image_point( center, &p->ci);

   p->size = size;
   p->parent = parent;
   if( parent) p->depth = parent->depth + 1;
   p->leaf = TRUE;

   depthcnt[p->depth]++;

   p->id = nmbox++;
   nleafs++;
   if( p->depth > maxdepth) maxdepth = p->depth;

   return p;
}

static void insert_mbox( struct MBOX *m, struct TILE *t)
{
   struct POINT p = t->c;

   if( !m->leaf)
   {
      int xi = (p.x - m->c.x)/m->size > 0 ? 1 : -1;
      int yi = (p.y - m->c.y)/m->size > 0 ? 1 : -1;
      int zi = (p.z - m->c.z)/m->size > 0 ? 1 : -1;

      int k = (zi >= 0 ? 0 : 4) +
              (yi >= 0 ? 0 : 2) + 
              (xi >= 0 ? 0 : 1);
         
      if( !m->kids[k])
      {
         double tsize = m->size/2;
         struct POINT c = { m->c.x + xi * tsize, 
                            m->c.y + yi * tsize,
                            m->c.z + zi * tsize };
         m->kids[k] = alloc_mbox( &c, m, tsize);
      }
        
      if( m->kids[k]->depth < FIXDEPTH)
      {
         if( m->kids[k]->leaf) nleafs--;
         m->kids[k]->leaf = FALSE;
      }
 
      insert_mbox( m->kids[k], t); 
      return;
   }

   //
   //  m is a leaf box.  Add this tile to the leaf.
   //

   m->tiles = realloc( m->tiles, (m->ntiles+1) * sizeof( struct TILE *));
   m->tiles[m->ntiles++] = t;

   if( !FIXDEPTH)   // Using adaptive tree?
   {
      //
      //  Maybe now this leaf box is too full, if so turn it into a branch box
      //  and push its tiles down a level.
      //

      if( m->ntiles > MAXLEAF)
      {
         m->leaf = FALSE;
         nleafs--;
   
         for( int i = 0; i < m->ntiles; i++) insert_mbox( m, m->tiles[i]);
         free( m->tiles);  m->tiles = NULL; m->ntiles = 0;
      }
   }
}

#define DEPTH_FIRST 1
#define BREADTH_FIRST 2

static void pscan_list( struct MBOX **list, void *N,
                        void fcn( struct MBOX *, void *, void *),
                        void *arg1, void *arg2)
{
   for( intptr_t i = 0; i < (intptr_t) N; i++) fcn( list[i], arg1, arg2);
}

static void pscan_octree_level( int depth, 
                                void fcn( struct MBOX *, void *, void *),
                                void *arg1, void *arg2)
{
   struct MBOX **list = depthlist[depth];
   int N = depthcnt[depth];

   for( int b = 0; b < N; )
   {
      //  This little heuristic gives a roughly equal load to all workers
      int r = N - b;  if( r > depth+1) r = depth+1;

      worker_run( pscan_list, list + b, (void *)(intptr_t) r, fcn, arg1, arg2);
      
      b += r;
   }

   worker_rendezvous();
}

static void pscan_octree( int mode,
                          void fcn( struct MBOX *, void *, void *),
                          void *arg1, void *arg2)
{
   if( mode == DEPTH_FIRST)
      for( int i = maxdepth; i >= 0; i--)
         pscan_octree_level( i, fcn, arg1, arg2);

   if( mode == BREADTH_FIRST)
      for( int i = 0; i <= maxdepth; i++)
         pscan_octree_level( i, fcn, arg1, arg2);
}

//
//  Work out maximum size and centroid of system.
//

static void setup_sizes( void)
{
   for( int i = 0; i < ntiles; i++) sum_3D( &WORLD.rc, tile_list[i]->c);

   scale_3D( &WORLD.rc, 1.0/ntiles);
   logmsg( 1, "centroid %.3e %.3e %.3e", WORLD.rc.x, WORLD.rc.y, WORLD.rc.z);

   for( int i = 0; i < ntiles; i++)
   {
      struct TILE *t = tile_list[i];
      double r;
      r = fabs( t->c.x - WORLD.rc.x); if( r > WORLD.msize) WORLD.msize = r;
      r = fabs( t->c.y - WORLD.rc.y); if( r > WORLD.msize) WORLD.msize = r;
      r = fabs( t->c.z - WORLD.rc.z); if( r > WORLD.msize) WORLD.msize = r;
   }

   logmsg( 1, "WORLD.msize %.3e", WORLD.msize);
}

static void fill_octree( void)
{
   root_mbox = alloc_mbox( &WORLD.rc, NULL, WORLD.msize * 1.001); 
   root_mbox->leaf = FALSE;
   nleafs--;

   //
   //  Install all the tiles in octree.
   //

   for( int i = 0; i < ntiles; i++) insert_mbox( root_mbox, tile_list[i]);

   //
   //  If depth is fixed but automatic, keep deepening the octree until we
   //  have an average of MAXLEAF or less tiles per leaf box.
   //

   void pushdown( struct MBOX *m)
   {
      if( !m->leaf)
      {
         for( int k = 0; k < 8; k++) if( m->kids[k]) pushdown( m->kids[k]);
         return;
      }

      m->leaf = FALSE;
      nleafs--;
   
      for( int i = 0; i < m->ntiles; i++) insert_mbox( m, m->tiles[i]);
      free( m->tiles);  m->tiles = NULL; m->ntiles = 0;
   }

   while( autodepth && FIXDEPTH < MAXDEPTH && ntiles/nleafs > MAXLEAF)
   {
      FIXDEPTH++;
      pushdown( root_mbox);     
   }

   //
   //  Make lists of all boxes at each depth.
   //

   for( int i = 0; i <= maxdepth; i++)
      depthlist[i] = array( depthcnt[i], struct MBOX *);

   int *depthidx = array( maxdepth + 1, int);

   void make_boxlists( struct MBOX *m)
   {
      depthlist[m->depth][depthidx[m->depth]++] = m;
      for( int k = 0; k < 8; k++) if( m->kids[k]) make_boxlists( m->kids[k]);
   }

   make_boxlists( root_mbox);
   free( depthidx);

   logmsg( 1, "octree boxes %d, %s %d, nleafs %d, "
              "tiles/leaf %.1f",
               nmbox, 
               FIXDEPTH ? "fixdepth" : "maxdepth", maxdepth,
               nleafs, ntiles/(double)nleafs);
   logmsg( 1, "spherical harminic degrees %d", MXP);
}

//
//  Build neighbour list and interaction list of each mbox.
//
//  Scan kids of neighbours of parent, any that are within 3.6 times the box
//  size must be immediate neighbours, otherwise they belong in the
//  interaction list.
//
//  All the siblings must also be immediate neighbours.
//

static void setup_lists( struct MBOX *mi,
                         __attribute__ ((unused)) void *arg1,
                         __attribute__ ((unused)) void *arg2)
{
   struct MBOX *mp = mi->parent;
   if( !mp) return;

   //
   //  Reserve space for the maximum possible number of neighbour mboxes
   //  and max size of interaction list.
   //
 
   mi->neibs = array( 26, struct MBOX *);
   mi->ilist = array( 189, struct MBOX *);

   for( int j = 0; j < mp->nneibs; j++) // Loop over parent's neighbour list
   {
      struct MBOX *mpn = mp->neibs[j];
      for( int k = 0; k < 8; k++)
      {
         struct MBOX *mk = mpn->kids[k];
         if( !mk) continue;

         double r = pythagoras( mi->c, mk->c);
   
         if( r > 3.6 * mi->size)
         {
            // Add to our interaction list

            if( mi->nilist == 189) bailout( "ilist more than 189");
            mi->ilist[mi->nilist++] = mk;
         }
         else
         {
            // Add mk to the neighbour list

            if( mi->nneibs == 26) bailout( "more than 26 neighbours");
            mi->neibs[mi->nneibs++] = mk;
         }
      }
   }

   // Add siblings to neighbour list

   for( int k = 0; k < 8; k++)
   {
      struct MBOX *mk = mp->kids[k];

      if( mk && mk != mi)
      {
         if( mi->nneibs == 26) bailout( "more than 26 neighbours");
         mi->neibs[mi->nneibs++] = mk;
      }
   }

   //
   //  Free up the list space we didn't use
   //

   mi->ilist = realloc( mi->ilist, mi->nilist * sizeof( struct MBOX *));
   mi->neibs = realloc( mi->neibs, mi->nneibs * sizeof( struct MBOX *));
}

//
//  Setup a potential matrix mi->PP for the direct Coulomb coupling from tiles
//  in mbox mi and its neighbour list.
//
//  Also set up a preconditioning matrix mi->C if required.
//

static void setup_pp( struct MBOX *mi,
                      __attribute__ ((unused)) void *arg1,
                      __attribute__ ((unused)) void *arg2)
{
   //
   //  Preconditioning, as described in section V of
   //    Fast Capacitance Extract of General Three-Dimensional Structures
   //    Nabors, Kim, White,
   //    IEEE Trans on Microwave Theory and Techniques, 40(7):1496-1506 July 92
   //    http://manual.freeshell.org/fastcap/mtt.ps
   //

   if( mi->leaf)
   {
      //
      //  Form mi->ptiles[], a list of all tiles in this box and all its
      //  neighbours.
      //

      mi->np = mi->ntiles;
      for( int k = 0; k < mi->nneibs; k++) mi->np += mi->neibs[k]->ntiles;
      mi->ptiles = array( mi->np, struct TILE *);

      int n = 0;
      for( int k = 0; k < mi->ntiles; k++) mi->ptiles[n++] = mi->tiles[k];
      for( int k = 0; k < mi->nneibs; k++)
         for( int j = 0; j < mi->neibs[k]->ntiles; j++)
            mi->ptiles[n++] = mi->neibs[k]->tiles[j];
   
      if( FLAG_PRECON)
      {
         //
         //  Construct a potential matrix mi->PP for the tiles in mi->ptiles
         //  and a copy mi->PM which will be inverted to form a preconditioner.
         //
   
         mi->PP = array( mi->np, PMTYPE *);
         PMTYPE **PM = array( mi->np, PMTYPE *);
   
         for( int j = 0; j < mi->np; j++)
         {
            mi->PP[j] = array( mi->np, PMTYPE);
            PM[j] = array( mi->np, PMTYPE);
   
            for( int k = 0; k < mi->np; k++)
               mi->PP[j][k] = PM[j][k] =
                  j == k ? mi->ptiles[j]->spc :
                           potential_coef( mi->ptiles[j], mi->ptiles[k]);
         }
   
         invert_matrix( mi->np, PM);
   
         // Just keep the first mi->ntiles rows of PM and assign to mi->C
         for( int r = mi->ntiles; r < mi->np; r++) free( PM[r]);
         mi->C = realloc( PM, mi->ntiles * sizeof( PMTYPE *));
   
         // Just keep the first mi->ntiles rows of PP
         for( int r = mi->ntiles; r < mi->np; r++) free( mi->PP[r]);
         mi->PP = realloc( mi->PP, mi->ntiles * sizeof( PMTYPE *));
      }
      else  // No preconditioning
      {
         // Construct a potential matrix mi->PP for the tiles in this box.
         // This is non-square: mi->PP[mi->ntiles][mi->np]
   
         mi->PP = array( mi->ntiles, PMTYPE *);
         for( int j = 0; j < mi->ntiles; j++)
         {
            mi->PP[j] = array( mi->np, PMTYPE);
   
            for( int k = 0; k < mi->np; k++)
               mi->PP[j][k] = 
                  j == k ? mi->ptiles[j]->spc :
                           potential_coef( mi->ptiles[j], mi->ptiles[k]);
         }
      }
   }
}

//
//  Initialise everything required for multipole (upward) expansions.
//

static void setup_multi( struct MBOX *mb,
                         __attribute__ ((unused)) void *arg1,
                         __attribute__ ((unused)) void *arg2)
{
   //
   //  Prepare multipole expansion coefficients for leaf boxes
   //

   if( mb->leaf)
   {
      mb->mecd = array( mb->ntiles, CXTYPE **);

      for( int k = 0; k < mb->ntiles; k++)
         mb->mecd[k] = coef_mec( &mb->tiles[k]->c, &mb->c);

      if( WORLD.g)
      {
         mb->meci = array( mb->ntiles, CXTYPE **);

         for( int k = 0; k < mb->ntiles; k++)
            mb->meci[k] = coef_mec( &mb->tiles[k]->ci, &mb->ci);
      }
   }

   //
   //  Calculate multipole translation coefficients of branch boxes
   //

   if( !mb->leaf)
   {
      mb->mtcd = array( 8, CXTYPE ****);

      for( int kid = 0; kid < 8; kid++)
         if( mb->kids[kid])
            mb->mtcd[kid] = coef_mtc( &mb->kids[kid]->c, &mb->c);

      if( WORLD.g)
      {
         mb->mtci = array( 8, CXTYPE ****);
   
         for( int kid = 0; kid < 8; kid++)
            if( mb->kids[kid])
               mb->mtci[kid] = coef_mtc( &mb->kids[kid]->ci, &mb->ci);
      }
   }

   //
   //  Allocate space for outgoing field components.
   //

   mb->mfd = alloc_fmm_field();
   if( WORLD.g) mb->mfi = alloc_fmm_field();
}

//
//  Initialise everything required for local (downward) expansions.
//

static void setup_local( struct MBOX *mb,
                         __attribute__ ((unused)) void *arg1,
                         __attribute__ ((unused)) void *arg2)
{
   //
   //  Coefficients for local expansion to tile potentials in leaf mboxes.
   //

   if( mb->leaf)
   {
      mb->lec = array( mb->ntiles, CXTYPE **);

      for( int k = 0; k < mb->ntiles; k++)
         mb->lec[k] = coef_lec( &mb->tiles[k]->c, &mb->c);
   }

   //
   //  Coefficients for local translation of parent's incoming field.
   //

   if( mb->parent) mb->ltc = coef_ltc( &mb->c, &mb->parent->c);

   //
   //  Coefficients for conversion of multipole fields from interaction list
   //  into local expansion field.
   //

   mb->m2lcd = array( mb->nilist, CXTYPE ****);

   for( int i = 0; i < mb->nilist; i++)
      mb->m2lcd[i] = coef_m2lc( &mb->ilist[i]->c, &mb->c);

   if( WORLD.g)
   {
      mb->m2lci = array( mb->nilist, CXTYPE ****);
   
      for( int i = 0; i < mb->nilist; i++)
         mb->m2lci[i] = coef_m2lc( &mb->ilist[i]->ci, &mb->c);
   }

   //
   //  Allocate space for incoming local field components.
   //

   mb->lf = alloc_fmm_field();
}

//
//  Initialise all the lists and coefficients for FMM passes.
//

static void init_fmm( void)
{
   Anm_setup();   // Precalculate some constants

   xlt = array( ntiles, int);  // For sanity checks only

   logmsg( 2, "preparing lists");
   pscan_octree( BREADTH_FIRST, setup_lists, NULL, NULL);

   logmsg( 2, "preparing potential matrices");
   pscan_octree( DEPTH_FIRST, setup_pp, NULL, NULL);

   logmsg( 2, "preparing multipole coefficients");
   pscan_octree( DEPTH_FIRST, setup_multi, NULL, NULL);

   logmsg( 2, "preparing local coefficients");
   pscan_octree( BREADTH_FIRST, setup_local, NULL, NULL);

   logmsg( 2, "coefficient cache size %d", nccache);
   for( int i = 1; i<= 3; i++)
      logmsg( 3, "class %d %d/%d", i, cachehit[i], cachecnt[i]);
}

//
//  Produce a preconditioned version of the tile charge array.
//

static void fmm_precon_pass( struct MBOX *mb, void *arg1, void *arg2)
{
   if( !mb->leaf) return;

   QVTYPE *q = (QVTYPE *) arg1;    // Array of source charges q[0 .. ntiles-1]
   QVTYPE *qp = (QVTYPE *) arg2;   // Array of preconditioned charges

   for( int j = 0; j < mb->ntiles; j++)
   {
      int tt = mb->tiles[j]->idx;
      qp[tt] = 0;

      for( int k = 0; k < mb->np; k++)
         qp[tt] += mb->C[j][k] * q[mb->ptiles[k]->idx];
   }
}

//
//  FMM upward pass.  Called depth-first for each mbox.
//

static void fmm_upward_pass( struct MBOX *mb,
                            void *arg1, __attribute__ ((unused)) void *arg2)
{
   QVTYPE *q = (QVTYPE *) arg1;   // Array of source charges q[0 .. ntiles-1]

   zero_fmm_field( mb->mfd);
   if( WORLD.g) zero_fmm_field( mb->mfi);
   mb->nmf = 0;  // Only for debug

   //
   //  For all leaf boxes, compute outgoing field components m->mf[][] for the
   //  tiles it contains.
   //  Ihler section 4.2 Eq 20.
   //

   if( mb->leaf)
   {
      for( int k = 0; k < mb->ntiles; k++)
         for( int l = 0; l < MXP; l++)
            for( int m = -l; m <= l; m++)
               mb->mfd[l][m] += mb->mecd[k][l][m] * q[mb->tiles[k]->idx];

      // If there's a ground plane, calculate outgoing field from image tiles
      if( WORLD.g)
         for( int k = 0; k < mb->ntiles; k++)
            for( int l = 0; l < MXP; l++)
               for( int m = -l; m <= l; m++)
                  mb->mfi[l][m] -= mb->meci[k][l][m] * q[mb->tiles[k]->idx];

      mb->nmf += mb->ntiles;  // Only for debug
   }

   //
   //  For all branch boxes, sum the kids' outgoing fields and translate to
   //  center of current box, ready for parent to make use of.
   //
   //  Ihler section 7.1.1 Eq 31.
   //
   
   if( !mb->leaf)
      for( int kid = 0; kid < 8; kid++)
      {
         struct MBOX *mk = mb->kids[kid];
         if( !mk) continue;

         apply_fmm_matrix( mb->mfd, mb->mtcd[kid], mk->mfd);

         // Do the same for the image field if we have a ground plane
         if( WORLD.g) apply_fmm_matrix( mb->mfi, mb->mtci[kid], mk->mfi);

         mb->nmf += mk->nmf;  // Only for debug
      }
}

//
//  FMM downward pass.  Called breadth first for each mbox.
//

static void fmm_downward_pass( struct MBOX *mb, void *arg1, void *arg2)
{
   QVTYPE *v = (QVTYPE *) arg1;        // Array of potentials v[0 .. ntiles-1]
   QVTYPE *q = (QVTYPE *) arg2;    // Array of source charges q[0 .. ntiles-1]

   zero_fmm_field( mb->lf);
   mb->nlf = 0;   // Only for debug

   //
   //  Direct terms for coupling between tiles inside each mbox and from
   //  tiles in neighbour boxes.
   //

   if( mb->leaf)
      for( int j = 0; j < mb->ntiles; j++)
         for( int k = 0; k < mb->np; k++)
         {
            int tt = mb->ptiles[j]->idx;
            int ts = mb->ptiles[k]->idx;
            v[tt] += mb->PP[j][k] * q[ts];

            xlt[tt]++;  // Only for debug
         }

   //
   //  Near field interaction from boxes in interaction list.
   //  Conversion of multipole expansion to local expansion.
   //  Ihler section 7.1.2, Greengard theorem 5.4
   //
 
   for( int is = 0; is < mb->nilist; is++)
   {
      struct MBOX *ms = mb->ilist[is];    // Source mbox

      apply_fmm_matrix( mb->lf, mb->m2lcd[is], ms->mfd);

      // Field from image of source box if we have a ground plane
      if( WORLD.g) apply_fmm_matrix( mb->lf, mb->m2lci[is], ms->mfi);

      mb->nlf += ms->nmf;   // Only for debug
   }

   //
   //  Local expansion of the far field coming down from parent mbox.
   //  Ihler 7.1.3 Eq 40.
   // 

   if( mb->parent)
   {
      struct MBOX *mp = mb->parent;
      apply_fmm_matrix( mb->lf, mb->ltc, mp->lf);
      mb->nlf += mp->nlf;  // Only for debug
   }

   //
   //  Evaluate mb->lf at each tile.
   //  Ihler 7.1.3 Eq 39.
   //

   if( mb->leaf)
      for( int k = 0; k < mb->ntiles; k++)   // Target tile
      {
         int tt = mb->tiles[k]->idx;

         for( int l = 0; l < MXP; l++)
            for( int m = -l; m <= l; m++)
               v[tt] += creal( mb->lec[k][l][m] * mb->lf[l][m]);
   
         xlt[tt] += mb->nlf;  // Only for debug
      }
}

static void compute_potentials_multi( QVTYPE *v, QVTYPE *q)
{
   for( int i = 0; i < ntiles; i++) xlt[i] = 0;  // XXX
   for( int i = 0; i < ntiles; i++) v[i] = 0;

   if( FLAG_PRECON)
   {
      QVTYPE *qp = array( ntiles, QVTYPE);
      pscan_octree( DEPTH_FIRST, fmm_precon_pass, q, qp);

      pscan_octree( DEPTH_FIRST, fmm_upward_pass, qp, NULL);
      if( root_mbox->nmf != ntiles)
         bailout( "root nmf %d/%d", root_mbox->nmf, ntiles);

      pscan_octree( BREADTH_FIRST, fmm_downward_pass, v, qp);
      free( qp);
   }
   else
   {
      pscan_octree( DEPTH_FIRST, fmm_upward_pass, q, NULL);
      if( root_mbox->nmf != ntiles)
         bailout( "root nmf %d/%d", root_mbox->nmf, ntiles);

      pscan_octree( BREADTH_FIRST, fmm_downward_pass, v, q);
   }

   for( int i = 0; i < ntiles; i++)
      if( xlt[i] < ntiles) bailout( "xlt %d %d\n", i, xlt[i]);
}

//
//  Diagnostic function to exercise the FMM transforms.
//

void diag_FMM( void)
{
   //
   //  A handful of source points clustered around 5,4,7, within a radius of
   //  about 1.
   //

   #define NTP 6

   struct TP {
      QVTYPE q;
      struct POINT p;
   } tps[NTP] = {
      { 12e-10, { 5.6, 4.6, 7.8 }},
      { 7e-10, { 4.9, 4.3, 6.2 }},
      { 14e-10, { 5.1, 3.4, 7.1 }},
      { 5e-10, { 4.7, 3.3, 7.2 }},
      { 10e-10, { 5.3, 4.5, 6.3 }},
      { 4e-10, { 4.2, 3.8, 6.8 }}
   };

   struct POINT OP = { -1, -2, 1 };  // Observation point

   //
   //  Field at OP calculated directly with Coulomb law
   //

   {
      QVTYPE v = 0;
      for( int i = 0; i < NTP; i++)
      {
         double r = pythagoras( tps[i].p, OP);
         v += tps[i].q/(4 * M_PI * EPSILON * r);
      }
      printf( "direct calc %.6e\n", v);
   }

   //
   //  Multipole field centered at MP1
   //

   struct POINT MP1 = { 5, 4, 7};  // A point near the center of the sources

   CXTYPE **mf1 = alloc_fmm_field();

   for( int i = 0; i < NTP; i++)
   {
      CXTYPE **mec1 = coef_mec( &tps[i].p, &MP1);

      for( int n = 0; n < MXP; n++)
         for( int m = -n; m <= n; m++) mf1[n][m] += mec1[n][m] * tps[i].q;
   }

   //  
   //  Evaluate the potential at OP from multipole field mf1 centered at MP1
   //

   {
      // Polar coords of OP relative to MP1
      double alpha, beta, p;
      to_polar( &OP, &MP1, &alpha, &beta, &p);

      QVTYPE v = 0;
      for( int l = 0; l < MXP; l++)
         for( int m = -l; m <= l; m++)
            v += creal( mf1[l][m] *
                        pow(1/p, l+1) *
                        Ynm( l, m, alpha, beta) * 
                        1/(EPSILON * 4 * M_PI));

      printf( "Eval at OP from mf1/MP1 %.6e\n", v);
   } 

   //
   //  Translate mf1 centered at MP1, to a new field mf2 centered at a
   //  point MP2, just outside the original circle.
   //

   struct POINT MP2 = { 4, 3, 6 };

   CXTYPE **mf2 = alloc_fmm_field();

   CXTYPE ****mtc1 = coef_mtc( &MP1, &MP2);
   apply_fmm_matrix( mf2, mtc1, mf1);

   //
   //  Evaluate the potential at OP from multipole field mf2 centered at MP2
   //

   {
      // Polar coords of OP relative to MP2
      double alpha, beta, p;
      to_polar( &OP, &MP2, &alpha, &beta, &p);

      QVTYPE v = 0;
      for( int l = 0; l < MXP; l++)
         for( int m = -l; m <= l; m++)
            v += creal( mf2[l][m] *
                        pow(1/p, l+1) *
                        Ynm( l, m, alpha, beta) * 
                        1/(EPSILON * 4 * M_PI));

      printf( "Eval at OP from mf2/MP2 %.6e\n", v);
   } 

   //
   //  Convert mf2 centered at MP2 to a local expansion field lf1 centered
   //  at LP1.  LP1 must be near the observation point.
   //

   struct POINT LP1 = { -1.5, -2.4, 0.7};

   CXTYPE **lf1 = alloc_fmm_field();
   CXTYPE ****m2lc = coef_m2lc( &MP2, &LP1);
   apply_fmm_matrix( lf1, m2lc, mf2);

   //
   //  Evaluate the potential at OP from local field lf1 centered at LP1
   //

   CXTYPE **lec1 = coef_lec( &OP, &LP1);

   {
      QVTYPE v = 0;
      for( int j = 0; j < MXP; j++)
         for( int k = -j; k <= j; k++) v += creal( lf1[j][k] * lec1[j][k]);

      printf( "Eval at OP from lf1/LP1 %.6e\n", v);
   }

   //
   //  Local field translation of lf1 at LP1 to lf2 centered at nearby LP2.
   //

   struct POINT LP2 = { -1.4, -2.3, 0.8};

   CXTYPE **lf2 = alloc_fmm_field();
   CXTYPE ****ltc1 = coef_ltc( &LP2, &LP1);
   apply_fmm_matrix( lf2, ltc1, lf1);

   //
   //  Evaluate the potential at OP from local field lf2 centered at LP2.
   //

   CXTYPE **lec2 = coef_lec( &OP, &LP2);

   {
      QVTYPE v = 0;
      for( int j = 0; j < MXP; j++)
         for( int k = -j; k <= j; k++) v += creal( lf2[j][k] * lec2[j][k]);

      printf( "Eval at OP from lf2/LP2 %.6e\n", v);
   }
}

//
//  Calculate the potentials v induced by the array of source charges q.
//

static void compute_potentials_direct_slave( QVTYPE *v, QVTYPE *q,
                                             int *i1, int *i2)
{
   for( int r = *i1; r < *i2; r++)
   {
      PMTYPE *a = PP[r];
      v[r] = 0;
      for( int c = 0; c < ntiles; c++) v[r] += a[c] * q[c];
   }
}

static void compute_potentials_direct( QVTYPE *v, QVTYPE *q)
{
   int r = ntiles / ncpu;
   int n1[MAXCPU], n2[MAXCPU];

   for( int n = 0; n < ncpu; n++)
   {
      n1[n] = n * r;
      n2[n] = n == ncpu - 1 ? ntiles : n * r + r;
  
      worker_run( compute_potentials_direct_slave, v, q, n1 + n, n2 + n, NULL);
   }

   worker_rendezvous();
}

static void (*compute_potentials)( QVTYPE *, QVTYPE *) = NULL;

//
//  Solve to obtain the charge distributions.
//

static void solve_charges( QVTYPE *v, QVTYPE *q)
{
   switch( is_method)
   {
      case ISM_GE: gaussian_elim( ntiles, PP, v, q);
                   break;

      case ISM_CG: cgm( ntiles, compute_potentials, v, q);
                   break;

      case ISM_GMRES: gmres( ntiles, compute_potentials, v, q);
                      break;
   }

   if( FLAG_PRECON)
   {
      QVTYPE *qp = array( ntiles, QVTYPE);
      pscan_octree( DEPTH_FIRST, fmm_precon_pass, q, qp);
      for( int i = 0; i < ntiles; i++) q[i] = qp[i];
      free( qp);
   }
}

static void solve_caps( void)
{
   struct TILE *bi;
   int i;

   if( is_method == ISM_GE)
      for( i = 0; i < ntiles; i++)
      {
         bi = tile_list[i];
         V[i] = -bi->electrode->volts;   
         Q[i] = 0;           
      }
   else
      for( i = 0; i < ntiles; i++)
      {
         bi = tile_list[i];
         V[i] = bi->electrode->volts;
         Q[i] = 0;     
      }

   solve_charges( V, Q);

   //
   //  Propagate the tile charges back to the electrodes.
   //

   for( i = 0; i < ntiles; i++) tile_list[i]->electrode->charge += Q[i];
}

static void do_capacitance( void)
{
   //
   //  If solving with direct coulomb calculation, make sure that the 
   //  potential matrix will fit given the address size
   //

   if( cp_method == CPM_DIR)
   {
      logmsg( 1, "memory for potential matrix: %.1f GB",
          (double) sizeof( PMTYPE) * ntiles * ntiles/(1024.0*1024.0*1024.0));

      if( sizeof( size_t) == 4 && ntiles > 0.8 * 65536/sqrt(sizeof(PMTYPE)))
      {
         logmsg( 0, "too many tiles (%d) for 32 bit machine", ntiles);
         logmsg( 0, "maybe try -m fmm");
         bailout( NULL);
      }

      if( sizeof( size_t) == 8 && ntiles > 0.8 * 4.294e9/sqrt(sizeof(PMTYPE)))
      {
         logmsg( 0, "too many tiles (%d) for 64 bit machine!", ntiles);
         logmsg( 0, "maybe try -m fmm");
         bailout( NULL);
      }
   }

   V = array( ntiles, QVTYPE);
   Q = array( ntiles, QVTYPE);

   if( cp_method == CPM_DIR)
   {
      PP = array2D( ntiles, PMTYPE);
      prepare_potential_matrix();
   }

   for( struct ELECTRODE *e = WORLD.electrodes; e; e = e->next)
   {
      logmsg( 2, "activating electrode %s", e->name);

      reset_all_electrodes();
      e->volts = 1.0;
      solve_caps();
      sum_charges( e);
   }

   logmsg( 1, "capacitance completed");
}

//
//  Calculate charge distribution for the given background field. 
//

static void do_efield( void)
{
   reset_all_electrodes();

   int i;

   // Unit vector along Efield
   struct POINT p; forepoint( WORLD.Efield, 1.0, &p, NULL);

   if( WORLD.g)
   {
      // Constrain E-field to be perpendicular to ground plane
      double a = inner_prod_3D( p, WORLD.g->norm);
      p.x = a * WORLD.g->norm.x;
      p.y = a * WORLD.g->norm.y;
      p.z = a * WORLD.g->norm.z;

      double m = mag_3D(p);
      if( !m) logmsg( 0, "E-field completely suppressed by ground plane");
      else
      if( m < 0.001)
         logmsg( 0, "E-field almost completely suppressed by ground plane");
   }

   //  Set V[] to the potential of the field at each tile location.  If a
   //  ground plane has been specified, use that as the zero reference plane
   //  for potential.  Otherwise place a zero potential at an arbitrary point
   //  (use the system's centroid  to minimise numerical errors) and later,
   //  adjust the potentials so that the total flux to infinity is zero.

   struct POINT refp = WORLD.g ? *WORLD.g->center : WORLD.rc;

   if( is_method == ISM_GE)
      for( i = 0; i < ntiles; i++)
      {
         struct TILE *t = tile_list[i];
         V[i] = -((t->c.x - refp.x) * p.x +
                  (t->c.y - refp.y) * p.y +
                  (t->c.z - refp.z) * p.z);
         Q[i] = 0;           
      }
   else
      for( i = 0; i < ntiles; i++)
      {
         struct TILE *t = tile_list[i];
         V[i] = (t->c.x - refp.x) * p.x +
                (t->c.y - refp.y) * p.y +
                (t->c.z - refp.z) * p.z;
         Q[i] = 0;           
      }

   solve_charges( V, Q);

   //
   //  Propagate the tile charges back to the electrodes and add up the total
   //  charge induced by the field.
   //

   QVTYPE total_charge = 0;
   for( i = 0; i < ntiles; i++)
   {
      tile_list[i]->electrode->charge += Q[i];
      total_charge += Q[i];
   }

   logmsg( 2, "efield total charge %.3e", total_charge);

   //
   //  Total C to infinity of all the electrodes.
   //

   double total_cinf = 0;
   struct C_STORE *cp;
   for( cp = Cstore, i = 0; i < Cstore_cnt; i++, cp++)
      if( cp->e1 == cp->e2) total_cinf += cp->C;

   logmsg( 2, "total cinf %.3e", total_cinf);

   //  Store the induced charge on each electrode.  If no ground plane, adjust
   //  the charge injected into each electrode so that the net displacement
   //  current from infinity is zero.

   for( struct ELECTRODE *e = WORLD.electrodes; e; e = e->next)
   {
      QVTYPE Q = e->charge;
      if( !WORLD.g) Q -= total_charge * find_C( e, e)/total_cinf;
      store_Q( e, Q);
   }
 
   logmsg( 1, "electric field completed");
}

///////////////////////////////////////////////////////////////////////////////
//  Inductance Calculation                                                   //
///////////////////////////////////////////////////////////////////////////////

//
//  Straight out of 'Numerical Recipes':
//
//  Computes Carlson's elliptic integral of the first kind, RF (x, y, z).
//  x, y, and z must be non-negative, and at most one can be zero.
//  TINY must be at least 5 times the machine underflow limit,
//  BIG at most one fifth the machine overflow limit.
//

#define ERRTOL 0.0025

#define TINY 1.5e-38
#define BIG 3.0e37

#define CF1 (1.0/24.0)
#define CF2 0.1
#define CF3 (3.0/44.0)
#define CF4 (1.0/14.0)

static inline double FMIN( double a, double b) { return a < b ? a : b; }
static inline double FMAX( double a, double b) { return a > b ? a : b; }

static double carlson_RF( double x, double y, double z)
{
   double alamb, ave, delx, dely, delz,
          e2, e3, sqrtx, sqrty, sqrtz, xt, yt, zt;

   if( FMIN( FMIN( x,y), z) < 0.0 || FMIN( FMIN( x+y, x+z), y+z) < TINY ||
       FMAX( FMAX( x,y), z) > BIG) bailout( "invalid arguments in carlson RF");

   xt = x; yt = y; zt = z;

   do {
      sqrtx = sqrt( xt); sqrty = sqrt( yt); sqrtz = sqrt( zt);

      alamb = sqrtx * (sqrty + sqrtz) + sqrty * sqrtz;

      xt = 0.25 * (xt+alamb); yt = 0.25 * (yt+alamb); zt = 0.25 * (zt+alamb);

      ave = (xt+yt+zt)/3;

      delx = (ave-xt)/ave; dely = (ave-yt)/ave; delz = (ave-zt)/ave;
   }
    while( FMAX( FMAX( fabs( delx), fabs( dely)), fabs( delz)) > ERRTOL);

   e2 = delx*dely - delz*delz;
   e3 = delx*dely * delz;

   return (1.0+(CF1*e2-CF2-CF3*e3)*e2+CF4*e3)/sqrt(ave);
}

#define CD1 (3.0/14.0)
#define CD2 (1.0/6.0)
#define CD3 (9.0/22.0)
#define CD4 (3.0/26.0)
#define CD5 (0.25*CD3)
#define CD6 (1.5*CD4)

static double carlson_RD( double x, double y, double z)
{
   double alamb, ave, delx, dely, delz,
          ea, eb, ec, ed, ee, fac, sqrtx, sqrty, sqrtz,
          sum, xt, yt, zt;

   if( FMIN(x,y) < 0 || FMIN(x+y,z) < TINY ||
       FMAX(FMAX(x,y),z) > BIG) bailout( "invalid arguments in carlson RD");

   xt = x;  yt = y; zt = z;
   sum = 0; fac = 1;

   do {
      sqrtx = sqrt(xt); sqrty = sqrt(yt); sqrtz = sqrt(zt);
      alamb = sqrtx * (sqrty + sqrtz) + sqrty * sqrtz;
      sum += fac/(sqrtz*(zt+alamb));

      fac = 0.25 * fac;
      xt = 0.25 * (xt+alamb); yt = 0.25 * (yt+alamb); zt = 0.25 * (zt+alamb);

      ave = 0.2 * (xt + yt + 3 * zt);
      delx = (ave - xt)/ave;
      dely = (ave - yt)/ave;
      delz = (ave - zt)/ave;
   }
    while(FMAX(FMAX(fabs(delx),fabs(dely)),fabs(delz)) > ERRTOL);

   ea = delx * dely;
   eb = delz*delz;
   ec = ea - eb;
   ed = ea - 6 * eb;
   ee = ed + ec + ec;

   return 3*sum + fac*(1 + ed*(-CD1 + CD5*ed - CD6*delz*ee) +
          delz * (CD2*ee + delz*(-CD3*ec+delz*CD4*ea)))/(ave * sqrt(ave));

}

static inline double elliptic_K( double p)
{
   return carlson_RF( 0, 1 - p*p, 1);
}

static inline double elliptic_E( double p)
{
   return carlson_RF( 0, 1 - p*p, 1) - p*p/3 * carlson_RD(0, 1 - p*p, 1);
}

//
//  Calculate M between circular filaments using the algorithm of
//  Babic, Sirois, Akyel, Girardi.
//  IEEE transactions on magnetics, vol 49 num 9 Sept 2010.
//

//  Assumes primary is horizontal and centered at 0,0,0
//  and Rp >= Rs

#define DPHI 0.1   // Default integration step, radians

static double compute_BSAG(
    double Rp, double Rs,                     // Radius of primary, secondary
    double xc, double yc, double zc,          // Secondary center
    double a, double b, double c,             // Secondary plane
    double dphi)                              // Integration step, radians
{
   double alpha = Rs/Rp;
   double beta = xc/Rp;
   double gamma = yc/Rp;
   double delta = zc/Rp;

   int sgnb = b >= 0 ? 1 : -1;

   int alt_def = FALSE;
   double el = sqrt(a*a  + c*c);
   if( el < 0.1) alt_def = TRUE;
 
   double eL = sqrt(a*a + b*b + c*c);

   double p1, p2, p3, p4, p5;
   if( !alt_def)
   {
      p1 = gamma * c/el;
      p2 = -(beta * el * el + gamma * a * b)/(el * eL);
      p3 = alpha * c/eL;
      p4 = -(beta * a * b - gamma * el * el + delta * b * c)/(el * eL);
      p5 = -(beta * c - delta * a)/el;
   }
   else
   {
      p1 = 0;
      p2 = -gamma * sgnb;
      p3 = 0;
      p4 = -beta * sgnb;
      p5 = delta;
   }

   double phi, sum = 0;   // Integral over phi: 0..2pi
  
   for( phi = 0; phi < 2*M_PI; phi += dphi)
   {
      double A0 = 1 + alpha*alpha + beta*beta + gamma*gamma + delta*delta +
                  2*alpha*(p4*cos(phi) + p5*sin(phi));
      double t1 = (1 - b*b*c*c/(el*el*eL*eL))*cos(phi)*cos(phi) +
                  c*c*sin(phi)*sin(phi)/(el*el) +
                  a*b*c/(el*el*eL)*sin(2*phi);
      double t2 = beta*beta + gamma*gamma -
                  2*alpha*(beta*a*b - gamma*el*el)/(el*eL)*cos(phi) -
                  2*alpha*beta*c/el*sin(phi);
  
      double V0;
      if( !alt_def) V0 = sqrt( alpha*alpha*t1 + t2);
      else
         V0 = sqrt(beta*beta + gamma*gamma + alpha*alpha*cos(phi)*cos(phi) -
                   2*alpha*beta*sgnb*cos(phi));

      double k = sqrt(4*V0/(A0 + 2*V0));
      double Psi = (1 - k*k/2) * elliptic_K(k) - elliptic_E(k);

      sum += dphi*((p1*cos(phi) + p2*sin(phi) + p3)*Psi) / (k*sqrt(V0*V0*V0));
   }

   return sum * MU * Rs/M_PI;
}

//
//  A set of tests to check the function of compute_BSAG().
//  Invoke with option -d bsag
//

static void diag_BSAG( void)
{
   double a, b, c;
   double xc, yc, zc;
   double Rp, Rs;
   double phi, theta;
   double M;

   // Example 1: 62-Grover
   phi = 0.64350111;
   a = b = 0;  c = 1;  xc = 0; yc = 0.2*sin(phi); zc = 0.2*cos(phi);
   Rs = Rp = 0.15;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example1 %.2f nH expect 45.31 nH\n", M * 1e9);
  
   // Example 2: 63-Grover
   Rs = Rp = 0.6096;
   a = b = 0; c = 1;
   xc = 0; yc = 1.21158; zc = 0.381; 
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example2 %.2f nH expect -24.57 nH\n", M * 1e9);

   // Example 3: 64-Grover
   Rs = Rp = 0.1524;
   a = b = 0; c = 1;
   zc = 0; yc = 0.4572; zc = 0;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example3a %.3f nH expect -7.562 nH\n", M * 1e9);

   zc = 0; yc = 0.30481; zc = 0;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example3b %.2f nH expect -85.97 nH\n", M * 1e9);

   zc = 0; yc = 0.0762; zc = 0;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example3c %.2f nH expect 268.8 nH\n", M * 1e9);

   // Example 4: 65-Grover
   phi = 1.1592795;
   Rs = 0.08; Rp = 0.1;
   a = b = 0; c = 1;
   xc = 0; yc = 0.5*sin(phi);  zc = 0.5*cos(phi);
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example4 %.3f nH expect -0.2483 nH\n", M * 1e9);

   // Example 5: 66-Grove
   phi = acos(0.6);
   Rs = 0.08; Rp = 0.1;
   a = b = 0; c = 1;
   xc = 0; yc = 0.2*sin(phi); zc = 0.2*cos(phi);
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example5 %.3f nH 4.465 nH\n", M * 1e9);

   // Example 7: 69-Grover
   theta = acos(0.3);
   Rs = 0.14; Rp = 0.2;
   a = 0; b = -sin(theta); c = cos(theta);
   xc = yc = zc = 0;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example7 %.2f nH 47.44 nH\n", M * 1e9);

   // Example 8: 70-Grover
   theta = acos(0.4);
   Rs = 0.0762;  Rp = 0.254;
   a = 0; b = -sin(theta); c = cos(theta);
   xc = 0; yc = 0; zc = 0.0762;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example8 %.2f nH 15.54 nH\n", M * 1e9);

   // Example 9: 71-Grover
   theta = 30.0/180 * M_PI;
   Rs = 0.1; Rp = 0.2;
   a = 0; b = -sin(theta); c = cos(theta);
   xc = 0; yc = 0; zc = 0.2;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example9 %.2f nH 29.44 nH\n", M * 1e9);

   // Example 10: 73-Grover
   theta = 60.0/180 * M_PI;
   Rs = 0.1; Rp = 0.16;
   a = 0; b = -sin(theta); c = cos(theta);
   xc = 0; yc = 0.05*sin(theta); zc = 0.2 - 0.05*cos(theta);
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example10 %.2f nH 13.61 nH\n", M * 1e9);

   // Example 11
   theta = M_PI/2;
   Rs = 0.1; Rp = 0.4;
   a = 0; b = -sin(theta); c = 0;
   xc = 0; yc = 0.2; zc = 0.1;
   M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
   printf( "Example11 %.2f nH -10.73 nH\n", M * 1e9);

   // Example 12
   theta = 60.0/180 * M_PI;
   Rs = 0.1; Rp = 0.16;

   struct {
      double phi;
      double M;
   } ex12[] = {
     { 0, 13.6113 },
     { M_PI/6, 14.4688 },
     { M_PI/4, 15.4877 },
     { M_PI/3, 16.8189},
     { M_PI/2, 20.0534},
     { 2*M_PI/3, 23.3253},
     { 3*M_PI/4, 24.6936},
     { 5*M_PI/6, 25.7493},
     { M_PI, 26.6433},
     { 7*M_PI/6, 25.7493},
     { 5*M_PI/4, 24.6936},
     { 4*M_PI/3, 23.3253 },
     { 3*M_PI/2, 20.0534},
     { 5*M_PI/3, 16.8189},
     { 7*M_PI/4, 15.4877},
     { 11*M_PI/6, 14.4688},
     { 2*M_PI, 13.6113}
   };

   xc = 0; yc = 0.05 * sin(theta); zc = 0.2 - 0.05*cos(theta);

   int i;
   for( i = 0; i < 17; i++)
   {
      a = sin(ex12[i].phi) * sin(theta);
      b = -cos(ex12[i].phi) * sin(theta);
      c = cos(theta);
      M = compute_BSAG( Rp, Rs, xc, yc, zc, a, b, c, DPHI);
      printf( "Example12-%d %.2f nH %.4f nH\n", i, M * 1e9, ex12[i].M);
   }
}

#define NF_DEFAULT 20                   // Number of filaments per coil section

//
//  Partial inductance between two straight segments.
//

static double partial_inductance( struct POINT *p1a, struct POINT *p1b,
                                  struct POINT *p2a, struct POINT *p2b)
{
   int N = 100;
   double sum = 0;
   struct POINT d1 = sub_3D( *p1b, *p1a);  scale_3D( &d1, 1.0/N);
   struct POINT d2 = sub_3D( *p2b, *p2a);  scale_3D( &d2, 1.0/N);

   struct POINT p1 = *p1a;
   p1.x += d1.x/2; 
   p1.y += d1.y/2; 
   p1.z += d1.z/2; 

   for( int i1 = 0; i1 < N; i1++)
   {
      struct POINT p2 = *p2a;
      p2.x += d2.x/2; 
      p2.y += d2.y/2; 
      p2.z += d2.z/2; 

      for( int i2 = 0; i2 < N; i2++)
      {
         double d = mag_3D( sub_3D( p2, p1));

         sum += 1/d;
         sum_3D( &p2, d2);
      }

      sum_3D( &p1, d1);
   }
   return sum * inner_prod_3D( d1, d2) * MU/(4 * M_PI); 
}

//
//  Mutual inductance between two circular filaments.  The primary radius rp
//  must be greater than or equal to the secondary radius rs.
//

static double Lmutual_circular( struct CSECT *sp, double hp, double rp,
                                struct CSECT *ss, double hs, double rs,
                                double dphi)
{
   // Center of the primary filament
   struct POINT p; forepoint( &sp->axis, hp, &p, &sp->end1);
   
   // Center of the secondary filament
   struct POINT s; forepoint( &ss->axis, hs, &s, &ss->end1);
   
   // Translate so that center of the primary filament is at the origin
   s = sub_3D( s, p);
   
   // Rotate position of secondary until primary axis is vertical
   s = unrotate( s, sp->axis.pa, sp->axis.azim);
   
   // Make unit vector in direction of secondary
   struct POINT v; forepoint( &ss->axis, 1, &v, NULL);
   
   // Rotate it so that primary axis is vertical
   v = unrotate( v, sp->axis.pa, sp->axis.azim);
   
   // Now we have the primary filament centered at 0,0,0 with axis
   // vertical (+z).   Secondary filament is centered at s.x,s.y,s.z
   // with axis given by unit vector v.x,v.y,v.z

   // Calculate the plane of the rotated secondary
   double azim = atan2( v.y, v.x);  // phi
   double r = sqrt(v.x*v.x + v.y*v.y);
   double pa = atan2( r, v.z);  // theta
   
   double a = cos(azim) * sin(pa);
   double b = sin(azim) * sin(pa);
   double c = cos(pa);
   
   return compute_BSAG( rp, rs, s.x, s.y, s.z, a, b, c, dphi);
}

//
//  Coupling between two polygonal filaments.
//

static double Lmutual_polygon( struct COIL *c1, struct CSECT *s1,
                               double t1,
                               struct COIL *c2, struct CSECT *s2,
                               double t2)
{
   double M = 0;

   struct POINT v;

   for( int i1 = 0; i1 < c1->npoints; i1++)
   {
      int j1 = (i1+1) % c1->npoints;
      struct POINT p1a, p1b;

      v = sub_3D( s1->end2_poly[i1], s1->end1_poly[i1]);
      scale_3D( &v, t1/s1->turns);
      p1a = add_3D( s1->end1_poly[i1], v);

      v = sub_3D( s1->end2_poly[j1], s1->end1_poly[j1]);
      scale_3D( &v, t1/s1->turns);
      p1b = add_3D( s1->end1_poly[j1], v);

      for( int i2 = 0; i2 < c1->npoints; i2++)
      {
         int j2 = (i2+1) % c2->npoints;
         struct POINT p2a, p2b;
 
         v = sub_3D( s2->end2_poly[i2], s2->end1_poly[i2]);
         scale_3D( &v, t2/s2->turns);
         p2a = add_3D( s2->end1_poly[i2], v);

         v = sub_3D( s2->end2_poly[j2], s2->end1_poly[j2]);
         scale_3D( &v, t2/s2->turns);
         p2b = add_3D( s2->end1_poly[j2], v);

         if( c1 == c2 && s1 == s2 && t1 == t2 && i1 == i2)
         {
            // Partial self inductance of a straight piece of wire
            struct POINT v = sub_3D( *c1->end1_poly[(i1+1)%c1->npoints],
                                     *c1->end1_poly[i1]);

            double s = mag_3D(v);
            double a = c1->wirad/s;
            M += MU/(2*M_PI) * s *
                   (a - sqrt(1+a*a) + log(1/a) + log(1 + sqrt(1+a*a)));
         }
         else
            M += partial_inductance( &p1a, &p1b, &p2a, &p2b);
      }
   }

   return M;
}

//
//  Polygonal filament of c1,s1 coupling to circular filament of s2.
//

static double Lmutual_mixed( struct COIL *c1, struct CSECT *s1,
                             double t1,
                             struct CSECT *s2,
                             double h2, double r2,
                             double dphi)
{
   double sum = 0;

   // Center of circular filament
   struct POINT s; forepoint( &s2->axis, h2, &s, &s2->end1);

   // Unit vector along axis of circular filament
   struct POINT av; forepoint( &s2->axis, 1, &av, NULL);

   struct POINT v;
   for( int i1 = 0; i1 < c1->npoints; i1++)
   {
      int j1 = (i1+1) % c1->npoints;

      struct POINT p1a, p1b;
      v = sub_3D( s1->end2_poly[i1], s1->end1_poly[i1]);
      scale_3D( &v, t1/s1->turns);
      p1a = add_3D( s1->end1_poly[i1], v);

      v = sub_3D( s1->end2_poly[j1], s1->end1_poly[j1]);
      scale_3D( &v, t1/s1->turns);
      p1b = add_3D( s1->end1_poly[j1], v);

      int N = 100;
      struct POINT d1 = sub_3D( p1b, p1a);  scale_3D( &d1, 1.0/N);

      struct POINT p1 = p1a;
      p1.x += d1.x/2; 
      p1.y += d1.y/2; 
      p1.z += d1.z/2; 

      for( int k = 0; k < N; k++)
      {
         for( double phi = 0; phi < 2 * M_PI; phi += dphi)
         {
            // Radius vector in the plane of the circular filament
            struct POINT rv =
               rotate( (struct POINT ){cos(phi), sin(phi), 0},
                        s2->axis.pa, s2->axis.azim);

            // Shift to center of circular filament
            struct POINT p2 = add_3D( rv, s);

            // Tangent vector at point q
            struct POINT tq = xprod_3D( rv, av);
            norm_3D( &tq);
            scale_3D( &tq, r2 * dphi);

            double d = mag_3D( sub_3D( p2, p1));

            sum += 1.0/d * inner_prod_3D( d1, tq);
         }

         sum_3D( &p1, d1);
      }
   }

   return sum * MU/(4 * M_PI) ;
}

//
//  Return the mutual inductance between two coil sections.
//

static double Lmutual( struct COIL *c1, struct CSECT *s1,
                       struct COIL *c2, struct CSECT *s2)
{
   // Number of circular filaments per coil section
   int NF1 = NF_DEFAULT;
   if( fabs( s1->turns) < NF1) NF1 = fabs( round(s1->turns));
   int NF2 = NF_DEFAULT;
   if( fabs( s2->turns) < NF2) NF2 = fabs( round(s2->turns));
   if( NF1 < 2) NF1 = 2;
   if( NF2 < 2) NF2 = 2;
  
   if( c1->loop) NF1 = 1;
   if( c2->loop) NF2 = 1;

   // Radius increment per filament
   double dR1 = (s1->radius2 - s1->radius1)/NF1;
   double dR2 = (s2->radius2 - s2->radius1)/NF2;

   // Axial increment per filament
   double dH1 = s1->length/NF1;
   double dH2 = s2->length/NF2;

   // Turns increment per filament
   double dT1 = s1->turns/NF1;
   double dT2 = s2->turns/NF2;

   double M = 0;
   int j, k;

   double dphi = DPHI;
   if( c1->dphi) dphi = c1->dphi;
   if( c2->dphi && c2->dphi < dphi) dphi = c2->dphi;

   for( j = 0; j < NF1; j++)   // For each filament of coil1 section
      for( k = 0; k < NF2; k++)  // For each filament of coil2 section
      {
         // Axial position of filaments relative to start of sections
         double h1 = (j + 0.5) * dH1;
         double h2 = (k + 0.5) * dH2;
     
         // Turns position of filaments relative to start of sections
         double t1 = (j + 0.5) * dT1;
         double t2 = (k + 0.5) * dT2;
      
         if( !c1->npoints && !c2->npoints)  // Both filaments circular?
         {
            // Radius of filaments
            double r1 = s1->radius1 + (j + 0.5) * dR1;
            double r2 = s2->radius1 + (k + 0.5) * dR2;

            if( c1 == c2 && s1 == s2 && j == k)
               M += MU * r1 * ((log(8*r1/c1->wirad) - 2) + 1.0/4);
            else
            {
               // For BSAG, the 'primary' must have the larger radius
               if( r2 >= r1)
                  M += Lmutual_circular( s2, h2, r2, s1, h1, r1, dphi);
               else
                  M += Lmutual_circular( s1, h1, r1, s2, h2, r2, dphi);
            }
         }
         else
         if( c1->npoints && c2->npoints) // Both filaments polygons?
         {
            M += Lmutual_polygon( c1, s1, t1, c2, s2, t2);
         }
         else
         if( c1->npoints && !c2->npoints) // c1 polygon, c2 circular?
         {
            double r2 = s2->radius1 + (k + 0.5) * dR2;

            M += Lmutual_mixed( c1, s1, t1, s2, h2, r2, dphi);
         }
         else
         if( !c1->npoints && c2->npoints) // c1 circular, c2 polygon?
         {
            double r1 = s1->radius1 + (j + 0.5) * dR1;

            M += Lmutual_mixed( c2, s2, t2, s1, h1, r1, dphi);
         }
      }

   //
   //  Return mutual inductance, scaled from the number of filaments, to
   //  the actual number of turns in these sections.
   //

   return M * s1->turns * s2->turns / (NF1 * NF2);
}

static void inductance_slave( struct COIL *c1, struct CSECT *s1,
                              struct COIL *c2, struct CSECT *s2)
{
   double L = Lmutual( c1, s1, c2, s2);

   if( WORLD.g) L += Lmutual( c1->image, s1->image, c2, s2);

   if( c1 == c2 && s1 == s2 && L < 0)
   {
      logmsg( 0, "unable to calculate Lself for %s", c1->name);
      logmsg( 0, "maybe wire radius too big for pitch?");
      bailout( NULL);
   }

   store_L( c1, s1->unit, c2, s2->unit, L);
}

static void do_inductance( void)
{
   for( struct COIL *c1 = WORLD.coils; c1; c1 = c1->next)
      for( int i1 = 0; i1 < c1->Ns; i1++)
         for( struct COIL *c2 = c1; c2; c2 = c2->next)
            for( int i2 = c1 == c2 ? i1 : 0; i2 < c2->Ns; i2++)
               worker_run( inductance_slave, c1, c1->sections + i1,
                                             c2, c2->sections + i2, NULL);

   worker_rendezvous();
   logmsg( 1, "inductance completed");
}

//
//  Calculate response to incident B field.
//

static void do_bfield( void)
{
   // Unit vector along Bfield
   struct POINT f; forepoint( WORLD.Bfield, 1.0, &f, NULL);

   if( WORLD.g)
   {
      // Constrain B-field to be parallel to ground plane
      double a = inner_prod_3D( f, WORLD.g->norm);
      f.x -= a * WORLD.g->norm.x;
      f.y -= a * WORLD.g->norm.y;
      f.z -= a * WORLD.g->norm.z;

      double m = mag_3D(f);
      if( !m) logmsg( 0, "B-field completely suppressed by ground plane");
      else
      if( m < 0.001)
         logmsg( 0, "B-field almost completely suppressed by ground plane");
   }

   for( struct COIL *c = WORLD.coils; c; c = c->next)
      for( int i = 0; i < c->Ns; i++)
      {
         struct CSECT *s = c->sections + i;

         // Unit vector along section axis
         struct POINT a; forepoint( c->axis, 1.0, &a, NULL);

         // Component of B along section axis
         double B = inner_prod_3D( f, a);

         // Average cross section area of section 
         double area;
         if( c->npoints) // Polygon
            area = polygon_area( c->axis, c->end1_poly, c->npoints);
         else  // Circular
         {
            double r = (s->radius2 + s->radius1)/2;
            area = M_PI * r * r;
         }

         double F = s->turns * area * B;   // Total linked flux
         store_F( c, i, F);
      }

   logmsg( 1, "magnetic field completed");
}

///////////////////////////////////////////////////////////////////////////////
//  Command Line Stuff                                                       //
///////////////////////////////////////////////////////////////////////////////

static void usage( void)
{
   fprintf( stderr, "lcng version %s\n\n", VERSION);

   fprintf( stderr,
       "usage:\n"
       "  lcng [options] system\n"
       "\noptions:\n"
       " -o spice   Generate spice sub-circuit\n"
       " -o list    Output a listing of R, L and C\n"
       " -o tiles   Generate a tiles file\n"
       " -o povray  Generate an input file for povray\n"
       " -o gnuplot Display system interactively using gnuplot\n"
       " -o geda    Output gEDA symbol\n"
       "\n"
       " options for -o can be combined comma separated\n"
       "\n"
       " -r dc      Use DC resistance for coils\n"
       " -r ac      Use AC resistance for coils (default)\n"
       "\n"
       "Mode options for capacitance solver\n"
       " -m dir        Use direct calculation for capacitance (default)\n"
       " -m fmm        Use fast multipole method for capacitance\n"
       " -m cgm        Conjugate gradient method (default with -m dir)\n"
       " -m gmres      Generalised minimum residuals (default with -m fmm)\n"
       " -m ge         Use Gaussian elimination (only with -m dir)\n"
       "Options available with -m fmm\n"
       " -m depth=num  Use fixed depth octree, 'num' gives depth (default 5)\n"
       " -m depth=auto Use fixed depth octree, automatically choose depth\n"
       " -m precon     Enable preconditioning (only with -m fmm)\n"
       " -m degree=num Maximum degree of spherical harmonics (default 6)\n"
       "\n"
       " options for -m can be combined comma separated\n"
       "\n"
       " -t factor  Scales the number of tiles by factor (default 1.0)\n"
       " -c cpus    Use number of threads in parallel (maximum 64, default 1)\n"
       "\n"
       " -NC        Don't do capacitance calculation\n"
       " -NL        Don't do inductance calculation\n"
       " -NR        Don't do resistance calculation\n"
       " -d bsag    Test BSAG routine\n"
       " -d fmm     Test FMM\n"
   );

   exit(1);
}

static void parse_output_options( char *s)
{
   while( s && *s)
   {
      char *p = strchr( s, ','); if( p) *p++ = 0;

      if( !strcmp( s, "spice")) OPT_spice = TRUE;
      else
      if( !strcmp( s, "list")) OPT_list = TRUE;
      else
      if( !strcmp( s, "tiles")) OPT_tiles = TRUE;
      else
      if( !strcmp( s, "povray")) OPT_povray = TRUE;
      else
      if( !strcmp( s, "gnuplot")) OPT_gnuplot = TRUE;
      else
      if( !strcmp( s, "geda")) OPT_geda = TRUE;
      else
      {
         fprintf( stderr, "unrecognised output option: %s\n", s);
         usage();
      }

      s = p;
   }
}

static void parse_diag_options( char *s)
{
   while( s && *s)
   {
      char *p = strchr( s, ','); if( p) *p++ = 0;

      if( !strcmp( s, "bsag")) DIAG_bsag = TRUE;
      else
      if( !strcmp( s, "fmm")) DIAG_fmm = TRUE;
      else
      {
         fprintf( stderr, "unrecognised diagnostic option: %s\n", s);
         usage();
      }

      s = p;
   }
}

static void parse_r_options( char *s)
{
   while( s && *s)
   {
      char *p = strchr( s, ','); if( p) *p++ = 0;

      if( !strcmp( s, "dc")) FLAG_RDC = TRUE;
      else
      if( !strcmp( s, "ac")) FLAG_RDC = FALSE;
      else
      {
         fprintf( stderr, "unrecognised -r option: %s\n", s);
         usage();
      }

      s = p;
   }
}

static void parse_m_options( char *s)
{
   while( s && *s)
   {
      char *p = strchr( s, ','); if( p) *p++ = 0;

      if( !strcmp( s, "ge")) is_method = ISM_GE;
      else
      if( !strcmp( s, "cgm")) is_method = ISM_CG;
      else
      if( !strcmp( s, "gmres")) is_method = ISM_GMRES;
      else
      if( !strcmp( s, "fmm")) cp_method = CPM_FMM;
      else
      if( !strcmp( s, "dir")) cp_method = CPM_DIR;
      else
      if( !strcmp( s, "precon")) FLAG_PRECON = TRUE;
      else
      if( !strcmp( s, "depth=auto"))
      {
         if( cp_method != CPM_FMM)
            bailout( "-m auto depth not available with direct solver");
         autodepth = TRUE;
      }
      else
      if( !strncmp( s, "depth=", 6))
      {
         if( cp_method != CPM_FMM)
            bailout( "-m depth not available with direct solver");
         FIXDEPTH = atoi( s+6);
      }
      else
      if( !strncmp( s, "degree=", 7)) MXP = atoi( s+7);
      else
      {
         fprintf( stderr, "unrecognised -m option: %s\n", s);
         usage();
      }

      s = p;
   }

   if( (is_method == ISM_GE || is_method == ISM_CG) && cp_method == CPM_FMM)
      bailout( "cannot use FMM with selected charge solver, use gmres");
}

int main( int argc, char *argv[])
{
   while( 1)
   {
      int c = getopt( argc, argv, "vm:o:d:N:r:t:c:?");

      if( c == 'v') VFLAG++;
      else
      if( c == 'm') parse_m_options( optarg);
      else
      if( c == 'r') parse_r_options( optarg);
      else
      if( c == 'c') ncpu = atoi( optarg);
      else
      if( c == 'o') parse_output_options( optarg);
      else
      if( c == 'd') parse_diag_options( optarg);
      else
      if( c == 't') tile_scale = atof( optarg);
      else
      if( c == 'N')
      {
         if( !strcmp( optarg, "C")) FLAG_NC = TRUE;
         else
         if( !strcmp( optarg, "L")) FLAG_NL = TRUE;
         else
         if( !strcmp( optarg, "R")) FLAG_NR = TRUE;
         else
         {
            fprintf( stderr, "unknown -N option [%c]\n", c);
            usage();
         }
      }
      else
      if( c == '?') usage();
      else
      if( c == -1) break;
      else
      {
         fprintf( stderr, "unknown option [%c]\n", c);
         usage();
      }
   }
  
   if( ncpu < 1 || ncpu >= MAXCPU) bailout( "invalid number of cpus"); 
   if( OPT_spice && FLAG_NC) bailout( "cannot have -o spice and -NC");
   if( OPT_spice && FLAG_NL) bailout( "cannot have -o spice and -NL");
   if( OPT_spice && FLAG_NR) bailout( "cannot have -o spice and -NR");

   if( DIAG_bsag)
   {
      diag_BSAG();
      return 0;
   }

   if( DIAG_fmm)
   {
      Anm_setup();
      diag_FMM();
      return 0;
   }

   //
   // The remaining argument should be the system name.
   //

   if( optind + 1 != argc) usage();
   strcpy( WORLD.sysname, argv[optind]);

   setup_constants(); 
   parse_input_file();

   worker_init();

   decompose_coils();
   build_coil_electrodes();

   setup_tiles();
   setup_sizes();

   if( OPT_geda) output_geda();
   if( OPT_tiles) output_tiles();
   if( OPT_gnuplot) output_gnuplot();
   if( OPT_povray) output_povray();

   if( OPT_spice || OPT_list)
   {
      if( !FLAG_NC) 
      {
         if( WORLD.g) calculate_image_points();

         if( cp_method == CPM_FMM)
         {
            fill_octree();
            init_fmm();
            compute_potentials = compute_potentials_multi;
         }
         else compute_potentials = compute_potentials_direct;
      
         do_capacitance();
         if( WORLD.Efield) do_efield();
      }

      if( !FLAG_NL && WORLD.coils)
      {
         if( WORLD.g) create_image_coils();

         do_inductance();
         if( WORLD.Bfield) do_bfield();
      }

      if( !FLAG_NR) do_resistance();
   }

   if( OPT_spice) output_spice();
   if( OPT_list) output_listing();

   return 0;
}

